/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */

/* AREADEF - Set up a graphics area with scaling */
#include "pl.x"

static int nareas = 17;
static char *areas[17] = {
"standard     1.2 2.5 7.4 7.0   1.5 1.5 9.0 6.2",
"standardp    1.2 3.5 7.4 8.0   1.5 1.5 9.0 6.2",
"square       1.2 2.0 7.2 8.0   2.2 0.8 8.2 6.8",
"whole        1.2 1.0 7.4 9.0   1.5 1.2 9.0 7.0",
"2hi          1.0 5.5 7.6 9.0   1.0 4.5 10.0 7.0",
"2lo          1.0 1.0 7.6 4.5   1.0 1.0 10.0 3.5",
"2left        1.0 1.0 4.0 9.5   1.0 1.0 5.25 6.5",
"2right       5.0 1.0 8.0 9.5   6.25 1.0 10.5 6.5",
"3hi          1.0 7.0 7.6 9.0   1.0 5.5 10.0 7.5",
"3mid         1.0 4.0 7.6 6.0   1.0 3.0 10.0 5.0",
"3lo          1.0 1.0 7.6 3.0   1.0 0.5 10.0 2.5",
"4nw          1.0 6.0 4.0 9.0   1.0 4.0 5.25 7.0",
"4ne          4.5 6.0 7.5 9.0   6.25 4.0 10.5 7.0",
"4sw          1.0 1.5 4.0 4.5   1.0 0.5 5.25 3.5",
"4se          4.5 1.5 7.5 4.5   6.25 0.5 10.5 3.5",
"slide        1 0.7 5.85 3.2    1 0.7 5.85 3.2",
"lifetab      1.0 1.0 7.0 4.5   1.5 1.0 7.5 4.5",
};


areadef( specfile, clonefiles, nclonefiles )
char *specfile, clonefiles[][MAXPATH];
int nclonefiles;
{
int i;
FILE *fp;
char buf[256];
char attr[40], val[256];
char lineval[256];
int nt;
FILE *openclonefile();
int icf;

int stat;
double xlo, xhi, ylo, yhi;
char areaname[50];
char xscaletype[50], yscaletype[50];
char title[256];
char title2[256];
char titledet[256];
char title2det[256];
char frame[256];
int nlines, maxlen;
int doframe;
int namedarea;
int align;
double adjx, adjy;
int gotarea, gotxrange, gotyrange;
int doxaxis, doyaxis;
char areacolor[40];
char xminstr[120], xmaxstr[120];
char yminstr[120], ymaxstr[120];
double height, width;
int locspec;
double pltab_ymax;
char pltab_row_y[50];
char linebottom[256];
char lineside[256];
double autowf, autowmin, autowmax;
double autohf, autohmin, autohmax;
int xcatsgot, ycatsgot;

strcpy( Eprogname, "pl proc areadef" );
 

/* -------------------------- */
/* defaults */
/* -------------------------- */
strcpy( xscaletype, "linear" );
strcpy( yscaletype, "linear" );
strcpy( areaname, "standard" );
if( Device == 'c' || Device == 'p' ) strcpy( areaname, "standardp" ); /* for sheet of paper */
strcpy( title, "" );
strcpy( title2, "" );
doframe = 0;
strcpy( frame, "" );
strcpy( titledet, "" );
strcpy( title2det, "" );
namedarea = 1;
/* doxaxis = doyaxis = 1; */
doxaxis = doyaxis = 0;
strcpy( areacolor, "" );
strcpy( Bigbuf, "" );
height = 4.0;
width = 6.0;
xlo = 1.5;
ylo = 5.0;
locspec = 0; 
strcpy( xminstr, "" );
strcpy( xmaxstr, "" );
strcpy( yminstr, "" );
strcpy( ymaxstr, "" );
strcpy( pltab_row_y, "-0.2 0.2" );
strcpy( linebottom, "" );
strcpy( lineside, "" );
gotarea = 1;
Catcompmethod = 0;
autowf = 0.0; autowmin = 0.0; autowmax = 20.0;
autohf = 0.0; autohmin = 0.0; autohmax = 20.0;
xcatsgot = 0; ycatsgot = 0;

/* -------------------------- */
/* open and process all clonefiles and then lastly the specfile */
/* -------------------------- */


for( icf = 0; icf <= nclonefiles; icf++ ) {

    if( icf == nclonefiles ) {
	fp = fopen( specfile, "r" );
	if( fp == NULL ) return( Eerr( 117, "Could not open to read", specfile ) );
	}
    else	{
	fp = openclonefile( clonefiles[icf] );
	if( fp == NULL ) continue;
	}


    while( fgets( buf, 255, fp ) != NULL ) {
	nt = sscanf( buf, "%s %s", attr, val );		/* val has 2nd token */
	if( nt < 1 ) continue;
	if( nt == 1 ) strcpy( val, "" );

	/* skip #proc on clone files */
	if( icf != nclonefiles && stricmp( attr, "#proc" )==0 ) continue; 
	i = 0;
	Egetok( buf, &i );
	while( buf[i] == ' ' || buf[i] == '\t' ) i++; /* skip over white space */
        strcpy( lineval, &buf[ i ] ); /* lineval has remainder of line, for sscanf-ing */
        lineval[ strlen( lineval ) - 1 ] = '\0'; /* strip off newline */
 
	/* strip off trailing colon if any.. */
	if( attr[ strlen( attr ) - 1 ] == ':' ) attr[ strlen( attr ) - 1 ] = '\0';

	/* attributes.. */

	if( stricmp( attr, "areaname" )==0 ) {
		strcpy( areaname, val );
		gotarea = 1;
		}
	else if( stricmp( attr, "rectangle" )==0 ) {
		getbox( "rectangle", lineval, &xlo, &ylo, &xhi, &yhi );
		namedarea = 0;
		gotarea = 1;
		}
	else if( stricmp( attr, "box" )==0 ) {
		sscanf( lineval, "%lf %lf", &width, &height );
		if( Using_cm ) { width /= 2.54; height /= 2.54; }
		namedarea = 0;
		gotarea = 1;
		}	

	else if( stricmp( attr, "location" )==0 ) {
		sscanf( lineval, "%lf %lf", &xlo, &ylo );
		if( Using_cm ) { xlo /= 2.54; ylo /= 2.54; }
		locspec = 1;
		}	

	else if( stricmp( attr, "autowidth" )==0 ) {
		nt = sscanf( lineval, "%lf %lf %lf", &autowf, &autowmin, &autowmax );
		}
		
	else if( stricmp( attr, "autoheight" )==0 ) {
		nt = sscanf( lineval, "%lf %lf %lf", &autohf, &autohmin, &autohmax );
		}

	else if( stricmp( attr, "xrange" )==0 || stricmp( attr, "xautorange" )==0 ) {
		if( strnicmp( val, "datafield", 9 )==0 ) strcpy( xminstr, lineval );
		else	{
			nt = sscanf( lineval, "%s %s", xminstr, xmaxstr );
			if( nt != 2 ) {
				Skipout = 1;
				return( Eerr( 105, "both min and max expected", "xrange" ));
				}
			}
		gotxrange = 1;
		}
	else if( stricmp( attr, "yrange" )==0 || stricmp( attr, "yautorange" )==0 ) {
		if( strnicmp( val, "datafield", 9 )==0 ) strcpy( yminstr, lineval );
		else	{
			nt = sscanf( lineval, "%s %s", yminstr, ymaxstr );
			if( nt != 2 ) {
				Skipout = 1;
				return( Eerr( 105, "both min and max expected", "yrange" ));
				}
			}
		gotyrange = 1;
		}

	else if( stricmp( attr, "xscaletype" )==0 ) strcpy( xscaletype, lineval );
	else if( stricmp( attr, "yscaletype" )==0 ) strcpy( yscaletype, lineval );
	else if( stricmp( attr, "xcategories" )==0 ) {
		strcpy( xscaletype, "categories" );  /* added scg 10/25/99 */
		/* put category names into array slots now.. */
		if( strnicmp( val, "datafield", 9 )==0 ) strcpy( Bigbuf, lineval );
		else getmultiline( "xcategories", lineval, fp, MAXBIGBUF, Bigbuf );
		setcats( 'x', Bigbuf );
		strcpy( xminstr, "0" );
		sprintf( xmaxstr, "%d", Ncats[0]+1 );
		gotxrange = 1;
		xcatsgot = 1;
		}
	else if( stricmp( attr, "ycategories" )==0 ) {
		strcpy( yscaletype, "categories" ); /* added scg 10/25/99 */
		/* put category names into array slots now.. */
		if( strnicmp( val, "datafield", 9 )==0 ) strcpy( Bigbuf, lineval );
		else getmultiline( "ycategories", lineval, fp, MAXBIGBUF, Bigbuf );
		setcats( 'y', Bigbuf );
		strcpy( yminstr, "0" );
		sprintf( ymaxstr, "%d", Ncats[1]+1 );
		gotyrange = 1;
		ycatsgot = 1;
		}
	else if( stricmp( attr, "xextracategory" )==0 ) {
		if( !xcatsgot ) addcat( 'x', "pre", lineval );
		else addcat( 'x', "post", lineval );
		}
	else if( stricmp( attr, "yextracategory" )==0 ) {
		if( !ycatsgot ) addcat( 'y', "pre", lineval );
		else addcat( 'y', "post", lineval );
		}

	else if( stricmp( attr, "catcompmethod" )==0 ) {
		if( stricmp( val, "beginswith" )==0 ) Catcompmethod = 0;
		else if( stricmp( val, "exact" )==0 ) Catcompmethod = -1;
		else if( strnicmp( val, "length=", 7 )==0 )  Catcompmethod = atoi( &val[7] );
		else if( strnicmp( val, "fuzzy=", 6 )==0 )  Catcompmethod = -15 + atoi( &val[6] );
		}

	else if( stricmp( attr, "frame" )==0 ) {
		if( strnicmp( val, "no", 2 )==0 ) doframe = 0;
		else { strcpy( frame, lineval ); doframe = 1; }
		}
	else if( stricmp( attr, "title" )==0 ) {
		strcpy( title, lineval );
		Econvertnl( title );
		}
	else if( stricmp( attr, "title2" )==0 ) {
		strcpy( title2, lineval );
		Econvertnl( title2 );
		}
	else if( stricmp( attr, "titledetails" )==0 ) strcpy( titledet, lineval );
	else if( stricmp( attr, "title2details" )==0 ) strcpy( title2det, lineval );
	else if( stricmp( attr, "areacolor" )==0 ) strcpy( areacolor, val );
	else if( stricmp( attr, "pltabrowy" )==0 ) strcpy( pltab_row_y, lineval );
	else if( stricmp( attr, "linebottom" )==0 ) {
		if( strnicmp( val, "no", 2 )!=0 ) strcpy( linebottom, lineval );
		}
	else if( stricmp( attr, "lineside" )==0 ) {
		if( strnicmp( val, "no", 2 )!= 0 ) strcpy( lineside, lineval );
		}
	else if( stricmp( attr, "axes" )==0 ) {
		if( stricmp( val, "none" )==0 ) { doxaxis = 0; doyaxis = 0; }
		else if( stricmp( val, "x" )==0 ) doxaxis = 1;
		else if( stricmp( val, "y" )==0 ) doyaxis = 1;
		else if( stricmp( val, "both" )==0 ) { doxaxis = 1; doyaxis = 1; }
		}
	else if( slmember( attr, "?axis.stubs ?axis.selflocatingstubs" )) {
		if( strnicmp( attr, "xaxis.", 6 )==0 ) doxaxis = 1; 
		if( strnicmp( attr, "yaxis.", 6 )==0 ) doyaxis = 1; 
		if( slmember( val, "none inc* file datafield* list categories usecategories" )) ;
		else getmultiline( "stubtext", lineval, fp, 500, Bigbuf ); /* just to skip past it..*/
		}
	else if( strnicmp( attr, "xaxis.", 6 )==0 ) { doxaxis = 1; continue; }
	else if( strnicmp( attr, "yaxis.", 6 )==0 ) { doyaxis = 1; continue; }
	else if( slmember( attr, "axisline tic* minortic*" )) continue;
	else Eerr( 1, "areadef attribute not recognized", attr );
	}
    fclose( fp );

    }


/* -------------------------- */
/* check for degenerate conditions */
/* -------------------------- */



/* -------------------------- */
/* now do the plotting work.. */
/* -------------------------- */

resetstacklist();  /* (obscure) - reset the bar graph list that tries to
			keep track of fields for automatic stacking */

if( stricmp( yscaletype, "pltab" )==0 ) {
	pltab_plotarea( &xlo, &ylo, &xhi, &yhi, &pltab_ymax );
	gotarea = 1;
	gotyrange = 1;
	namedarea = 0;
	}

if( strnicmp( yscaletype, "pltabrow", 8 )==0 ) {
	pltab_plotarea( &xlo, &ylo, &xhi, &yhi, &pltab_ymax );
	/* still need yrange and area specs */
	gotarea = 1;
	namedarea = 0;
	}

if( stricmp( xscaletype, "categories" )==0 && Ncats[0] < 1 ) {
	Skipout = 1;
	return( Eerr( 5972, "xscaletype defined as categories but no xcategories given", "" ) );
	}

if( stricmp( yscaletype, "categories" )==0 && Ncats[1] < 1 ) {
	Skipout = 1;
	return( Eerr( 5973, "yscaletype defined as categories but no ycategories given", "" ) );
	}

if( locspec ) { /* location and box specified, calculate xhi and yhi.. */
	xhi = xlo + width;
	yhi = ylo + height;
	}

/* determine area.. */
if( namedarea ) {
	RETRY:
	strcat( areaname, " " );
	for( i = 0; i < nareas; i++ ) {
		if( strnicmp( areas[i], areaname, strlen( areaname ) )==0 ) {
			if( Ecurpaper == 0 ) sscanf( areas[i], "%*s %lf %lf %lf %lf", &xlo, &ylo, &xhi, &yhi );
			else if( Ecurpaper == 1 ) sscanf( areas[i], "%*s %*s %*s %*s %*s %lf %lf %lf %lf", &xlo, &ylo, &xhi, &yhi );
			break;
			}
		}
	if( i == nareas ) {
		areaname[ strlen( areaname ) -1 ] = '\0';
		Eerr( 110, "warning, areaname not recognized, using 'standard'..", areaname );
		strcpy( areaname, "standard" );
		goto RETRY;
		}
	}
		
if( !gotarea && !locspec ) {  /* && !locspec added scg 11/21/00 */
	Skipout = 1;
	return( Eerr( 130, "No plotting area has been specified", "" ) );
	}
if( !gotxrange ) {
	Skipout = 1;
	return( Eerr( 130, "No xrange has been specified", "" ) );
	}
if( !gotyrange ) {
	Skipout = 1;
	return( Eerr( 130, "No yrange has been specified", "" ) );
	}

/* set scaling type and special units if any.. */
stat = Escaletype( xscaletype, 'x' );
if( stat != 0 ) Escaletype( "linear", 'x' );
if( strnicmp( xscaletype, "log", 3 )==0 ) Esetunits( 'x', "linear" );
else Esetunits( 'x', xscaletype );

stat = Escaletype( yscaletype, 'y' );
if( stat != 0 ) Escaletype( "linear", 'y' );
if( strnicmp( yscaletype, "log", 3 )==0 ) Esetunits( 'y', "linear" );
else Esetunits( 'y', yscaletype );


/* if autoranging is specified, do it now.. */
if( strnicmp( xminstr, "datafield", 9 ) == 0 ) autorange( 'x', xminstr, xminstr, xmaxstr );
if( strnicmp( yminstr, "datafield", 9 ) == 0 ) autorange( 'y', yminstr, yminstr, ymaxstr );

/* if scaletype is log but there are data values = 0.0, go to log+1 scaling.. */
if( stricmp( xscaletype, "log" )==0 && atof( xminstr ) <= 0.0 ) Escaletype( "log+1", 'x' );
if( stricmp( yscaletype, "log" )==0 && atof( yminstr ) <= 0.0 ) Escaletype( "log+1", 'y' );

/* if using autowidth or autoheight, revise plotting area now.. */
if( autowf != 0.0 ) {
	double xmin, xmax;
	if( strncmp( xscaletype, "categories" )== 0 ) {
		xmin = atof( xminstr );
		xmax = atof( xmaxstr );
		}
	else	{
		xmin = Econv( 'x', xminstr );
		xmax = Econv( 'x', xmaxstr );
		}
	xhi = xlo + ((xmax-xmin) * autowf);
	if( xhi - xlo < autowmin ) xhi = xlo + autowmin;
	else if( xhi - xlo > autowmax ) xhi = xlo + autowmax;
	}
if( autohf != 0.0 ) {
	double ymin, ymax;
	if( strncmp( yscaletype, "categories" )== 0 ) {
		ymin = atof( yminstr );
		ymax = atof( ymaxstr );
		}
	else	{
		ymin = Econv( 'y', yminstr );
		ymax = Econv( 'y', ymaxstr );
		}
	yhi = ylo + ((ymax-ymin) * autohf);
	if( yhi - ylo < autohmin ) yhi = ylo + autohmin;
	else if( yhi - ylo > autohmax ) yhi = ylo + autohmax;
	}

	


/* set the scaling for the plot area.. */
stat = Esetscale( 'x', xlo, xhi, xminstr, xmaxstr );

if( strnicmp( yscaletype, "pltab", 5 )==0 )  /* (special case) */
	stat += Escale_y( ylo, yhi, 0.0, pltab_ymax ); 
else stat += Esetscale( 'y', ylo, yhi, yminstr, ymaxstr );

if( Debug ) fprintf( Diagfp, "// areadef: lowerleft: %g,%g  upperright: %g,%g\n",
		xlo, ylo, xhi, yhi );
if( Debug ) fprintf( Diagfp, "// areadef:   xrange is %s to %s.   yrange is %s to %s.\n",
		xminstr, xmaxstr, yminstr, ymaxstr );

if( stat != 0 ) return( stat );

/* handle pltabrow (a variant of pltab) .. */
if( strnicmp( yscaletype, "pltabrow", 8 )==0 ) {
	char rowval[50], subscaletype[50];
	double ry, pltr_lo, pltr_hi;

	/* get row value and subscale type  */
	nt = sscanf( yscaletype, "%*s %s %s", rowval, subscaletype );
	if( nt == 1 ) strcpy( subscaletype, "linear" );
	/* get absolute Y position of center of requested row */
	ry = Eay( Econv( Y, rowval ) );
	if( Econv_error() ) {
                fprintf( Diagfp, "ignoring row %s\n", rowval );
                return( 0 );
                }
	stat = Escaletype( subscaletype, 'y' );
	if( stat != 0 ) Escaletype( "linear", 'y' );
	if( strnicmp( subscaletype, "log", 3 )==0 ) Esetunits( 'y', "linear" );
	else Esetunits( 'y', subscaletype );
	nt = sscanf( pltab_row_y, "%lf %lf", &pltr_lo, &pltr_hi );
	if( nt != 2 ) { pltr_lo = -0.2; pltr_hi = 0.2; }
	stat = Esetscale( 'y', ry+pltr_lo, ry+pltr_hi, yminstr, ymaxstr );
	}
	
suppress_twin_warn( 0 );

/* set DMS variables to hold plot area bounds and scale min/max.. */
setfloatvar( "AREATOP", EYhi );
setfloatvar( "AREABOTTOM", EYlo );
setfloatvar( "AREALEFT", EXlo );
setfloatvar( "AREARIGHT", EXhi );
setcharvar( "XMIN", xminstr );
setcharvar( "XMAX", xmaxstr );
setcharvar( "YMIN", yminstr );
setcharvar( "YMAX", ymaxstr );


if( strlen( areacolor ) > 0 ) {
	Ecblock( EXlo, EYlo, EXhi, EYhi, areacolor, 0 );
	}

/* draw a frame of the graphics area */
if( doframe ) {
	if( stricmp( frame, "bevel" )==0 ) {
		Ecblockdress( EXlo, EYlo, EXhi, EYhi, 0.1, "0.6", "0.8", 0.0, "" );
		}
	else	{
		linedet( "frame", frame, 1.0 );
		Ecblock( EXlo, EYlo, EXhi, EYhi, "", 1 );
		}
	}


/* draw bottom line */
if( strlen( linebottom ) > 0 ) {
	linedet( "linebottom", linebottom, 1.0 );
	Emov( EXlo, EYlo );
	Elin( EXhi, EYlo );
	}

/* draw side line */
if( strlen( lineside ) > 0 ) {
	linedet( "lineside", lineside, 1.0 );
	Emov( EXlo, EYlo );
	Elin( EXlo, EYhi );
	}

/* title */
if( strlen( title ) > 0 ) {
	textdet( "titledetails", titledet, &align, &adjx, &adjy, 0, "R" );
	Emeasuretext( title, &nlines, &maxlen );
	if( align == '?' ) align = 'L';
	if( align == 'L' ) Emov( EXlo + adjx, 
				EYhi + ((nlines-1) * Ecurtextheight ) + adjy + 0.02 );
	else if( align == 'C' ) 
		Emov( EXlo+((EXhi-EXlo) / 2.0 ) + adjx, 
			EYhi + ((nlines-1) * Ecurtextheight ) +adjy + 0.02 );
	else if( align == 'R' ) 
		Emov( EXhi + adjx, EYhi + ((nlines-1) * Ecurtextheight ) +adjy + 0.02);
	Edotext( title, align );
	}
/* title 2 */
if( strlen( title2 ) > 0 ) {
	textdet( "title2details", title2det, &align, &adjx, &adjy, 0, "R" );
	Emeasuretext( title2, &nlines, &maxlen );
	if( align == '?' ) align = 'L';
	if( align == 'L' ) Emov( EXlo + adjx, 
		EYhi + ((nlines-1) * Ecurtextheight ) + adjy + 0.02 );
	else if( align == 'C' ) 
		Emov( EXlo+((EXhi-EXlo) / 2.0 ) + adjx, 
			EYhi + ((nlines-1) * Ecurtextheight ) +adjy + 0.02 );
	else if( align == 'R' ) Emov( EXhi + adjx, 
		EYhi + ((nlines-1) * Ecurtextheight ) +adjy + 0.02 );
	Edotext( title2, align );
	}


if( doxaxis ) axis( specfile, clonefiles, nclonefiles, 'x', 6 );

if( doyaxis ) axis( specfile, clonefiles, nclonefiles, 'y', 6 );

return( 0 );
}
