/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#ifdef PLOTICUS
#include "dms.d"
#endif
/*
condex()
This routine takes a "condex" conditional expression as a single string argument.

* Field specifiers, variables, etc. must be converted to literals before 
  calling condex.

* If the expression is true, 1 is returned.
  If the expression is false, 0 is returned.
  If there was an error in the expression, -1 is returned.
  A 0-length expression is always true.

  For condex expression syntax, do man dmscondex

* Any programs using this module need to include the following function stub
	custom_function( name, parms, value )
	char *name, *parms, *value;
	{ return( -1 ); }
  This function can be expanded to hold definitions for application-specific
  functions if desired - see eval_function() herein as a guide.  In general
  the code for each function should return 0 if successful, -1 if not.

* Programs can retrieve the value of the previous condex expression (usually
  used when functions are involved which produce a result).  To do this, call
  this routine:  char *condex_results().  The results are returned.

*/
#include <stdio.h>
#define NUMBER 0
#define ALPHA 1

#define NCLAUSE 12
#define NTOKS 30
#define MAXTOK 255

static char LastResults[256] = "";

static int evalclause(), evalstmt(), yield();

/* ================================== */

condex( cond )
char cond[];
{
int s[NCLAUSE], i, j, k, rtn, ii, ix, negate;
char args[NTOKS][MAXTOK], *getok(), tok[MAXTOK];
int argc;
int ixtmp;
char nexttok[MAXTOK];
int typ, stat, p;
double atof();


strcpy( LastResults, "" );

/* break cond into tokens */
ix = 0;
/* check for leading "not:" */
strcpy( tok, getok( cond, &ix ) );
if( strcmp( tok, "not:" )==0 ) negate = 1;
else 	{
	negate = 0;
	ix = 0;
	}

for( i = 1; ; i++ ) {
	strcpy( args[ i ], getok( cond, &ix ) );
	if( args[i][0] == '\0' ) break;
	}
argc = i;


/* special function $exists() - if there is anything between the parens other than
   white space or BLANKFIELD, return 1.  Otherwise return 0.
   This is the only "function" that can take args containing white space.
   It must be alone in the expression and may not be NOTed or ANDed / ORed with any other expression. */
if( strncmp( args[1], "$exists(", 8 )==0 || strncmp( args[1], "$notexists(", 11 )==0 ) {
	int eflag, found;  /* found indicates that we've found a non-space, non-missing character */
	if( args[1][1] == 'n' ) eflag = 0;
	else eflag = 1;

	if( strcmp( args[1], "$exists(=)" )==0 ) return( 0 );
	else if( strcmp( args[1], "$exists()" )==0 ) return( 0 );
	else if( strcmp( args[1], "$notexists(=)" )==0 ) return( 1 );
	else if( strcmp( args[1], "$notexists()" )==0 ) return( 1 );

	for( i = 0; i < strlen( cond ); i++ ) if( cond[i] == '(' ) break;
	for( j = i+1, found = 0; j < strlen( cond ); j++ ) {
		if( cond[j] == ')' ) break;
		else if( cond[j] != ' ' && cond[j] != '\t' && cond[j] != '=' ) { found = 1; break; }
		}
	if( eflag ) return( found );
	else return( !found );
	}

#ifdef CUT

strcpy( nexttok, "" );
sscanf( cond, "%s %s", tok, nexttok );
if( strncmp( tok, "$exists(", 8 )==0 ) {
	if( strlen( tok ) == 9 ) return( 0 ); /* $exists() */
	else if( strlen( tok ) < 9 && strcmp( nexttok, ")" )==0 ) return( 0 ); /* $exists(  ) */
	else if( strcmp( tok, "$exists(=)" )==0 ) return( 0 );            /* $exists(=) */
	else if( strlen( tok ) > 9 && strlen( nexttok )==0 ) return( 1 ); /* $exists(xxx) */
	else if( strlen( tok ) > 8 && strcmp( nexttok, ")" )!=0 ) return( 1 ); /* $exists(xx yy) */
	else return( 1 ); /* ? */
	}
else if( strncmp( tok, "$notexists(", 11 )==0 ) {
	if( strlen( tok ) == 12 ) return( 1 ); /* $notexists() */
	else if( strcmp( tok, "$notexists(=)" )==0 ) return( 1 );            /* $notexists(=) */
	else if( strlen( tok ) < 12 && strcmp( nexttok, ")" )==0 ) return( 1 ); /* $notexists(  ) */
	else if( strlen( tok ) > 12 && strlen( nexttok )==0 ) return( 0 ); /* $notexists(xxx) */
	else if( strlen( tok ) > 11 && strcmp( nexttok, ")" )!=0 ) return( 0 ); /* $notexists(xx yy) */
	else return( 0 ); /* ? */
	}
#endif  /* CUT */




/* if there is just one argument, it must be a function.
   Just evalute, and store the value in static variable lastresults. */
if( argc == 2 ) {
	stat = yield( args[1], &typ );
	if( stat == -1 ) { 
		err( 15001, 0, "expression error", cond );
		return( -1 ); 
		}
	if( typ == NUMBER ) { /* if evaluates to 0, function returns 0; otherwise 1 */
		sprintf( LastResults, "%g", atof(args[1]) );
		if( strcmp( LastResults, "0" )==0 ) return( 0 );
		else return( 1 );
		}
	strcpy( LastResults, args[1] );
	return( 1 );
	}
	


/* do "clauses" */
for( i = 0; i < NCLAUSE; i++ ) s[i] = 0;
i = 0;
j = k = 1;
while( 1 ) {
	if( j==argc && strcmp( args[j], "||" )==0 ) { 
		err( 15002, 0, "expression error", cond );
		return( -1 ); 
		}
	if( j == argc || strcmp( args[j], "||" )==0 ) {
		s[ i ] = evalclause( args, k, j-1 );
		if( s[ i ] == -1 ) {
			err( 15003, 0, "expression error", cond );
			return( -1 );
			}
		k = j+1;
		i++;
		}
	j++;
	if( j > argc ) break;
	}
rtn = ( s[0] || s[1] || s[2] || s[3] || s[4] || s[5] || s[6] || s[7] || s[8] || s[9] || s[10] || s[11] );
if( negate ) return( ! rtn );
else return( rtn );
}

/* ================ */
/* EVALCLAUSE - evaluate a clause */
static int
evalclause( args, start, stop )
char args[NTOKS][MAXTOK];
int start, stop;
{
int s[NCLAUSE], i, j, k, rtn, ii;
char stmt[80];

for( i = 0; i < NCLAUSE; i++ ) s[i] = 1;
i = 0;
j = k = start;
while( 1 ) {
	if( j==stop && strcmp( args[j], "&&" )==0 ) {  return( -1 ); }

	if( j==stop || strcmp( args[j], "&&" )==0 ) {
		s[ i ] = evalstmt( args, k );
		if( s[ i ] == -1 ) return( -1 );
		k = j+1;
		i++;
		}
	j++;
	if( j > stop ) break;
	}
rtn = ( s[0] && s[1] && s[2] && s[3] && s[4] && s[5] && s[6] && s[7] && s[8] && s[9] && s[10] && s[11] );
return( rtn );
}

/* ============ */
/* EVALSTMT - evaluate a statement */
static int
evalstmt( args, start )
char args[NTOKS][MAXTOK];
int start;
{
char r1[MAXTOK], r2[MAXTOK], op[MAXTOK], val1[MAXTOK], val2[MAXTOK];
int i, t1, t2, i1, i2;
double diff;
double atof(), a2f();


strcpy( r1, args[start] ); 
strcpy( op, args[start+1] );
strcpy( r2, args[start+2] );
if( smember( r2, "&& ||" ) || r2[0] == '\0' ) { return( -1 ); } /* got off track */


/* assign data type for value-- types are alpha, number, date */
i1 = yield( r1, &t1 );
i2 = yield( r2, &t2 );
if( i1 < 0 || i2 < 0 ) return( -1 );

if( t1 == NUMBER && t2 == NUMBER ) diff = atof( r1 ) - atof( r2 );
else diff = strcmp( r1, r2 );
	
/* number compared against alpha always unequal */
if(( smember( op, "ne !=" )) &&  ((t1 == NUMBER && t2 == ALPHA) || (t1 == ALPHA && t2 == NUMBER )) ) 
	return( 1 );

else if( smember( op, "in" )) {
	for( i = 0; i < strlen( r2 ); i++ ) if( r2[i] == ',' ) r2[i] = ' ' ;
	return( smember( r1, r2 ) );
	}
else if( smember( op, "!in ni" )) {
	for( i = 0; i < strlen( r2 ); i++ ) if( r2[i] == ',' ) r2[i] = ' ' ;
	return( ! smember( r1, r2 ) );
	}

else if( smember( op, "like" )) return( ! wildcmp( r1, r2, strlen( r2 ), 0 ) );
else if( smember( op, "!like" )) return( wildcmp( r1, r2, strlen( r2 ), 0 ) );
else if( smember( op, "inlike" )) {
	for( i = 0; i < strlen( r2 ); i++ ) if( r2[i] == ',' ) r2[i] = ' ' ;
	return( slmember( r1, r2 ) );
	}
else if( smember( op, "!inlike" )) {
	for( i = 0; i < strlen( r2 ); i++ ) if( r2[i] == ',' ) r2[i] = ' ' ;
	return( ! slmember( r1, r2 ) );
	}


/* for the rest of the relational operators, operands must be of same type for comparison .. */
else if( (t1 == NUMBER && t2 == ALPHA) || (t1 == ALPHA && t2 == NUMBER ) ) return( 0 );

else if( smember( op, "eq = ==" )) { if( diff == 0 ) return( 1 ); else return( 0 ); }
else if( smember( op, "ne !=" )) { if( diff != 0 ) return( 1 ); else return( 0 ); }
else if( smember( op, "gt >" )) { if( diff > 0 ) return( 1 ); else return( 0 ); }
else if( smember( op, "ge >=" )) { if( diff >= 0 ) return( 1 ); else return( 0 ); }
else if( smember( op, "lt <" )) { if( diff < 0 ) return( 1 ); else return( 0 ); }
else if( smember( op, "le <=" )) { if( diff <= 0 ) return( 1 ); else return( 0 ); }

return( -1 );
}


/* ========================= */
/* YIELD - determine data type, evaluate functions.
   		Yield is a recursive function. 
   		Returns -1 for bad expression */
static int
yield( v, t )
char v[];
int *t;
{
int t2, i, i1, p;
double atof(), a2f();
int status;


/* if v is a $function call, evaluate it .. */
if( v[0] == '$' && isalpha( v[1] ) ) {
	/* fprintf( stderr, "[%s =", v ); */ /* For function tracing uncomment these 2 lines */
	status = function_call( v ); /* v will be modified here */
	/* fprintf( stderr, " %s]\n", v ); */
	if( status == -1 ) return( -1 );
	}

/* figure out what type v is */
if( v[ strlen( v )-1 ] == ' ' ) { *t = ALPHA; v[ strlen( v ) -1 ] = '\0'; } /* forced alpha */
										
else if( goodnum( v, &p ) ) *t = NUMBER;
else *t = ALPHA;


return( 1 );
}


/* ============================= */
/* CONDEX_RESULTS - make results of most recent condex expression available. */
char *
condex_results()
{
return( LastResults );
}

