/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include "pl.x"

/* ========================== */
/* DEVAVAIL - returns 1 if specified device driver is available
	in this build, or 0 if not. */
devavail( dev )
char *dev;
{
if( dev[0] == 'x' ) {
#ifdef NOX11
	return( 0 );
#endif
	return( 1 );
	}

else if( dev[0] == 'g' ) {
#ifdef GDPNG
	return( 0 );
#endif
	return( 1 );
	}

else if( strcmp( dev, "png" )==0 ) {
#ifdef GDGIF
	return( 0 );
#endif
	return( 1 );
	}
else if( Emember( dev[0], "pce" )) return( 1 );
return( 0 );
}


/* ========================== */
/* DEVNAMEMAP - translate between internal and external 
	representation of output device code */
devnamemap( s, t, mode )
char *s; /* internal (Device) */
char *t; /* external (command line opt[1] or DEVICE */
int mode; /* 1 = set s given t    2 = set t given s */
{
char c;

if( mode == 1 ) {
	c = tolower(t[0]);

	/* old monochrome ps */
	if( c == 'b' ) *s = 'p';

	/* eps */
	else if( c == 'e' ) *s = 'e';

	/* x11 */
	else if( c == 'x' ) {
		*s = 'x';
		if( !devavail( "x" ) ) {
			fprintf( Errfp, "X11 not supported in this build.\n" );
			exit(1);
			}
		}

	/* gif */
	else if( c == 'g' ) {
		*s = 'g';
		if( !devavail( "g" ) ) {
			fprintf( Errfp, "GIF not supported in this build, try -png.\n" );
			exit(1);
			}
		}

	/* png */
	else if( strcmp( t, "png" )==0 ) { 
		*s = 'g'; 
		if( !devavail( "png" )) {
			fprintf( Errfp, "PNG not supported in this build, try -gif.\n" );
			exit(1);
			}
		Png = 1; 
		}

	/* postscript */
	else if( c == 'p' ) *s = 'c';

	else Eerr( 8026, "unrecognized device rep", t );
	return( 0 );
	}
else if( mode == 2 ) {
	if( *s == 'g' ) {
		if( Png ) strcpy( t, "png" );
		else strcpy( t, "gif" );
		}
	else if( *s == 'p' ) strcpy( t, "bwps" );
	else if( *s == 'c' ) strcpy( t, "ps" );
	else if( *s == 'e' ) strcpy( t, "eps" );
	else if( *s == 'x' ) strcpy( t, "x11" );
	else Eerr( 8025, "unrecognized internal device rep", "" );
	return( 0 );
	}
else Eerr( 8027, "invalid devnamemap mode", "" );
}

/* ============================== */
/* MAKEOUTFILENAME - given script file name, make a default output file
   for when no output file is specified by user.

   The rule is:
    If scriptfile ends in .p, substitute either .eps or .gif for .p 
    Otherwise use out.eps or out.gif.

   Returns 0 if ok; 1 if usage error 
*/
makeoutfilename( scriptfn, outfn, dev )
char *scriptfn, *outfn;
char dev;
{
int len;

len = strlen( scriptfn );
if( strcmp( &scriptfn[ len-2 ], ".p" )==0 ) len -= 2;
else if( strcmp( &scriptfn[ len-3 ], ".pl" )==0 ) len -= 3;
else if( strcmp( &scriptfn[ len-4 ], ".htm" )==0 ) len -= 4;
else if( strcmp( &scriptfn[ len-4 ], ".pls" )==0 ) len -= 4;
else if( strcmp( &scriptfn[ len-5 ], ".html" )==0 ) len -= 5;

if( smember( &scriptfn[ len ], ".p .pl .pls .htm .html" )) {
	if( dev == 'g' ) {
		strcpy( outfn, scriptfn );
		if( Png ) strcpy( &outfn[ len ], ".png" );
		else strcpy( &outfn[ len ], ".gif" );
		return( 0 );
		}
	else if( dev == 'e' ) {
		strcpy( outfn, scriptfn );
		strcpy( &outfn[ len ], ".eps" );
		return( 0 );
		}
	}
else 	{
	if( dev == 'g' ) { 
		if( Png ) strcpy( outfn, "out.png" );
		else strcpy( outfn, "out.gif" ); 
		return( 0 ); 
		}
	else if( dev == 'e' ) { 
		strcpy( outfn, "out.eps" ); 
		return( 0 ); 
		}
	}
return( 1 );
}

