/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* DRAWCOMMANDS - execute a list of draw commands */
#include "pl.x"

static int do_drawcommands();

drawcommands( specfile, clonefiles, nclonefiles )
char *specfile, clonefiles[][ MAXPATH ];
int nclonefiles;
{
int i;
FILE *fp;
char buf[256];
char attr[40], val[256];
char lineval[256];
int nt;
FILE *openclonefile();
int icf;

int stat;
int align;
double adjx, adjy;

char dcfile[256];


strcpy( Eprogname, "pl proc drawcommands" );
/* -------------------------- */
/* DEFAULTS */
/* All variables must start out with default values, */
/* since no variable is guaranteed to be set */
/* -------------------------- */
strcpy( dcfile, "" );
strcpy( Bigbuf, "" );

/* -------------------------- */
/* open and process clonefile and specfile */
/* -------------------------- */
for( icf = 0; icf <= nclonefiles; icf++ ) {
 
    if( icf == nclonefiles ) {
        fp = fopen( specfile, "r" );
        if( fp == NULL ) return( Eerr( 117, "Could not open to read", specfile ) );
        }
    else        {
        fp = openclonefile( clonefiles[icf] );
        if( fp == NULL ) continue;
        }


    while( fgets( buf, 255, fp ) != NULL ) {
	nt = sscanf( buf, "%s %s", attr, val );		/* val has 2nd token */
	if( nt < 1 ) continue;
	if( nt == 1 ) strcpy( val, "" );
	if( icf < nclonefiles && stricmp( attr, "#proc" )==0 ) continue;
	i = 0;
        Egetok( buf, &i );
	while( buf[i] == ' ' || buf[i] == '\t' ) i++; /* skip over white space */
        strcpy( lineval, &buf[ i ] ); /* lineval has remainder of line, for sscanf-ing */
        lineval[ strlen( lineval ) - 1 ] = '\0'; /* strip off newline */
 
	/* strip off trailing colon if any.. */
	if( attr[ strlen( attr ) - 1 ] == ':' ) attr[ strlen( attr ) - 1 ] = '\0';


	/* attributes.. */

	if( stricmp( attr, "file" )==0 ) strcpy( dcfile, val );

	else if( stricmp( attr, "commands" )==0 ) {
		getmultiline( "commands", lineval, fp, 50000, Bigbuf );
		}

	else Eerr( 1, "drawcommands attribute not recognized", attr );
	}
    fclose( fp );
    }


/* -------------------------- */
/* overrides and degenerate cases */
/* -------------------------- */



/* -------------------------- */
/* now do the plotting work.. */
/* -------------------------- */
if( strlen( Bigbuf ) > 0 ) {
	sprintf( buf, "%s_Z", Tmpname );
	fp = fopen( buf, "w" );
	if( fp == NULL ) return( Eerr( 522, "Cannot open draw commands file", buf ));
	fprintf( fp, "%s\n", Bigbuf );
	fclose( fp );
	do_drawcommands( buf );
	}
else if( strlen( dcfile ) > 0 ) do_drawcommands( dcfile );
return( 0 );
}

/* ===================================== */
/* DO_DRAWCOMMANDS - draw commands interpreter.
   Similar to ../src/drawfile.c but replicated here to include operations
	in scaled space, and special units conversions.

	The commands follow the structure of the API,
	except that movs lins and paths are used instead of movu, linu, and pathu
*/
static int
do_drawcommands( filename )
char *filename;
{
FILE *fp;
char buf[512];
int nt;
char op[80];
double x, y;
int oldtextsize;
char oldcolor[40];
int oldlinetype;
double oldlinewidth, oldpatfact;
char sx[80], sy[80];

if( strlen( filename ) < 0 ) return( 0 );

fp = fopen( filename, "r" );
if( fp == NULL ) return( 0 );
               
oldtextsize = Ecurtextsize;
strcpy( oldcolor, Ecurcolor );
oldlinetype = Ecurlinetype;
oldlinewidth = Ecurlinewidth;
oldpatfact = Ecurpatternfactor;
Etextsize( 12 );
Ecolor( "black" );
Elinetype( 0, 1.0, 1.0 );

while( fgets( buf, 511, fp ) != NULL ) {
	buf[ strlen( buf ) - 1 ] = '\0';

	nt = sscanf( buf, "%s %lf %lf", op, &x, &y );	

	if( nt < 1 ) continue; /* blank line */
	if( op[0] == '/' && op[1] == '/' ) continue; /* comment */

	if( smember( op, "mov lin" ) && nt != 3 ) { 
		if( nt != 6 ) Eerr( 2582, "drawfile error on this line", buf );
		continue;
		}

	if( strcmp( op, "mov" )==0 ) Emov( x, y );
	else if( strcmp( op, "lin" )==0 ) Elin( x, y );
	else if( strcmp( op, "text" )==0 ) {
		Econvertnl( &buf[5] );
		Etext( &buf[5] );
		}	
	else if( strcmp( op, "path" )==0 ) Epath( x, y );
	else if( smember( op, "movs lins paths" )) { /* scaled space operators */
		sscanf( buf, "%*s %s %s", sx, sy );
		if( op[0] == 'm' ) Emov( Eu(X, sx ), Eu(Y, sy ) );
		else if( op[0] == 'l' ) Elin( Eu(X, sx ), Eu(Y, sy ) );
		else if( op[0] == 'p' ) Epath( Eu(X, sx ), Eu(Y, sy ) );
		}
	else if( smember( op, "movp linp pathp" )) { /* posex operators */
		sscanf( buf, "%*s %s %s", sx, sy );
		Eposex( sx, X, &x );
		Eposex( sy, Y, &y );
		if( op[0] == 'm' ) Emov( x, y );
		else if( op[0] == 'l' ) Elin( x, y );
		else if( op[0] == 'p' ) Epath( x, y );
		}
	else if( strcmp( op, "fill" )==0 ) Efill();
	else if( strcmp( op, "centext" )==0 ) {
		Econvertnl( &buf[8] );
		Ecentext( &buf[8] );
		}
	else if( strcmp( op, "rightjust" )==0 ) {
		Econvertnl( &buf[10] );
		Erightjust( &buf[10] );
		}	
	else if( strcmp( op, "cblock" )==0 ) {
		double x1, y1, x2, y2;
		char color[40];
		int outline;
		nt = sscanf( buf, "%*s %lf %lf %lf %lf %s %d",
			&x1, &y1, &x2, &y2, color, &outline );
		if( nt != 6 ) Eerr( 2849, "drawfile error on this line", buf );
		else Ecblock( x1, y1, x2, y2, color, outline );
		}
	else if( strcmp( op, "color" )==0 ) Ecolor( &buf[6] );
	else if( strcmp( op, "textsize" )==0 ) Etextsize( atoi( &buf[9] ) );
	else if( strcmp( op, "linetype" )==0 ) {
		double z;
		sscanf( buf, "%*s %*s %*s %lf", &z );
		Elinetype( (int)x, y, z );
		}
	else if( strcmp( op, "mark" )==0 ) {
		char sym[80];
		double r;
		sscanf( buf, "%*s %s %s %s %lf", sx, sy, sym, &r );
		Eposex( sx, X, &x );
		Eposex( sy, Y, &y );
		Emark( x, y, sym, r );
		}
	else if( strcmp( op, "clr" )==0 ) Eclr(); /* allow? */
	else if( strcmp( op, "bkcolor" )==0 ) Ebackcolor( &buf[8] ); /* persist? */
	}
fclose( fp );
Eflush();
/* restore previous settings.. */
Ecolor( oldcolor );
Etextsize( oldtextsize );
Elinetype( oldlinetype, oldlinewidth, oldpatfact );
return( 0 );
}
