/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* EXEC_PLFILE - execute one ploticus control file.
 *
 *  This is called by pl.c and by pltab.c
 *
 *  ED* variables should be initialized before call.
 *
 *  Returns 0 if ok, or a non-zero error code 
 */
#include "pl.x"

#define NONDRAWINGPROCS "page print originaldata tabulate getdata defineunits rangebar catslide processdata datesettings endproc"

exec_plfile( specfile )
char *specfile;
{
int i, j;
char buf[256];
char buf2[256];
char tmpfilename[120];
int stat;
struct scripterstate ss;
int musthave_fail; /* so that all #musthave errors display before quitting */
int first;
char procname[40];
FILE *fp;
int nt;
char clonefiles[MAXCLONEFILES][MAXPATH];
int nclonefiles;
int lastbs;

char data[50][DATAMAXLEN+1]; /* to appease scripter()  - but not used for anything.. */
char recordid[80];	    /* '' */
char saveas_name[80];
char last_proctok[20]; /* either #proc or #procdef */

strcpy( recordid, "" ); /* BSD */

/* open spec file.. */
stat = scripter_open( &ss, specfile );
if( stat != 0 ) {
	Eerr( 22, "Cannot open scriptfile", specfile );
	vermessage();
	}

Npages = 0;


musthave_fail = 0;
first = 1;
fp = NULL;
lastbs = 0;
nclonefiles = 0;
Dsize = 0;
Dsel = 0;
sprintf( tmpfilename, "%s_K", Tmpname );
strcpy( saveas_name, "" );
strcpy( last_proctok, "" );


/* data array initializations.. */
ND[0] = ND[1] = 0;
Nrecords[0] = Nrecords[1] = 0;
Nfields[0] = Nfields[1] = 0;
StartD[0] = StartD[1] = 0;

/* also initialize the do-nothing 'data' array... */
for( i = 0; i < 50; i++ ) strcpy( data[i], "" );

/* Begin reading through spec file using scripter..
   Get one #proc chunk at a time, put it in tmpfile, then execute 
	appropriate plot routine.. */

while( 1 ) {
	/* do script processing.. */
	stat = scripter( &ss, buf, recordid, data );

	if( stat == 408 ) { 
	 	musthave_fail = 1; 
	 	continue; 
		}
        else if( stat > 255 || musthave_fail ) 
		return( Eerr( 23, "Fatal script processing error.", "" ) );

	else if( Skipout ) { /* error that is fatal for this script */
		Skipout = 0;
		return( 1 );
		}

        else if( stat != 0 ) break;

	/* if( Debug ) fprintf( Diagfp, ">>>%s", buf ); */

	nt = sscanf( buf, "%s", buf2 );

	if( nt > 0 && strnicmp( buf2, "#proc", 5 )==0 && !lastbs ) {  /* #proc or #procdef break */

		if( ! first && fp != NULL ) {  

		    fclose( fp );  /* close the tmpfile.. */

		    /* if saveas was specified, do it here.. */
		    if( strlen( saveas_name ) > 0 ) saveas( tmpfilename, saveas_name );

		    if( stricmp( last_proctok, "#proc" )==0 ) {  /* execute proc 
								   that was just read.. */

			/* proc page can do the Einit, or we can do it here.. */
			if( stricmp( procname, "page" )==0 ) Initialized = 1;
			if( !Initialized && !slmember( procname, NONDRAWINGPROCS )) {
				Einit( Inputfile, Device );
				Epaper( Landscape );
				Initialized = 1;

				/* we are ready to draw.. safe to say page 1.. scg 11/28/00 */
				if( stricmp( procname, "page" )!=0 ) Npages = 1;


				if( Bkcolorgiven ) {
					/* EPS color=transparent - best to do nothing.. scg 1/10/00*/
					if( Device == 'e' && strcmp( Ecurbkcolor, "transparent" )==0 ) ;

					else Eclr(); 
					}
				}


			/* ------------------------------------------------- */
			/* now execute the appropriate plotting procedure... */
			/* ------------------------------------------------- */
			proc_call( procname, tmpfilename, clonefiles, nclonefiles );
			if( Initialized ) Eflush(); /* get output onto screen.. */
			}
		    }

		/* if( stricmp( buf2, "#procend" )==0 ) continue; */ /* added scg 3/16/99 */

		nclonefiles = 0;
		strcpy( saveas_name, "" );
		strcpy( last_proctok, buf2 );

		if( sscanf( buf, "%*s %s", procname ) < 1 ) {
			return( Eerr( 24, "#proc must be followed by a procedure name", "" ) );
			}

		if( procname[ strlen( procname ) - 1 ] == ':' ) 
			procname[ strlen( procname ) - 1 ] = '\0';

		first = 0;
		fp = fopen( tmpfilename, "w" );
		if( fp == NULL ) {
			return( Eerr( 25, "Cannot open tmp file", tmpfilename ) );
			}

		}
	else 	{
		if( fp == NULL ) {
			if( nt < 1 ) continue;
			/* else Eerr( 26, "Warning, ignoring this control file line", buf ); */
			}
		else 	{
			/* copy line over to tmp file, checking for exceptions, such
				as "Clone" */

			if( !lastbs && strnicmp( buf2, "#clone", 6 )==0 ) {
				sscanf( buf, "%*s %s", clonefiles[nclonefiles++] );
				}

			else if( !lastbs && strnicmp( buf2, "#saveas", 7 )==0 ) {
				sscanf( buf, "%*s %s", saveas_name );
				}

			else if( !lastbs && strnicmp( buf2, "#intrailer", 10 )==0 ) {
				long loc;
				int seekstate;
				char tok[80];
				/* remember current location in control file, then
				   scan forward until we reach #data , copy dats to
				   tmp file, then seek back to where we left off. 
				   Does not work from within an #include.  */
				loc = ftell( ss.fp );
				seekstate = 0;
				while( fgets( buf, 255, ss.fp ) != NULL ) {
					sscanf( buf, "%s %s", buf2, tok );
					if( seekstate == 0 ) {
						if( stricmp( buf2, "#proc" ) ==0 &&
							stricmp( tok, "trailer" )==0 ) {
							/* fprintf( fp, "Data:\n" ); */
							seekstate = 1;
							}
						else continue;
						}
					else fprintf( fp, "%s", buf );
					}
				fseek( ss.fp, loc, 0 /*SEEK_SET BSD*/ );  /* now go back.. */
				}

			else 	{
				fprintf( fp, "%s", buf );
				}
			if( buf[ strlen( buf ) - 2 ] == '\\' ) lastbs = 1;
			else lastbs = 0;
			}
		}
	}

/* last case.. */
if( fp != NULL ) {
	/* no need to code for proc page here; it can never be the last proc */
	fclose( fp );
	if( !Initialized && !slmember( procname, NONDRAWINGPROCS ) ) {
		Einit( Inputfile, Device );
		Epaper( Landscape );
		Initialized = 1;
		}
	proc_call( procname, tmpfilename, clonefiles, nclonefiles );
	if( Initialized ) Eflush(); /* get output onto screen.. */
	nclonefiles = 0;
	}

return( 0 );
}

/* ========================= */
proc_call( procname, tmpfilename, clonefiles, nclonefiles )
/* char *procname, *tmpfilename, clonefiles[MAXCLONEFILES][]; BSD */
char *procname, *tmpfilename, *clonefiles[MAXCLONEFILES]; 
int nclonefiles;
{
int stat;
int n;

if( Debug ) { fprintf( Diagfp, "Executing %s\n", procname ); fflush( Diagfp ); }

if( stricmp( procname, "areadef" )==0 ) {
	stat = areadef( tmpfilename, clonefiles, nclonefiles );
	if( stat != 0 ) {
		Eerr( 10, "(Fatal) areadef failed due to problems with area specification or xrange/yrange.", "" );
		exit(1);
		}
	}
else if( stricmp( procname, "page" )==0 ) page( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "xaxis" )==0 ) 
	axis( tmpfilename, clonefiles, nclonefiles, 'x', 0 );
else if( stricmp( procname, "yaxis" )==0 ) 
	axis( tmpfilename, clonefiles, nclonefiles, 'y', 0 );
else if( stricmp( procname, "getdata" )==0 ) getdata( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "lineplot" )==0 ) lineplot( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "bars" )==0 ) bars( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "rangebar" )==0 ) rangebar( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "scatterplot" )==0 ) scatterplot( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "annotate" )==0 ) annotate( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "legend" )==0 ) legend( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "tabulate" )==0 ) tabulate( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "curvefit" )==0 ) curvefit( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "rangesweep" )==0 ) rangesweep( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "pie" )==0 ) pie( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "drawcommands" )==0 ) drawcommands( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "line" )==0 ) line( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "symbol" )==0 ) onesymbol( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "import" )==0 ) import( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "print" )==0 ) print( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "transform" )==0 ||
         stricmp( procname, "processdata" )==0 ) processdata( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "originaldata" )==0 ) {
	Dsel = 0;
	setintvar( "NRECORDS", Nrecords[Dsel] );
	setintvar( "NFIELDS", Nfields[Dsel] );
	}
else if( stricmp( procname, "legendentry" )==0 ) legendentry( tmpfilename, clonefiles, nclonefiles);
else if( stricmp( procname, "breakaxis" )==0 ) breakaxis( tmpfilename, clonefiles, nclonefiles);
else if( stricmp( procname, "trailer" )==0 ) ; /* do nothing */
else if( stricmp( procname, "endproc" )==0 ) ; /* do nothing */
else if( stricmp( procname, "bevelrect" )==0 ) bevelrect( tmpfilename, clonefiles, nclonefiles );
else if( stricmp( procname, "defineunits" )==0 ) defineunits( tmpfilename, clonefiles, nclonefiles);
else if( stricmp( procname, "catslide" )==0 ) catslide( tmpfilename, clonefiles, nclonefiles);
else if( stricmp( procname, "datesettings" )==0 ) datesettings( tmpfilename, clonefiles, nclonefiles);
else return( Eerr( 101, "procedure name unrecognized", procname ) );

if( Initialized ) Eflush();
n = report_convmsgcount();
if( Debug && n > 0 ) {
	fprintf( Diagfp, "note: pl proc %s encountered %d unplottable data values\n", procname, n );
	zero_convmsgcount();
	}
return( 0 );
}
	
/* ========================= */
Ehandle_events( x, y, e )
double x, y;
int e;
{
if( e == E_EXPOSE || e == E_RESIZE ) Erestorewin();
return( 0 );
}
