/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include <stdio.h>
#ifdef PLOTICUS
#include "dms.d"
#endif

/* =============================================== */
/* FUNCTION_CALL - parse function name, and any arguments, then call
   function evaluator.  V is returned with the result value. */

function_call( v )
char *v;
{
int i, j;
char name[30];
char parameters[ 1024 ];
int status;

/* v should look like this:  $name([arg1,arg2..argN) */

i = 0;
getchunk( name, v, &i, "(" );
if( strlen( name ) < 1 ) return( -1 );
strcpy( parameters, &v[i+1] );

/* if( parameters[ strlen( parameters ) - 1 ] == ')' )
 *	parameters[ strlen( parameters ) - 1 ] = '\0';
 */
 
/* search backwards from end of paramters to find closing ')' */
for( j = strlen( parameters ) - 1; j >= 0; j-- ) {
	if( parameters[j] == ')' ) {
		parameters[j] = '\0';
		break;
		}
	}

status = eval_function( name, parameters, v );
if( status != 0 ) return( -1 );
return( 0 );
}


/* ======================================== */
/* EVAL_FUNCTION - implement general purpose condex functions. 

	Return 0 if ok, -1 if named function doesn't exist or some other error. 
	Functions which produce a result which should always be taken as
   	alpha rather than number should append a space to the end of the result. 

	Note: functions that update multiple data items are found in scriptops.c
*/

eval_function( name, parms, value )
char *name; 	/* function name */
char *parms;	/* function parameter list (generally comma delimited) */
char *value;	/* the result value gets copied into this string */
{
int i, stat, typ;
char arg[256];
int result;
double accum;
char curop;
double atof();
char *getok();
char tok[256];
char tok2[256];



/* ------------ */
/* Arithmetic   */
/* ------------ */
/* $arith() - L to R arithmetic expression evaluator, e.g.  4+3*2-1 (unary plus/minus allowed) */
/* $arithl() - same as above except lazy (non-numeric arguments are silently allowed and treated as 0) */
if( smember( name, "$arith $arithl" )) {
	i = 0;
	accum = 0.0;
	curop = '+';
	parms[ strlen( parms ) +1 ] = '\0'; /* append extra null so loop terminates properly - scg 11/13/00 */
	if( parms[0] == '-' ) arg[0] = '-'; /* check for unary minus on first arg */
	else arg[0] = '+';
	while( 1 ) {

		getchunk( &arg[1], parms, &i, "+-/*%" ); /* arg[1] because sign is always in arg[0] */
		if( strlen( &arg[1] ) < 1 ) break;
			
		if( ! goodnum( arg, &stat ) && name[6] != 'l' )  
			{ err( 15003, 0, "arith: bad value", arg ); return( -1 ); }
		if( curop == '+' ) accum = accum + atof( arg );
		else if( curop == '-' ) accum = accum - atof( arg );
		else if( curop == '/' ) accum = accum / atof( arg );
		else if( curop == '*' ) accum = accum * atof( arg );
		else if( curop == '%' ) accum = (int)(accum) % (int)(atof( arg ));
		else { err( 15004, "arith:bad operator", "" ); return( -1 ); }
		curop = parms[i];
		if( parms[i+1] == '-' ) arg[0] = '-'; /* check for unary minus on next arg */
		else arg[0] = '+';
		i++;
		}
	sprintf( value, "%g", accum );
	return( 0 );
	}

/* ----------- */
/* $inrange() - (custom ploticus function)  */
else if( strcmp( name, "$inrange" )==0 ) {
	stat = custom_function( name, parms, value );
	if( stat != 0 ) {
		err( 15005, 0, "funtion not defined", name );
		return( -1 );
		}
	return( 0 );
	}

/* ------------ */
/* $data_to_absolute(axis,val) - (custom ploticus function) given a data coordinate, return the absolute location.
   Example: #set AVAL = $data_to_absolute(Y,@DVAL)
 */
else if( strcmp( name, "$data_to_absolute" )==0 ) {
	double f, Econv(), Ea();
	i = 0;
	getchunk( arg, parms, &i, "," ); /* axis */
	getchunk( tok, parms, &i, "," ); /* value */
	f = Econv( tolower(arg[0]), tok );
	sprintf( value, "%g", Ea( tolower(arg[0]), f ) );
	return( 0 );
	}


/* --------------- */
/* $daysdiff(d1,d2) - return # of days between two dates.  Old name: $datecmp() */
else if( strcmp( name, "$daysdiff" )==0 || strcmp( name, "$datecmp" )==0 ) {
	long diff;
	i = 0;
	getchunk( arg, parms, &i, "," ); /* d1 */
	getchunk( tok, parms, &i, "," ); /* d2 */
	stat = daysdiff( arg, tok, &diff );
	if( stat != 0 ) {
		err( 15006, 0, "$daysdiff error", parms );
		return( -1 );
		}
	sprintf( value, "%ld", diff );
	return( 0 );
	}
/* --------------- */
/* $dateinrange(testdate,earlydate,latedate) - return 1 if testdate is w/in interval
	exclusive, 0 if not */
else if( strcmp( name, "$dateinrange" )==0 ) {
	long diff;
	char latedate[30];
	i = 0;
	getchunk( arg, parms, &i, "," ); /* test */
	getchunk( tok, parms, &i, "," ); /* early */
	getchunk( latedate, parms, &i, "," ); /* early */
	stat = dateinrange( arg, tok, latedate );
	if( stat == -1 ) {
		err( 15007, 0, "$dateinrange error", parms );
		return( -1 );
		}
	sprintf( value, "%d", stat );
	return( 0 );
	}

/* ----------- */
/* $isnumber(s) - return 1 if argument is a valid number, 0 if not. */
else if( strcmp( name, "$isnumber" )==0 ) {
	int i;
	if( strlen( parms ) < 1 ) {
		sprintf( value, "0" ) ;
		return( 0 );
		}
	stat = goodnum( parms, &i );
	sprintf( value, "%d", stat );
	return( 0 );
	}

/* ----------- */
/* $formatfloat(n,format) */
else if( strcmp( name, "$formatfloat" )==0 ) {
	double atof();
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	strcat( tok, " " ); /* force result to be seen as a string */
	sprintf( value, tok, atof( arg ) );
	return( 0 );
	}

/* ----------- */
/* $inr(n,low,hi) - see if n is in range low to high */
else if( strcmp( name, "$inr" )==0 ) {
	double atof();
	char hi[30];
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	getchunk( hi, parms, &i, "," );
	if( !goodnum( arg, &i ) ) strcpy( value, "0" );
	else if( atof( arg ) >= atof( tok ) && atof( arg ) <= atof( hi ) ) strcpy( value, "1" );
	else strcpy( value, "0" );
	return( 0 );
	}

/* ------------- */
/* $max(a,b,c) - return the greatest of a, b, c, etc. (any number of args may be used) */
else if( strcmp( name, "$max" )==0 ) {
	double max;
	int prec;
	max = -9999999.0;
	i = 0;
	while( 1 ) {
		getchunk( arg, parms, &i, "," ); /* $max */
		if( strlen( arg ) < 1 ) break;
		if( !goodnum( arg, &prec ) ) continue; /* added scg 11/15/00 */
		if( atof( arg ) > max ) max = atof( arg );
		}
	sprintf( value, "%g", max );
	return( 0 );
	}
	

#ifndef PLOTICUS
/* ---------- */
/* $checksumgood(s) -  validate s which is an integer containing a trailing check digit, 
   using the algorithm in $ps/clibx/checkdig  */
/* ---------- */
else if( strcmp( name, "$checksumgood" )==0 ) {
	int i;
	i = checkdig( parms, strlen( parms ) -1 );
	if( i == parms[ strlen( parms ) -1 ] ) sprintf( value, "1" );
	else sprintf( value, "0" );
	return( 0 );
	}

/* ----------- */
/* $checksumencode(i) - return i with checksum digit appended. */
else if( strcmp( name, "$checksumencode" )==0 ) {
	long i, atol();
	char tmp[20];
	int cs;
	i = atol( parms );
	sprintf( tmp, "%ld", i );
	cs = checkdig( tmp, strlen( tmp ) );
	sprintf( value, "%s%c ", tmp, cs );
	return( 0 );
	}
	

/* -------------------------- */
/* $checksumnext(s) take s which is a number including trailing checksum 
   digit, increment number and recompute new checksum digit. */
else if( strcmp( name, "$checksumnext" )==0 ) {
	int cs;
	long i, atol();
	char tmp[20];
	parms[ strlen( parms ) - 1 ] = '\0'; /* strip off existing checksum digit */
	i = atol( parms );
	i++;
	sprintf( tmp, "%ld", i );
	cs = checkdig( tmp, strlen( tmp ) );
	sprintf( value, "%s%c ", tmp, cs );
	return( 0 );
	}
#endif /* PLOTICUS */

	


/* ------------ */
/* NEW date functions */
/* ------------ */
/* $setdatefmt(fmt) - set the date format.  fmt may be any of the supported format strings. */
else if( strcmp( name, "$setdatefmt" )==0 ) {
	stat = setdatefmt( parms );
	if( stat != 0 ) {
		err( 15008,0, "$setdatefmt given an invalid date format", parms );
		return( -1 );
		}
	sprintf( value, "0" );
	return( 0 );
	}

/* -------------- */
/* $setdateparms(parm,value) - set a date parameter.
   See dates.c setdateparms() */
else if( strcmp( name, "$setdateparms" )==0 ) {
	i = 0;
	getchunk( arg, parms, &i, "," ); /* parm */
	getchunk( tok, parms, &i, "," ); /* value */
	stat = setdateparms( arg, tok );
	if( stat != 0 ) {
		err( 15009, 0, "$setdateparms error ", parms );
		return( -1 );
		}
	sprintf( value, "0" );
	return( 0 );
	}

/* -------------- */
/* $defaultdatefmt() - set the date format to the default */
else if( strcmp( name, "$defaultdatefmt" )==0 ) {
	defaultdatefmt();
	return( 0 );
	}
/* --------------- */
/* $todaysdate() - return today's date.  */
else if( strcmp( name, "$todaysdate" )==0 || strcmp( name, "$today" )==0 ) {
	stat = todaysdate( arg );
	sprintf( value, "%s ", arg );
	return( 0 );
	}
/* --------------- */
/* $formatdate(date,fmt) - format the date, using fmt which is any of 
	the supported format strings) */

else if( strcmp( name, "$formatdate" )==0 ) {
	i = 0;
	getchunk( arg, parms, &i, "," ); /* date */
	/* getchunk( tok, parms, &i, "," ); */ /* fmt */
	strcpy( tok, &parms[i + 1] );
	stat = formatdate( arg, tok, tok2 );
	if( stat != 0 ) {
		err( 15010, 0, "$formatdate failed", parms );
		return( -1 );
		}
#ifdef PLOTICUS
	/* convert underscore to space - ok for ploticus scg 6/23/99 */
	for( i =0; i< strlen( tok2 ); i++ ) if( tok2[i] == '_' ) tok2[i] = ' ';
#endif
	sprintf( value, "%s ", tok2 );
	return( 0 );
	}
/* --------------- */
/* $daysdiff(d1,d2)  -- moved above for performance */
/* --------------- */
/* $dateinrange(testdate,earlydate,latedate) -- moved above for performance */
/* --------------- */
else if( strcmp( name, "$daysdiff_m" )==0 ) {
	long diff;
	char fmt1[20], fmt2[20];
	i = 0;
	getchunk( arg, parms, &i, "," ); /* d1 */
	getchunk( fmt1, parms, &i, "," ); /* fmt1 */
	getchunk( tok, parms, &i, "," ); /* d2 */
	getchunk( fmt2, parms, &i, "," ); /* fmt1 */
	stat = daysdiff_m( arg, fmt1, tok, fmt2, &diff );
	if( stat != 0 ) {
		err( 15011, 0, "$daysdiff_m error", parms );
		return( -1 );
		}
	sprintf( value, "%ld", diff );
	return( 0 );
	}
/* --------------- */
/* $dateadd(date,ndays) - return new date given date + ndays */
else if( strcmp( name, "$dateadd" )==0 ) {
	int ndays, j;
	char result[80];
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	ndays = atoi( tok );
	stat = dateadd( arg, ndays, result );
	if( stat != 0 ) {
		err( 15012, 0, "$dateadd error", parms );
		return( -1 );
		}
	sprintf( value, "%s ", result );
	return( 0 );
	}
/* ---------------- */
/* $weekday(date) - return day of week (abbreviated) for the given date */
else if( strcmp( name, "$weekday" )==0 ) {
	stat = weekday( parms, tok );
	if( stat != 0 ) {
		err( 15013, 0, "$weekday error", parms );
		return( -1 );
		}
	sprintf( value, "%s", tok );
	return( 0 );
	}
/* ----------------- */
/* $yearsold(birthdate,testdate) - return # years age (whole number) 
    between birthdate and testdate */
else if( strcmp( name, "$yearsold" )==0 ) {
	int nyears;
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	stat = yearsold( arg, tok, &nyears );
	if( stat != 0 ) {
		err( 15014, 0, "$yearsold error", parms );
		return( -1 );
		}
	sprintf( value, "%d ", nyears );
	return( 0 );
	}



/* ------------ */
/* Time */
/* ------------ */
else if( strcmp( name, "$settimefmt" )==0 ) {
	stat = settimefmt( parms );
	if( stat != 0 ) {
		err( 15015, 0, "$settimefmt - invalid format", parms );
		return( -1 );
		}
	strcpy( value, "0" );
	return( 0 );
	}

/* other time functions from time.c not installed here as yet.. */

/* $time() - return current time in hh:mm:ss format */
else if( strcmp( name, "$time" )==0 ) {
	int hr, min, sec;
	systime( &hr, &min, &sec );
	sprintf( value, "%02d:%02d:%02d", hr, min, sec );
	return( 0 );
	}

#ifndef PLOTICUS
/* ------------ */
/* $timesec() - return number of seconds since midnight */
else if( strcmp( name, "$timesec" )==0 ) {
	int hr, min, sec;
	systime( &hr, &min, &sec );
	sprintf( value, "%ld", (((hr*(long)60) + min)*(long)60) + sec );
	return( 0 );
	}

/* ------------ */
/* $minutesdiff() - return number of minutes between two hh:mm times from the same day. */
else if( strcmp( name, "$minutesdiff" )==0 ) {
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	sprintf( value, "%d", timecmp( arg, tok ) );
	return( 0 );
	}
#endif /* PLOTICUS */

/* ------------ */
/* strings      */
/* ------------ */

/* ------------ */
/* $substring(s,from,nchar) - return a substring, e.g. $substring(abcde,3,99) --> cde */
else if( strcmp( name, "$substring" )==0 ) {
	int fromchar, nchar;
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	fromchar = atoi( tok );
	getchunk( tok, parms, &i, "," );
	nchar = atoi( tok );
	substring( value, arg, fromchar, nchar );
	strcat( value, " " ); /* force alpha */
	return( 0 );
	}

/* ------------ */
/* $strlen(s) - return length of string s */
else if( strcmp( name, "$strlen" )==0 ) {
	sprintf( value, "%d", strlen( parms ) );
	return( 0 );
	}

/* ------------ */
/* $strcat(s,t) - concatenate s and t */
else if( strcmp( name, "$strcat" )==0 ) {
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	sprintf( value, "%s%s", arg, tok );
	strcat( value, " " ); /* force alpha */
	return( 0 );
	}


/* ------------ */
/* $char(n,string) - return the nth character of string (first is 1) */
else if( strcmp( name, "$char" )==0 ) {
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	sprintf( value, "%c", tok[atoi(arg)-1] );
	return( 0 );
	}

/* ------------ */
/* $lowerc(string) - return the lower-case equivalent of string */
else if( strcmp( name, "$lowerc" )==0 ) {
	strcpy( value, parms );
	for( i = 0; i < strlen( value ); i++ ) value[i] = tolower( value[i] );
	strcat( value, " " ); /* force alpha */
	return( 0 );
	}

/* ------------ */
/* $upperc(string) - return the lower-case equivalent of string */
else if( strcmp( name, "$upperc" )==0 ) {
	strcpy( value, parms );
	for( i = 0; i < strlen( value ); i++ ) value[i] = toupper( value[i] );
	strcat( value, " " ); /* force alpha */
	return( 0 );
	}

/* ------------ */
/* $padr(len,string) - right-pad the string to a length of len */
else if( strcmp( name, "$padr" )==0 ) {
	int len;
	i = 0;
	getchunk( arg, parms, &i, "," );
	i++;
	strcpy( value, &parms[i] );
	len = atoi( arg );
	if( len < 1 ) { err( 15016,0, "$padr: bad arg 1", "" ); return( -1 ); }
	while( strlen( value ) < len ) strcat( value, "_" );
	strcat( value, " " ); /* force alpha */
	return( 0 );
	}

/* ------------ */
/* $change(s,t,string) - change every occurance of character s to t in string */
/* the special word 'space' may be used for t  */
else if( strcmp( name, "$change" )==0 ) {
	char s[10], t[10];
	i = 0;
	getchunk( s, parms, &i, "," );
	getchunk( t, parms, &i, "," );
	if( strcasecmp( t, "space" )==0 ) strcpy( t, " " );
	/* getchunk( arg, parms, &i, "," ); */
	strcpy( arg, &parms[i+1] );
	for( i = 0; i < strlen( arg ); i++ ) if( arg[i] == s[0] ) arg[i] = t[0];
	sprintf( value, "%s ", arg );
	return( 0 );
	}
	
/* ------------- */
/* $string(s) - force s to be treated as a string */
else if( strcmp( name, "$string" )==0 ) {
	sprintf( value, "%s ", parms );
	return( 0 );
	}

else if( strcmp( name, "$fuzzymatch" )==0 ) {
	char s[100], t[100];
	i = 0;
	getchunk( s, parms, &i, "," );
	getchunk( t, parms, &i, "," );
	getchunk( arg, parms, &i, "," );
	stat = fuzzymatch( s, t, strlen(t), atoi(arg) );
	sprintf( value, "%d", stat );
	return( 0 );
	}

/* ------------ */
/* List processing */
/* ------------ */

/* $count(str,list) - count the number of times 'str' appears in 'list'.  List is
	a comma-delimited list.  Example: $count(hello,aba,gabba,jabba) would yield 0 */
else if( strcmp( name, "$count" )==0 ) {
	i = 0;
	getchunk( arg, parms, &i, "," );
	i++;
	result = 0;
	while( i < strlen( parms ) ) {
		getchunk( tok, parms, &i, "," );
		if( strlen( tok ) < 1 ) break;
		if( strcmp( tok, arg )==0 ) result++;
		}

	sprintf( value, "%d", result );
	return( 0 );
	}

/* $nmember(n,list) - return the nth member of the list.  List is a comma-delimited list. */
else if( strcmp( name, "$nmember" )==0 ) {
	int j, n;
	i = 0;
	getchunk( arg, parms, &i, "," );
	n = atoi( arg );
	j = 0;
	while( 1 ) {
		j++;
		getchunk( tok, parms, &i, "," );
		if( strlen( tok ) < 1 ) { strcpy( value, " " ); break; }
		else if( j == n ) { sprintf( value, "%s ", tok ); break; }
		}
	return( 0 );
	}

#ifndef PLOTICUS
/* ------------ */
/* Database */
/* ------------ */
/* $dba_get(command)
	Retrieve one data item using a dba command (see dbaccess.c).
	Since spaces are not allowed in condex function paramters,
	the single quote (') should be used instead of space.  Also, since 
	condex() will try to convert @vars, a double @ should be used:
	Example:
	 $dba_get(datamode'from'clinics'where'@@number'='01)  

	If no matching records are found and everything else is ok, "" is returned.
	For other dba_get errors, the string "errorN" is returned.
 */
else if( strcmp( name, "$dba_get" )==0 ) {
	for( i = 0; i < strlen( parms ); i++ ) if( parms[i] == '\'' ) parms[i] = ' ';
	stat = dba_get( value, parms );
	if( stat == 1 ) strcpy( value, "" );
	else if( stat != 0 ) sprintf( value, "error%d", stat );
	strcat( value, " " ); /* force alpha */
	return( 0 );
	}

/* ------------ */
/* $dba_next(ch) - get next record if any when doing multi-item retrieval.
	Yields 0 if another record was read; 1 on EOF */

else if( strcmp( name, "$dba_next" )==0 ) {
	int ch;
	ch = atoi( parms );
	if( ch < 1 || ch > 3 ) ch = 1;
	stat = dba_next( ch );
	sprintf( value, "%d", stat );
	return( 0 );
	}

/* ------------- */
/* $dba_force_new_query(ch) - force this query to begin anew, even if it is
	the same as the previous query. */

else if( strcmp( name, "$dba_force_new_query" )==0 ) {
	int ch;
	ch = atoi( parms );
	if( ch < 1 || ch > 3 ) ch = 1;
	stat = dba_force_new_query( ch );
	sprintf( value, "%d", stat );
	return( 0 );
	}

/* ------------- */
/* $dbfileexists(file) - returns 1 if named file exists, 0 otherwise */
else if( strcmp( name, "$dbfileexists" )==0 ) {
	stat = dbfileexists( parms );
	sprintf( value, "%d", stat );
	return( 0 );
	}

/* ------------- */
/* $dba_update(varvaluepairs,dbacommand) - updates the specified fields on the data record */
else if( strcmp( name, "$dba_update" )==0 ) {
	int i;
	for( i = 0; i < strlen( parms ); i++ ) if( parms[i] == '\'' ) parms[i] = ' ';
	i = 0;
	getchunk( arg, parms, &i, "," );
	stat = dba_update( arg, &parms[i+1], "if_file_available" );
	if( stat != 0 ) sprintf( value, "error%d", stat );
	else sprintf( value, "0" );
	return( 0 );
	}

/* ------------- */
/* $dba_insert(varvaluepairs,dbacommand) - insert a new data record and set the specified fields
	to the given values.  The dbacommand is used to select the datafile, and also to ensure
	that the record to be inserted does not already exist. */
else if( strcmp( name, "$dba_insert" )==0 ) {
	int i;
	for( i = 0; i < strlen( parms ); i++ ) if( parms[i] == '\'' ) parms[i] = ' ';
	i = 0;
	getchunk( arg, parms, &i, "," );
	stat = dba_insert( arg, &parms[i+1], "if_file_available" );
	if( stat != 0 ) sprintf( value, "error%d", stat ); /* changed from "if(stat != 1)" 3/30/99 */
	else sprintf( value, "1" );
	return( 0 );
	}
/* ------------- */
/* $dba_delete(dbacommand) - delete the record specified by the dbacommand */
else if( strcmp( name, "$dba_delete" )==0 ) {
	int i;
	for( i = 0; i < strlen( parms ); i++ ) if( parms[i] == '\'' ) parms[i] = ' ';
	stat = dba_delete( parms, "if_file_available" );
	if( stat != 0 ) sprintf( value, "error%d", stat );
	else sprintf( value, "0" );
	return( 0 );
	}

/* ------------- */
/* $dba_serialnum(dbacommand) - retrieves a number using the dbacommand, increments it,
	and stores it back in the named table. Result is the serial number, after
	incrementing. */

else if( strcmp( name, "$dba_serialnum" )==0 ) {
	int i;
	for( i = 0; i < strlen( parms ); i++ ) if( parms[i] == '\'' ) parms[i] = ' ';
	i = 0;
	stat = dba_serialnum( parms, arg );
	if( stat != 0 ) sprintf( value, "error%d", stat );
	else sprintf( value, "%s", arg );
	return( 0 );
	}

/* ------------- */
/* $dna_noitem_ok(mode) - mode=1: allow requests for non-existant items; they
	will result in NO-ITEM.  mode=0: don't. */

else if( strcmp( name, "$dba_noitem_ok" )==0 ) {
	dba_noitem_ok( atoi( parms ) );
	sprintf( value, "0" );
	return( 0 );
	}

/* -------------- */
/* script execution */
/* -------------- */
/* $scripter_call(scriptfile) - execute named scripter script.  Result is the
	first line of results.  If the script produces no results, the script's
	exit code is the result. */
else if( strcmp( name, "$scripter_call" )==0 ) {
	/* stat = */
	scripter_call( value, parms );
	/* if( stat != 0 ) sprintf( value, "%d ", stat );
	 * else 	{ */
	for( i = 0; i < strlen( value ); i++ ) if( value[i] == ' ' ) value[i] = '_';
	strcat( value, " " ); /* force alpha */
	/* } */
	return( 0 );
	}
#endif /* PLOTICUS */

/* --------------- */
/* validation */
/* --------------- */
/* $validate(val,datatype) - return 1 if val is a valid instance of datatype, 0 if not. */
else if( strcmp( name, "$validate" )==0 ) {
	i = 0;
	getchunk( arg, parms, &i, "," );
	getchunk( tok, parms, &i, "," );
	for( i = 0; i < strlen( tok ); i++ ) if( tok[i] == '\'' ) tok[i] = ' ';
	sprintf( value, "%d",  validate( arg, tok ) );
	return( 0 );
	}

/* -------------------------- */
/* external command execution */
/* -------------------------- */

else if( strcmp( name, "$system" )==0 ) {
	FILE *pfp, *popen();
	for( i = 0; i < strlen( parms ); i++ ) if( parms[i] == '\'' ) parms[i] = ' ';
	pfp = popen( parms, "r" );
	if( pfp == NULL ) {
		sprintf( value, "error" );
		return( 0 );
		}
	if( fgets( arg, 80, pfp ) == NULL ) {
		/* err( 15018, 0, "$system: no results.\n", parms ); */ 
			/* ? scg 4/2/99 - case in sst $h/scripter/surglogs where
				this is acceptable and should not give a system error */
		sprintf( value, "" );
		pclose( pfp );
		return( 0 );
		}
	else 	{
		if( arg[ strlen( arg ) - 1 ] == '\n' ) arg[ strlen( arg ) - 1 ] = '\0';
		/* convert spaces to underscores.. */
		for( i = 0; i < strlen( arg ); i++ ) if( arg[i] == ' ' ) arg[i] = '_';
		strcpy( value, arg );
		strcat( value, " " );
		pclose( pfp );
		return( 0 );
		}
	}


/* ---------
   MISC
   --------- */
/* $messageline(msg) */
#ifndef PLOTICUS
else if( strcmp( name, "$messageline" )==0 ) {
	int i;
	for( i = 0; i < strlen( parms ); i++ ) if( parms[i] == '\'' ) parms[i] = ' ';
	message( parms );
	strcpy( value, "0" );
	return( 0 );
	}

/* $userinput(msg,len,datatype) */
else if( strcmp( name, "$userinput" )==0 ) {
	int i;
	char uin[80];
	for( i = 0; i < strlen( parms ); i++ ) if( parms[i] == '\'' ) parms[i] = ' ';
	i = 0;
	getchunk( arg, parms, &i, "|" );
	message( arg );
	getchunk( tok, parms, &i, "|" );
	inp( uin, atoi(tok), &parms[i+1] );
	sprintf( value, " %s", uin );
	return( 0 );
	}

/* ------------ */
else if( strcmp( name, "$self" )==0 ) {
	strcpy( value, parms );
	return( 0 );
	}
#endif /* PLOTICUS */

/* ------------ */
/* try a custom function.. */
else	{
	stat = custom_function( name, parms, value );
	if( stat != 0 ) {
		err( 15020, 0, "function not defined", name );
		return( -1 );
		}
	else return( 0 );
	}
}

