/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* Render an automatic legend. 
   Plotting procs supply legend entries */

#include "pl.x"

#define MAXLEGENT 30            /* max # of legend entries */
#define MAXLEGTEXT 2000         /* max amount of legend text, including labels
                                        and details attributes */
 
static int NLE = 0;
static int LEavail = 0;
static int LEtype[MAXLEGENT];
static int LEparm1[MAXLEGENT];
static int LEparm2[MAXLEGENT];
static int LEparm3[MAXLEGENT];
static int LElabel[MAXLEGENT];
static char Ltext[MAXLEGTEXT];


legend( specfile, clonefiles, nclonefiles  )
char *specfile, clonefiles[][ MAXPATH ];
int nclonefiles;
{
int i;
FILE *fp;
char buf[256];
char attr[40], val[256];
char lineval[256];
int nt;
FILE *openclonefile();
int icf;

int stat;
int align;
double adjx, adjy;

double x, y;
char format;
int outline;
int nlines, maxlen;
double yy;
char textdetails[256];
int reverseorder;
int j;
double seglen;
double hsep, msep;
int do_outline;
int colortext;
int specifyorder;
int ix, k;
double swatchsize;
int noclear;


strcpy( Eprogname, "pl proc legend" );
/* -------------------------- */
/* DEFAULTS */
/* All variables must start out with default values, */
/* since no variable is guaranteed to be set */
/* -------------------------- */
x = -9999.0;
y = -9999.0;
format = 'm';
strcpy( textdetails, "" );
reverseorder = 0;
seglen = 0.5;
msep = 0.0;
hsep = 1.2;
do_outline = 0;
colortext = 0;
strcpy( Bigbuf, "" );
specifyorder = 0;
ix = 0;
swatchsize = 0.1;
noclear = 0;


/* -------------------------- */
/* open and process all clonefiles and lastly, specfile */
/* -------------------------- */
for( icf = 0; icf <= nclonefiles; icf++ ) {
 
    if( icf == nclonefiles ) {
        fp = fopen( specfile, "r" );
        if( fp == NULL ) return( Eerr( 117, "Could not open to read", specfile ) );
        }
    else        {
        fp = openclonefile( clonefiles[icf] );
        if( fp == NULL ) continue;
        }

    while( fgets( buf, 255, fp ) != NULL ) {
	nt = sscanf( buf, "%s %s", attr, val );		/* val has 2nd token */
	if( nt < 1 ) continue;
	if( nt == 1 ) strcpy( val, "" );
	if( icf < nclonefiles && stricmp( attr, "#proc" )==0 ) continue;
	i = 0;
        Egetok( buf, &i );
	while( buf[i] == ' ' || buf[i] == '\t' ) i++; /* skip over white space */
        strcpy( lineval, &buf[ i ] ); /* lineval has remainder of line, for sscanf-ing */
        lineval[ strlen( lineval ) - 1 ] = '\0'; /* strip off newline */
 
	/* strip off trailing colon if any.. */
	if( attr[ strlen( attr ) - 1 ] == ':' ) attr[ strlen( attr ) - 1 ] = '\0';


	/* attributes.. */

	if( stricmp( attr, "location" )==0 ) {
		/* note: location x = beginning of LABEL; y = top */
		if( strlen( lineval ) > 0 ) getcoords( "location", lineval, &x, &y );
		}
	else if( stricmp( attr, "textdetails" )==0 ) strcpy( textdetails, lineval );
	else if( stricmp( attr, "seglen" )==0 ) seglen = atof( val );
	else if( stricmp( attr, "sep" )==0 ) {
		hsep = atof( val );
		msep = atof( val );
		}
	else if( stricmp( attr, "colortext" )==0 ) {
		if( strnicmp( val, YESANS, 1 )==0 ) colortext = 1;
		else colortext = 0;
		}
		
	else if( stricmp( attr, "outlinecolors" )==0 ) {
		if( strnicmp( val, YESANS, 1 )==0 ) do_outline = 1;
		else do_outline = 0;
		}
	else if( stricmp( attr, "noclear" )==0 ) {
		if( strnicmp( val, YESANS, 1 )==0 ) noclear = 1;
		else noclear = 0;
		}
	else if( stricmp( attr, "format" )==0 ) {
		if( strlen( val ) > 0 ) format = tolower(val[0]);
		else format = 'm';
		}
	else if( stricmp( attr, "specifyorder" )==0 ) {
		getmultiline( "specifyorder", lineval, fp, MAXBIGBUF, Bigbuf );
		specifyorder = 1;
		}

	else if( stricmp( attr, "swatchsize" )==0 ) {
		if( strlen( val ) > 0 ) {
			swatchsize = atof( val );
			if( Using_cm ) swatchsize /= 2.54;
			}
		else swatchsize = 0.1;
		}

	else if( stricmp( attr, "reverseorder" )==0 ) {
		if( strnicmp( val, YESANS, 1 )==0 ) reverseorder = 1;
		else reverseorder = 0;
		}
	else if( stricmp( attr, "reset" )==0 ) {
		NLE = 0;
		return( 0 );
		}

	else Eerr( 1, "attribute not recognized", attr );
	}
    fclose( fp );
    }

/* -------------------------- */
/* overrides and degenerate cases */
/* -------------------------- */

if( NLE < 1 ) return( Eerr( 5839, "Cannot do legend, no legend entries given yet", "" ) );


/* -------------------------- */
/* now do the plotting work.. */
/* -------------------------- */

/* default location */
if( x < -9000.0 ) { 
	if( scalebeenset() ) { x = EXhi - 1.5; y = EYhi - 0.1; }
	else { x = 6.0; y = 2.0; }
	}

ix = 0;

textdet( "textdetails", textdetails, &align, &adjx, &adjy, -2, "R" );
y -= Ecurtextheight; 
for( i = 0; i < NLE; i++ ) {
	/* fprintf( stderr, "%d|%s|%s|%s\n", LEtype[i],
		&Ltext[LElabel[i]],
		&Ltext[LEparm1[i]],
		&Ltext[LEparm2[i]] ); fflush ( stderr );
 	 */
	if( specifyorder ) {
		/* get next line in orderspec.. */
		Egetchunk( buf, Bigbuf, &ix, "\n" );
		if( ix >= strlen( Bigbuf ) ) break;

		/* now search for matching entry.. */
		for( k = 0; k < NLE; k++ ) {
			if( strnicmp( buf, &Ltext[LElabel[k]], strlen(buf) )==0) { 
				j = k;
				break;
				}
			}
		if( k == NLE ) {
			Eerr( 2894, "No legend entry matched", buf );
			continue;
			}
		}
	else if( reverseorder ) j = ((NLE-1)-i);
	else j = i;
		
	yy = y+(Ecurtextheight*0.4);

	/* draw swatch(es), depending on type */
	if( LEtype[j] == LEGEND_COLOR ) {
		char color[40];
		sscanf( &Ltext[LEparm1[j]], "%s", color );
		if( slmember( color, "white gr?y(1* 1*" )) outline = 1;
		else outline = do_outline;
		if( outline ) Elinetype( 0, 0.5, 1.0 );
		Ecblock( x-(swatchsize+0.1), y, x-0.1, y+swatchsize, color, outline );
		}
	else if( LEtype[j] == LEGEND_LINE ) {
		linedet( &Ltext[LElabel[j]], &Ltext[LEparm1[j]], 1.0 );
		Emov( x-(seglen+0.1), yy );
		Elin( x-0.1, yy );
		}
	else if( LEtype[j] == LEGEND_SYMBOL ) {
		double radius;
		char symcode[40];
		symdet( "symbol", &Ltext[LEparm1[j]], symcode, &radius );
		Emark( x-0.17, yy, symcode, radius );
		}
	else if( LEtype[j] == LEGEND_TEXT ) {
		/* parm1 is text, parm2 is textdetails */
		textdet( &Ltext[LElabel[j]], &Ltext[LEparm2[j]], &align, &adjx, &adjy, -2, "R" );
		Emov( x-0.1, y );
		Erightjust( &Ltext[LEparm1[j]] ); 
		}
	else if( LEtype[j] == LEGEND_LINE + LEGEND_SYMBOL ) {
		/* parm1 is linedetails, parm2 is symboldetails */
		double radius;
		char symcode[40];
		linedet( &Ltext[LElabel[j]], &Ltext[LEparm1[j]], 1.0 );
		Emov( x-(seglen+0.1), yy );
		Elin( x-0.1, yy );
		symdet( "symbol", &Ltext[LEparm2[j]], symcode, &radius );
		Emark( x-(seglen), yy, symcode, radius );
		Emark( x-0.1, yy, symcode, radius );
		}
	else if( LEtype[j] == LEGEND_TEXT + LEGEND_SYMBOL ) {
		/* parm1 is text, parm2 is textdetails, parm3 is symboldetails */
		double radius;
		char symcode[40];
		symdet( "symbol", &Ltext[LEparm3[j]], symcode, &radius );
		Emark( x-0.17, yy, symcode, radius );
		textdet( &Ltext[LElabel[j]], &Ltext[LEparm2[j]], &align, &adjx, &adjy, -2, "R" );
		Emov( x-0.17, y+(Ecurtextheight*0.2) );
		Ecentext( &Ltext[LEparm1[j]] ); 
		}
	

	/* get #lines and maxlen of label */
	Emeasuretext( &Ltext[LElabel[j]], &nlines, &maxlen );

	/* render label */
	if( ! colortext ) textdet( "textdetails", textdetails, &align, &adjx, &adjy, -2, "R" );
	Emov( x, y );
	Etext( &Ltext[LElabel[j]] );

	/* position for next line.. */
	if( format == 'm' ) {
		if( msep > 0.0 ) y -= msep;
		else y -= ((Ecurtextheight * nlines) + 0.03 );
		}
	else if( format == 's' ) {
		if( hsep > 0.0 ) x += ((Ecurtextwidth * maxlen ) + hsep);
		else x += ((Ecurtextwidth * maxlen ) + 1.2);
		}
	}

/* now reset so a new legend can be accumulated.. */
if( !noclear ) NLE = 0;
return( 0 );
}


/* ===================================== */
/* ADD_LEGENT - used by procs to add a legend entry */

add_legent( typ, label, parm1, parm2, parm3 )
int typ; 
char *label, *parm1, *parm2, *parm3;
{

Econvertnl( label );


LEtype[ NLE ] = typ;
LElabel[ NLE ] = LEavail;
if( LEavail + strlen( label ) >= MAXLEGTEXT ) 
	return( Eerr( 8478, "Sorry, too much legend content", "" ) );
strcpy( &Ltext[ LEavail ], label );
LEavail += (strlen( label ) + 1);

if( LEavail + strlen( parm1 ) >= MAXLEGTEXT ) 
	return( Eerr( 8478, "Sorry, too much legend content", "" ) );
LEparm1[ NLE ] = LEavail;
strcpy( &Ltext[ LEavail ], parm1 );
LEavail += (strlen( parm1 ) + 1);

if( LEavail + strlen( parm2 ) >= MAXLEGTEXT ) 
	return( Eerr( 8478, "Sorry, too much legend content", "" ) );
LEparm2[ NLE ] = LEavail;
strcpy( &Ltext[ LEavail ], parm2 );
LEavail += (strlen( parm2 ) + 1);

if( LEavail + strlen( parm3 ) >= MAXLEGTEXT ) 
	return( Eerr( 8478, "Sorry, too much legend content", "" ) );
LEparm3[ NLE ] = LEavail;
strcpy( &Ltext[ LEavail ], parm3 );
LEavail += (strlen( parm3 ) + 1);


NLE++;
return( 0 );
}

