/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include "pl.x"

static int Supress_convmsg = 0;
static int N_convmsg = 0;

/* ====================== */
/* DA - access D as a 2-D array; return as char *    */
char *
da( r, c )
int r, c;
{
int base;
base = StartD[Dsel];
if( r >= Nrecords[Dsel] ) { 
	fprintf( Errfp, "warning: %s data reference (rec=%d field=%d) is out of bounds\n", 
		Eprogname, r+1, c+1 ); 
	return( "" );
	}
if( c >= Nfields[Dsel] ) { 
	fprintf( Errfp, "warning: %s data reference (rec=%d field=%d) is out of bounds\n", 
		Eprogname, r+1, c+1 ); 
	return( "" );
	}
return( D[ base + ( r * Nfields[Dsel] ) + c ] );
}
/* ====================== */
/* FDA - access D as a 2-D array; return as double; 
   For non-plottables 0.0 is returned, but Econv_error may be called to see if 
   there was a conversion error */
double fda( r, c, ax )
int r, c;
char ax;
{
return( Econv( ax, da( r, c ) ) );
}


/* ====================== */
/* OPENCLONEFILE - open a clone file.  First try as a tmp file name, then
   as a normal path name.. */
FILE *
openclonefile( clonefile )
char *clonefile;
{
static FILE *fp;
char buf[200];
if( strlen( clonefile ) < 1 ) return( NULL );
sprintf( buf, "%s.%s", Tmpname, clonefile ); 
fp = fopen( buf, "r" );
if( fp == NULL ) fp = fopen( clonefile, "r" );
if( fp == NULL ) return( NULL );
else return( fp );
}

/* ====================== */
/* SAVEAS - copy spec file to named tmp file.. */
saveas( specfile, newname )
char *specfile, *newname;
{
char newfile[255];
sprintf( newfile, "%s.%s", Tmpname, newname ); 
Efilecopy( specfile, newfile, "w" );
return( 0 );
}

/* ====================== */
/* NUM - convert string to double, and ensure that it is proper numeric
   Result is s converted to double.
   Return is 0 on ok, 1 on non-numeric. */
num( s, result )
char s[];
double *result;
{
int i;
int nt;

nt = sscanf( s, "%lf", result );
if( nt > 0 ) return( 0 );
else	{
	*result = 0.0;
	return( 1 );
	}
/* ..replaced by above.. scg 5/23/00 
 * if( Egoodnum( s, &i ) ) {
 *	*result = atof( s );
 *	return( 0 );
 *	}
 * else 	{
 *	*result = 0.0;
 *	return( 1 );
 *	}
 */
}

/* ============================= */
/* GETCOORDS - extract a coordinate pair (two posex's) from val */
getcoords( parmname, val, x, y )
char *parmname, *val;
double *x, *y;
{
char px[40], py[40], dummy[40];
int nt;
int stat;

nt = sscanf( val, "%s %s", px, py, dummy );
if( nt != 2 ) return( Eerr( 55, "Two values expected", parmname ) );

stat = Eposex( px, X, x );
stat += Eposex( py, Y, y );
if( stat != 0 ) return( Eerr( 201, "Error on coord pair", parmname ) );
else return( 0 );
}

/* ============================= */
/* GETBOX - extract two coordinate pairs (four posex's) from val */
getbox( parmname, val, x1, y1, x2, y2 )
char *parmname, *val;
double *x1, *y1, *x2, *y2;
{
char px1[40], py1[40], px2[40], py2[40], dummy[40];
int nt;
int stat;

nt = sscanf( val, "%s %s %s %s", px1, py1, px2, py2, dummy );
if( nt != 4 ) return( Eerr( 55, "Two values expected", parmname ) );
stat = Eposex( px1, X, x1 );
stat += Eposex( py1, Y, y1 );
stat += Eposex( px2, X, x2 );
stat += Eposex( py2, Y, y2 );
if( stat != 0 ) return( Eerr( 201, "Error on box specification", parmname ) );
else return( 0 );
}

/* ============================= */
/* GETMULTILINE - get a multi-line text item from IPL spec file.
   End is marked by an empty (length=0) line.
   Leading white space is removed from all lines.
   Example
	title:  Number of days difference 
	 	\
	        Subgroup: 1B

	frame: yes
*/
   
getmultiline( parmname, firstline, fp, maxlen, result )
char *parmname, *firstline;
FILE *fp;
int maxlen;
char *result;
{
char buf[256];
char tok[80];
int nt;
int i;

/* strip off leading white space.. */
for( i = 0; i < strlen( firstline ); i++ ) 
	if( firstline[i] != ' ' && firstline[i] != '\t' ) break;

if( i < strlen( firstline ) ) sprintf( result, "%s\n", &firstline[i] );


while( fgets( buf, 255, fp ) != NULL ) {
	buf[ strlen( buf ) -1 ] = '\0';

	/* strip off leading white space.. */
	for( i = 0; i < strlen( buf ); i++ ) 
		if( buf[i] != ' ' && buf[i] != '\t' ) break;

	if( i >= strlen( buf )) break; /* end at first empty line */

	if( strlen( result ) + strlen( &buf[i] ) > maxlen ) {
		return( Eerr( 60, "Warning, multiline text truncated", parmname ) );
		}

	nt = sscanf( &buf[i], "%s", tok );
	if( nt == 1 && strcmp( tok, "\\" )==0 ) strcat( result, "\n" );
	else 	{
		strcat( result, &buf[i] );
		strcat( result, "\n" );
		}
	}
return( 0 );
}



/* ============================= */
/* TEXTDET - parse a text details setting and execute it */
/* attributes are: size=N color=COLOR style=B|I|BI|R align=L|C|R adjust=x,y */
/* ============================= */
textdet( parmname, spec, align, adjx, adjy, sizehint, stylehint )
char *parmname, *spec;
int *align;
double *adjx, *adjy;
int sizehint;
char *stylehint;
{
char at[5][80];
int nt;
int i;
int p;
char str[80];
char font[80];
char color[80];
double tmp;

/* defaults.. */
strcpy( color, Estandard_color );
strcpy( font, stylehint );
EPSfontname( Estandard_font, font );
p = Estandard_textsize + sizehint;
*align = '?'; /* align and offset will be sent back to caller.. */
*adjx = 0.0;
*adjy = 0.0;

/* parse spec.. */
nt = sscanf( spec, "%s %s %s %s %s", at[0], at[1], at[2], at[3], at[4] );

if( stricmp( at[0], "yes" )==0 || 
	strnicmp( at[0], "no", 2 )==0 ) nt = 0; /* drop thru.. */

for( i = 0; i < nt; i++ ) {
	if( strnicmp( at[i], "size=", 5 )==0 ) p = atoi( &at[i][5] );
	else if( strnicmp( at[i], "style=", 6 )==0 ) {
		if( !slmember( &at[i][6], "B I BI R" )) {
			Eerr( 82, "warning, text style must be B, I, BI, or R", &at[i][6] );
			strcpy( font, "R" );
			}
		else strcpy( font, &at[i][6] );
		EPSfontname( Estandard_font, font );
		/* was: EPSfontname( Ecurfont, font ); */
		}

	else if( strnicmp( at[i], "font=", 5 )==0 ) strcpy( font, &at[i][5] );
	else if( strnicmp( at[i], "color=", 6 )==0 ) strcpy( color, &at[i][6] );

	else if( strnicmp( at[i], "align=", 6 )==0 ) *align = at[i][6];
	else if( strnicmp( at[i], "adjust=", 7 )==0 ) {
		if( Eflip ) sscanf( &at[i][7],  "%lf,%lf", adjy, adjx );
		else sscanf( &at[i][7],  "%lf,%lf", adjx, adjy );
		}
	else 	{
		sprintf( str, "warning: %s is an unrecognized text details subattribute", at[i] );
		Eerr( 71, str, parmname );
		}
	}
Ecolor( color );
Efont( font );
Etextsize( p ); 
if( !Emember( *align, "LCR?" )) {
	Eerr( 72, "warning, align must be either L, C, or R", parmname );
	*align = '?';
	}
if( Using_cm ) { *adjx /= 2.54; *adjy /= 2.54; }
return( 0 );
}

/* ============================= */
/* LINEDET - parse a line details setting and execute it */
/* attributes are: width=W style=N color=COLOR dashscale=S */
/* ============================= */
linedet( parmname, spec, defaultwidth )
char *parmname, *spec;
double defaultwidth;
{
char at[5][80];
int nt;
int i;
int stat;
int style;
double width;
double ds;
char str[80];
char color[80];

/* defaults.. */
style = 0;
width = defaultwidth;
ds = 1.0;
strcpy( color, Estandard_color );

/* parse the spec.. */
nt = sscanf( spec, "%s %s %s %s %s", at[0], at[1], at[2], at[3], at[4] );

if( stricmp( at[0], "yes" )==0 || 
	strnicmp( at[0], "no", 2 )==0 ) nt = 0; /* drop thru.. */

for( i = 0; i < nt; i++ ) {
	if( strnicmp( at[i], "width=", 6 )==0 || strnicmp( at[i], "thick=", 6 )==0 ) {
		stat = num( &at[i][6], &width );
		if( stat != 0 || width < 0.1 ) return( Eerr( 73, "invalid line width", parmname ) );
		}
	else if( strnicmp( at[i], "style=", 6 )==0 ) style = atoi( &at[i][6] );
	else if( strnicmp( at[i], "color=", 6 )==0 ) strcpy( color, &at[i][6] );
	else if( strnicmp( at[i], "dashscale=", 10 )==0 ) ds = atof( &at[i][10] );
	else 	{
		sprintf( str, "%s is an unrecognized line details subattribute", at[i] );
		return( Eerr( 72, str, parmname ) );
		}
	}
Elinetype( style, width, ds );
Ecolor( color );
	
return( 0 );
}

/* ================== */
/* SYMDET - parse a point symbol detail spec and build a symbol code  */
/* shape=A style=S radius=R linewidth=W symcode= giffile=F gifscale=sx,sy */
/* =================== */
symdet( parmname, spec, symcode, radius )
char *parmname, *spec;
char *symcode; /* new symbol code copied into here */
double *radius;
{
char at[5][80];
int nt;
int i;
int stat;
int shape;
char style;
double linewidth;
char fillcolor[80];
char linecolor[80];
int symcodedone;
char str[128];
int doing_img;
double sx, sy;

/* defaults.. */
shape = 6;
style = 'a';
linewidth = 0.5; /* nice & thin */
/* strcpy( fillcolor, Estandard_color ); */
strcpy( fillcolor, "" );
strcpy( linecolor, Estandard_color );
*radius = 0.05;
symcodedone = 0;
doing_img = 0;
sx = 1.0; sy = 1.0;

/* parse the spec.. */
nt = sscanf( spec, "%s %s %s %s %s", at[0], at[1], at[2], at[3], at[4] );

if( stricmp( at[0], "yes" )==0 || 
	strnicmp( at[0], "no", 2 )==0 ) nt = 0; /* drop thru.. */


for( i = 0; i < nt; i++ ) {
	if( strnicmp( at[i], "linewidth=", 10 )==0 ) {
		stat = num( &at[i][10], &linewidth );
		if( stat != 0 || linewidth < 0.1 ) return( Eerr( 73, "invalid line width", parmname ) );
		}
	else if( strnicmp( at[i], "shape=", 6 )==0 ) {
		if( atoi( &at[i][6] ) > 0 ) shape = atoi( &at[i][6] );
		else if( strnicmp( &at[i][6], "tri", 3 )==0 ) shape = 1;
		else if( strnicmp( &at[i][6], "downtri", 7 )==0 ) shape = 2;
		else if( strnicmp( &at[i][6], "di", 2 )==0 ) shape = 3;
		else if( strnicmp( &at[i][6], "sq", 2 )==0 ) shape = 4;
		else if( strnicmp( &at[i][6], "pe", 2 )==0 ) shape = 5;
		else if( strnicmp( &at[i][6], "ci", 2 )==0 ) shape = 6;
		else if( strnicmp( &at[i][6], "lefttri", 7 )==0 ) shape = 8;
		else if( strnicmp( &at[i][6], "righttri", 8 )==0 ) shape = 7;
		else if( strnicmp( &at[i][6], "niceci", 6 )==0 ) shape = 9;
		}
	else if( strnicmp( at[i], "style=", 6 )==0 ) {
		if( strlen( &at[i][6] ) == 1 ) style = at[i][6];
		else if( strnicmp( &at[i][6], "out", 3 )==0 ) style = 'a';
		else if( strnicmp( &at[i][6], "spo", 3 )==0 ) style = 's';
		else if( strnicmp( &at[i][6], "fil", 3 )==0 ) {
			style = 'n';
			if( strlen( fillcolor ) < 1 ) strcpy( fillcolor, Estandard_color );
			}
		}
	
	else if( strnicmp( at[i], "fillcolor=", 10 )==0 ) strcpy( fillcolor, &at[i][10] );
	else if( strnicmp( at[i], "linecolor=", 10 )==0 ) strcpy( linecolor, &at[i][10] );
	else if( strnicmp( at[i], "radius=", 7 )==0 ) *radius = atof( &at[i][7] );
	else if( strnicmp( at[i], "sym", 3 ) == 0 ) { /* return a specified sym code as is.. */
		strcpy( symcode, at[i] );
		symcodedone = 1;
		}
#ifdef CUTSCG
	else if( strnicmp( at[i], "giffile", 7 ) == 0 ) {
		if( Device != 'g' || Png ) 
			Eerr( 7825, "gif symbols may only be used with -gif", "" );
		else	{
			strcpy( str, &at[i][8] );
			doing_img = 1;
			}
		}
#endif
	else if( strnicmp( at[i], "pngfile", 7 ) ==0 ) {    
		if( !Png ) 
			Eerr( 7825, "png symbols may only be used with -png", "" );
		else	{
			strcpy( str, &at[i][8] );
			doing_img = 1;
			}
		}
	else if( /* strnicmp( at[i], "gifscale", 8 ) == 0 || */
		strnicmp( at[i], "pngscale", 8 ) == 0 ) { 
		nt = sscanf( &at[i][9], "%lf,%lf", &sx, &sy );
		if( nt == 1 ) sy = sx;
		}
	else 	{
		sprintf( str, "%s is an unrecognized symboldetails subattribute", at[i] );
		return( Eerr( 72, str, parmname ) );
		}
	}

if( doing_img ) {
	if( sx < 0.001 ) sx = 1.0;
	if( sy < 0.001 ) sy = 1.0;
	stat = Eimload( str, sx, sy );
	if( stat != 0 ) Eerr( 5893, "warning, error on image load", str );
	strcpy( symcode, "gif" );
	return( 0 );
	}
if( !symcodedone ) sprintf( symcode, "sym%d%c%s", shape, style, fillcolor );
Elinetype( 0, linewidth, 1.0 );
Ecolor( linecolor );
if( Using_cm ) *radius /= 2.54; 
return( 0 );
}


/* ================= */
double avg( g, h )
double g, h;
{
return( (g + h) / 2.0 );
}

/* ================== */
/* GETRANGE - get a low/high range */
getrange( lineval, lo, hi, ax, deflo, defhi )
char *lineval;
double *lo, *hi;
char ax;
double deflo, defhi;
{
char s1[80], s2[80];
int nt;
nt = sscanf( lineval, "%s %s", s1, s2 );
if( nt == 2 ) {
	*lo = Econv( ax, s1 );
	*hi = Econv( ax, s2 );
	return( 0 );
	}
else if( nt == 1 ) {
	*lo = Econv( ax, s1 );
	*hi = defhi;
	return( 0 );
	}
else if( nt <= 0 ) {
	*lo = deflo;
	*hi = defhi;
	return( 0 );
	}
}
	

/* ============================= */
/* DO_SELECT - See if a row of data meets a condex selection condition. 
   Send back result of 1 if yes, 0 if no.
   Returns 0 if ok or an error code > 1 */
do_select( selectex, row, result )
char *selectex;
int row;
int *result;
{
int stat;
char recordid[10];
int i, j, len;
char *s;
char buf[256];
char *df[MAXITEMS];

 
strcpy( recordid, "" ); /* not used */
 
/* set up pointers to data items.. */
for( i = 0; i < Nfields[Dsel]; i++ ) df[i] = da( row, i );

stat = plvalue_subst( buf, selectex, df, FOR_CONDEX );
*result = 0;
if( stat > 1 ) return( stat );
*result = condex( buf );
return( 0 );
}

/* ============================= */
/* DO_SUBST - do value_subst substitution on data fields references such as @1 */
do_subst( out, in, row )
char *out; /* result */
char *in;	/* template with variable references e.g. @1 */
int row;	/* row of data to access */
{
int stat;
char recordid[10];
char data[MAXITEMS][DATAMAXLEN+1];
int i, j, len;
char *s;
 
strcpy( recordid, "" ); /* not used */
 
/* copy data items into data array and change embedded blanks to underscores.. */
for( i = 0; i < Nfields[Dsel]; i++ ) {
	s = da( row, i );
	len = strlen( s );
	for( j = 0; j < len; j++ ) {
		if( s[j] == ' ' ) data[i][j] = '_';
		else data[i][j] = s[j];
		}
	data[i][j] = '\0';
	}

stat = value_subst( out, in, data, recordid, NORMAL );
return( stat );
}
/* ============================= */
/* RM_CHAR - remove all instances of character c from buf */
rm_char( buf, c )
char *buf;
char c;
{
int i, j;
for( i = 0, j = 0; i < strlen( buf ); i++ ) {
	if( buf[i] == '@' ) continue;
	else buf[j++] = buf[i];
	}
buf[j] = '\0';
return( 0 );
}

/* ============================== */
/* CLOSE_TO - test two floating point numbers to see if 
	they are within a small tolerance. */

close_to( a, b, tol )
double a, b;
double tol;
{
if( a == b ) return( 1 );
else if( a > b && a - b < tol ) return( 1 );
else if( b > a && b - a < tol ) return( 1 );
else return( 0 );
}

/* =============================== */
/* FILE_TO_BUF - read file or execute command and place contents into buf.
   Shell expandable file name is ok.
   Returns 0 if ok, 1 if file not available */

file_to_buf( filename, mode, result, buflen )
char *filename;
int mode; /* 1 = file   2 = command */
char *result;
int buflen;
{
FILE *fp, *popen();
char buf[1000];

if( mode == 1 ) sprintf( buf, "cat %s", filename );
else strcpy( buf, filename );
fp = popen( buf, "r" );
if( fp == NULL ) return( 1 );
strcpy( result, "" );
while( fgets( buf, 999, fp ) != NULL ) {
	if( strlen( result ) + strlen( buf ) >= buflen ) {
		Eerr( 7254, "warning: truncated, capacity reached", filename );
		return( 2 );
		}
	strcat( result, buf );
	}
pclose( fp );
return( 0 );
}

/* ========================= */
/* SETFLOATVAR - set a DMS var to a double value */
setfloatvar( varname, f )
char *varname;
double f;
{
char buf[80];
int stat;
stat = declarevar( varname );
if( stat != 0 ) return( Eerr( 3889, "Sorry, too many variables, cannot declare", varname ) );
sprintf( buf, "%g", f );
stat = setvar( varname, buf );
if( stat != 0 ) return( Eerr( 3890, "Error on setting variable", varname ) );
return( 0 );
}
/* ========================= */
/* SETINTVAR - set a DMS var to an integer value */
setintvar( varname, n )
char *varname;
int n;
{
char buf[80];
int stat;
stat = declarevar( varname );
if( stat != 0 ) return( Eerr( 3891, "Sorry, too many variables, cannot declare", varname ) );
sprintf( buf, "%d", n );
stat = setvar( varname, buf );
if( stat != 0 ) return( Eerr( 3892, "Error on setting variable", varname ) );
return( 0 );
}

/* ========================== */
/* SETCHARVAR - set a DMS var to a char string value */
setcharvar( varname, s )
char *varname;
char *s;
{
int stat;
stat = declarevar( varname );
if( stat != 0 ) return( Eerr( 3893, "Sorry, too many variables, cannot declare", varname ) );
stat = setvar( varname, s );
if( stat != 0 ) return( Eerr( 3894, "Error on setting variable", varname ) );
return( 0 );
}

/* =========================== */
/* CONV_MSG - print a message to Errfp for a data item of invalid type */
conv_msg( row, col, aname )
int row, col;
char *aname;
{
N_convmsg++;
if( Supress_convmsg ) return( 0 );
fprintf( Errfp, "%s warning, %s skipping unplottable '%s' (rec=%d field=%d)\n",
	Eprogname, aname, da(row,col), row+1, col+1 );
return( 0 );
}

/* =========================== */
/* OTH_MSG - print a message to Errfp for a data item of invalid type */
oth_msg( msg )
char *msg;
{
N_convmsg++;
if( Supress_convmsg ) return( 0 );
fprintf( Errfp, "%s %s\n", Eprogname, msg );
return( 0 );
}

/* =========================== */

/* =========================== */
/* SUPPRESS_CONVMSG - suppress invalid type messages */
suppress_convmsg( mode )
int mode;
{
Supress_convmsg = mode;
return( 0 );
}

/* ============================= */
/* ZERO_CONVMSGCOUNT - zero out the conv msg counter */
zero_convmsgcount()
{
N_convmsg = 0;
return( 0 );
}
/* ============================= */
/* REPORT_CONVMSGCOUNT - report on what the conv msg count is.. */
report_convmsgcount()
{
return( N_convmsg );
}



/* ========================== */
/* SCALEBEENSET - return 1 if scaling has been set, 0 if not */
scalebeenset()
{
if( EDXhi - EDXlo > 0.0000001 ) return( 1 );
else return( 0 );
}

/* ===================== */
/* CATITEM - add a new data item */
catitem( s, d, avail )
char *s;
int *d;
int *avail;
{
D[ (*d)++ ] = &Databuf[ *avail ];
strcpy( &Databuf[ (*avail) ], s );
(*avail) += (strlen( s ) + 1);
return( 0 );
}

/* ======================== */
/* DEFAULTINC - given a min and a max, estimate a reasonable default inc */
static double threshvals[40] = { 0.000001, 0.00001, 0.0001, 0.001, 0.002, 0.005,
	0.01, 0.02, 0.05, 
	0.1, 0.2, 0.5, 
	1.0, 2.0, 5.0, 
	10.0, 20.0, 50.0, 
	100.0, 200.0, 500.0, 
	1000.0, 2000.0, 5000.0, 
	10000.0, 20000.0, 50000.0, 
	100000.0, 200000.0, 500000.0, 
	1000000.0, 2000000.0, 5000000.0, 
	10000000.0, 20000000.0, 50000000.0 };

defaultinc( min, max, inc )
double min, max, *inc;
{
double diff, h, fabs();
int i;

diff = max - min;
diff = fabs( diff );
h = diff / 10.0;  /* we want to have about 10 tics on an axis.. */

for( i = 0; ; i++ ) {
	if( h < threshvals[i] ) break;
	if( threshvals[i] > 49999999.0 ) break;
	}

*inc = threshvals[i];
return( 0 );
}

/* ======================== */
/* FMOD  */
double fmod( a, b )
double a, b;
{
double x, y;
 
x = a / b;
y = (int) (x) * b;
return( a - y );
}

/* ======================== */
/* REWRITENUMS - rewrite numbers, supplying a spacer (comma in US)
	every 3 zeros in large numbers.  If spacer is dot (.), European
	mode is presumed, decimal point is written as a comma, and
	large numbers are spaced using dot.   

 	Parameter s is modified.
*/

rewritenums( num )
char *num;
{
int i, j, decplace;
char s[40], tmp[40];

if(  Bignumspacer == '\0' ) return( 0 ); /* do nothing */

sscanf( num, "%s", s ); /* strip off any leading spaces */
decplace = strlen( s );
for( i = 0; i < strlen( s ); i++ ) {
	if( s[i] == '.' ) {
		decplace = i;
		if( Bignumspacer == '.' ) tmp[i] = ',';
		else tmp[i] = '.';
		}
	else tmp[i] = s[i];
	}
tmp[i] = '\0';

if( decplace > 4 ) {   /* add spacers.. */
	for( i = 0, j = 0; i < decplace; i++ ) {
		if( i > 0 && (decplace-i) % 3 == 0 ) {
			s[j++] = Bignumspacer;
			s[j++] = tmp[i];
			}
		else s[j++] = tmp[i];
		}
	s[j] = '\0';
	strcat( s, &tmp[i] ); /* append decimal point and rightward */
	}
else strcpy( s, tmp );

strcpy( num, s );
return( 0 );
} 


/* ================================= */
/* DO_LOGO - put on user/dir/filename identifier */
do_logo()
{
char buf[200];


#ifdef SOLARIS2
if( Emember( 'I', Flags ) && !Emember( 'N', Flags ) && Emember( Device, "pc" )) {
	char *login, *getlogin(), *cwd, *getcwd(), blanks[8];
	int m, d, y, hr, min, sec;
	char shortcwd[80];

	strcpy( blanks, " " );
	Esysdate( &m, &d, &y );
	Esystime( &hr, &min, &sec );

	login = NULL;
	login = getlogin(); /* portability */
	if( login == NULL ) login = blanks;
	if( strcmp( login, "pubsrc" )==0 ) return( 0 );

	cwd = NULL;
	cwd = getcwd( (char *)NULL, 255 );  /* portability */
	if( cwd == NULL ) cwd = blanks;
	else 	{
		if( strlen( cwd ) > 40 ) {
			sprintf( shortcwd, "...%s", &cwd[ strlen(cwd) - 40 ] );
			}
		else strcpy( shortcwd, cwd );
		}

	sprintf( buf, " %s  %s/%s  %02d/%02d/%02d  %02d:%02d", 
		login, shortcwd, Inputfile, m, d, y, hr, min );
	Etextsize( 6 );
	Emov( 0.3, 0.15 ); 
	Etext( buf );
	}
#endif
return( 0 );
}

