/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* Parse data into fields by assigning char * pointers to beginning
   of each field, and inserting nulls into the data buffer. 

   Delimmethods are: space = any amount of white space/ double quotes
		     comma = comma/double quotes
*/

/* #include "pl.x" */
#include <stdio.h>
#include <ctype.h>

extern FILE *Errfp;

static int shortwarnissued = 0;
static int longwarnissued = 0;


/* ============================================ */
parsedata( data, delimmethod, commentmarker, field, maxd, nr, nf, nd )
/* char *data; */
unsigned char *data; /* for LOCALE scg 3/15/00 */
char *delimmethod;
char *commentmarker;
char *field[]; /* array of pointers to fields */
int maxd; /* max # of elements in above array */
int *nr; /* number of newline-delimited records -  returned */
int *nf; /* number of fields per record */
int *nd; /* total number of fields */
{
int i, j, ip;
int datalen;
char buf[512];
int gettingws, gettingqs, leadingws;
int first;
char errstr[40];
int fieldsln;
char delim;
int disp; /* for error message display */
int lastnl;
static char * show();


/* Nrecords[Dsel] = 0; */
*nr = 0;
ip = 0;

datalen = strlen( data );

/* get overall delimiting method.. (w or c) */
if( strcasecmp( delimmethod, "comma" )==0 )  delim = 'c';
else if( strcasecmp( delimmethod, "tab" )==0 )  delim = 't';
else delim= 'w';

/* go through data and convert delimiters to 0, 1, or 2.. */
datalen = strlen( data );
gettingws = 0;
gettingqs = 0;
leadingws = 1;
/* lastnl = 0; */
lastnl = -1;
disp = 0;
for( i = 0; i < datalen; i++ ) {
	if( delim == 'w' && ( data[i] == ' ' || data[i] == '\t' ) ) {
		if( leadingws ) { data[i] = 0; continue; }
		else if( gettingws )data[i] = 1;
		else if( !gettingqs ) { data[i] = 0; gettingws = 1; }
		}

	else if( delim == 'c' && data[i] == ',' && !gettingqs ) data[i] = 0;
	else if( delim == 't' && data[i] == '\t' && !gettingqs ) data[i] = 0; /* added scg 7/7/99 */

	else if( data[i] == '"' ) {
		if( gettingqs && data[i-1] != '\\' ) { data[i] = 0; gettingqs = 0; }
		else if( delim == 'w' && !gettingws && !leadingws ) 
			;
			/* fprintf( Errfp, 
			   "Getdata warning, quote (\") should be preceded by whitespace: (%s)\n", 
				show(&data[disp],30) ); */
		else	{ 
			data[i] = 1; 
			gettingqs = 1; 
			}
		}
	else if( data[i] == '\n' ) { 
		if( gettingqs ) { 
			/* fprintf( Errfp, "Getdata warning, unmatched quotes (\"): (%s)\n", 
				show(&data[disp],30) ); */
			if( i > 0 )data[i-1] = 0;
			gettingqs = 0;
			}
		data[i] = 2; 
		leadingws = 1; 
		if( i - lastnl > 1 ) (*nr)++; /* Nrecords[Dsel]++; */ /* don't count it if its a blank line */
		disp = i+1;
		lastnl = i;
		}

	/* end of line comments (ws delimit only)*/
	else if( delim == 'w' && gettingws && 
	   strncmp( &data[i], commentmarker, strlen( commentmarker ) )==0 ) {
		leadingws = 1;
		(*nr)++; /* Nrecords[Dsel]++;  */
		/* skip to eol */
		for( ; i < datalen; i++ ) {
			if( data[i] == '\n' ) { data[i] = 2; break; }
			else data[i] = 1;
			}
		}
	
	else	{
		leadingws = 0;
		gettingws = 0;
		}
	}
if( data[i-1] != 2 ) (*nr)++; /* Nrecords[Dsel]++;   */



/* now, go through again, and wherever there is a character preceded by
   a 0 1 or 2, assign a pointer to it.  Change 2 to 0. 
   Also keep track of number of fields found on each line, to ensure
   that each line has same # of fields as first line; this is necessary
   to ensure that data array is uniform.
*/
first = 1;  /* flag: 1 if doing first line, 0 otherwise */
fieldsln = 0;  /* count # of fields on each line */
if( data[0] > 2 ) /* D[ip++] */ field[ip++] = &data[0];
disp = 0; 


for( i = 1; i < datalen; i++ ) {

		
	if( data[i] == 2 ) {
		if( first && ip > 0 ) { 
			(*nf) = ip; 
			first = 0; 
			}
		else	{
			/* handle a line with less than required # of fields.. */
			if( fieldsln > 0 && fieldsln < (*nf) /* Nfields[Dsel]*/ ) {
				if( !shortwarnissued ) 
					Eerr( 279, "warning: one or more records is short on fields.  Filling in nulls and continuing..", "" );
				shortwarnissued = 1;
				for( j = fieldsln; j < (*nf) /* Nfields[Dsel]*/; j++ )  /* set remaining fields */
					/* D[ip++] */ field[ip++] = "";
				}
			}
		fieldsln = 0;
		data[i] = 0;
		disp = i + 1;
		}
	else if( data[i] > 2 && data[i-1] < 3 ) {
		/* D[ip++] = &data[i]; */
		field[ip++] = &data[i];
		if( ip >= maxd ) return( Eerr( 411, "Sorry, max # data points exceeded", "" ) );
		fieldsln++;
		/* handle a line with greater than required # of fields.. */
		if( !first && fieldsln > (*nf) /* Nfields[Dsel]*/ ) {
			ip--; /* back up data counter..*/
			/* skip ahead to just before next newline */
			for( ; i < datalen; i++ ) if( data[i] == 2 ) { i--; break; }
			if( !longwarnissued ) 
				Eerr( 280, "warning: one or more records has too many fields.  Extra fields ignored and continuing..", "" );
			longwarnissued = 1;
			}
		}
	}
/* ND[Dsel] = ip; */ /* remember size of D */
*nd = ip;
if( ip % (*nr) /* Nrecords[Dsel]*/ != 0 ) return( 410 ); /* records have different #s of fields */
return( 0 );
}


/* ====================== */
/* for error messages, make presentable..*/
static char *show( s, len )
char *s;
int len;
{
int i;
static char rtn[40];
for( i = 0; i < len; i++ ) {
	if( s[i] == 2 || s[i] == '\n' ) break;
	/* else if( s[i] == 0 ) rtn[i] = '$'; */
	/* else if( s[i] == 1 ) rtn[i] = '!'; */
	else if( isprint( s[i] ) ) rtn[i] = s[i];
	/* else rtn[i] = '-'; */
	else rtn[i] = ' ';
  	}
rtn[i] = '\0';
return( rtn );
}
