/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include "pl.x"
#define TORAD 0.0174532
#define MAXSLICE 40


pie( specfile, clonefiles, nclonefiles  )
char *specfile, clonefiles[][ MAXPATH ];
int nclonefiles;
{
int i, j;
FILE *fp;
char buf[256];
char attr[40], val[256];
char lineval[256];
int nt;
FILE *openclonefile();
int icf;

int stat;
int align;
double adjx, adjy;
int df;
double cx, cy, radius;
double theta;
char color[MAXSLICE][40];
int lblfld;
double total;
double ux, uy, stop;
int first;
int ncolors;
double starttheta;
double sin(), cos();
double fval;
double x, y;
double oldx, oldy;
int nexpl;
double expl[MAXSLICE];
char outlinedetails[256];
char lablinedetails[256];
char labelmode[50];
int lbltextgiven;
int ibb;
char label[256];
char textdetails[256];
double stheta;
double lblfarout;
char pctstr[30];
char pctfmt[30];


strcpy( Eprogname, "pie" );
/* -------------------------- */
/* DEFAULTS */
/* All variables must start out with default values, */
/* since no variable is guaranteed to be set */
/* -------------------------- */
theta = 0.0;
cx = cy = -1.0;
radius = -1.0;
lblfld = -1;
strcpy( labelmode, "legend" );
strcpy( Bigbuf, "" );
ncolors = 0;
starttheta = 90.0 * TORAD;
total = 0.0;
nexpl = 0;
strcpy( outlinedetails, "" );
strcpy( lablinedetails, "" );
/* strcpy( labelmode, "labelonly" ); */
lbltextgiven = 0;
strcpy( textdetails, "" );
lblfarout = 0.0;
strcpy( pctfmt, "%.1f" );

/* -------------------------- */
/* open and process all clonefiles and lastly, specfile */
/* -------------------------- */
for( icf = 0; icf <= nclonefiles; icf++ ) {
 
    if( icf == nclonefiles ) {
        fp = fopen( specfile, "r" );
        if( fp == NULL ) return( Eerr( 117, "Could not open to read", specfile ) );
        }
    else        {
        fp = openclonefile( clonefiles[icf] );
        if( fp == NULL ) continue;
        }

    while( fgets( buf, 255, fp ) != NULL ) {
	nt = sscanf( buf, "%s %s", attr, val );		/* val has 2nd token */
	if( nt < 1 ) continue;
	if( nt == 1 ) strcpy( val, "" );
	if( icf < nclonefiles && stricmp( attr, "#proc" )==0 ) continue;
	i = 0;
        Egetok( buf, &i );
	while( buf[i] == ' ' || buf[i] == '\t' ) i++; /* skip over white space */
        strcpy( lineval, &buf[ i ] ); /* lineval has remainder of line, for sscanf-ing */
        lineval[ strlen( lineval ) - 1 ] = '\0'; /* strip off newline */
 
	/* strip off trailing colon if any.. */
	if( attr[ strlen( attr ) - 1 ] == ':' ) attr[ strlen( attr ) - 1 ] = '\0';


	/* attributes.. */

	if( stricmp( attr, "datafield" )==0 ) df = fref( val ) -1 ;
	else if( stricmp( attr, "center" )==0 ) getcoords( "center", lineval, &cx, &cy );
	else if( stricmp( attr, "radius" )==0 ) Elenex( val, X, &radius ); /* use X units.. */
	else if( stricmp( attr, "firstslice" )==0 ) starttheta = ((360-atof( val )) * TORAD ) + 90.0 * TORAD;
	else if( stricmp( attr, "total" )==0 ) total = atof( val );
	else if( stricmp( attr, "colors" )==0 ) {
		int ix;
		for( i = 0, ix = 0; i < MAXSLICE; i++ ) {
			strcpy( color[i], Egetok( lineval, &ix ) );
			if( strlen( color[i] ) < 1 ) break;
			}
		ncolors = i;
		}
	else if( stricmp( attr, "labels" )==0 ) {
		getmultiline( "labels", lineval, fp, MAXBIGBUF, Bigbuf );
		lbltextgiven = 1;
		}
	else if( stricmp( attr, "labelfield" )==0 ) lblfld = fref( val ) - 1;
	else if( stricmp( attr, "outlinedetails" )==0 ) strcpy( outlinedetails, lineval );
	else if( stricmp( attr, "lablinedetails" )==0 ) strcpy( lablinedetails, lineval );
	else if( stricmp( attr, "textdetails" )==0 ) strcpy( textdetails, lineval );
	else if( stricmp( attr, "labelmode" )==0 ) strcpy( labelmode, val );
	else if( stricmp( attr, "labelfarout" )==0 ) lblfarout = atof( val );
	else if( stricmp( attr, "pctformat" )==0 ) strcpy( pctfmt, val );
	else if( stricmp( attr, "explode" )==0 ) {
		int ix;
		for( i = 0, ix = 0; i < MAXSLICE; i++ ) {
			strcpy( buf, Egetok( lineval, &ix ));
			if( strlen( buf ) < 1 ) break;
			else expl[i] = atof( buf );
			}
		nexpl = i;
		}
	else Eerr( 1, "attribute not recognized", attr );
	}
    fclose( fp );
    }



/* -------------------------- */
/* overrides and degenerate cases */
/* -------------------------- */

if( Nrecords[Dsel] < 1 ) return( Eerr( 17, "No data has been read yet w/ proc getdata", "" ) );
if( df < 0 || df >= Nfields[Dsel] ) return( Eerr( 2840, "invalid datafield", "" ) );
if( cx < 0.0 || cy < 0.0 ) return( Eerr( 2841, "invalid center", "" ) );
if( radius < 0.0 || radius > 5.0 ) return( Eerr( 2842, "invalid radius", "" ) );


if( lblfarout < 0.001 && strnicmp( labelmode, "label", 5 )==0 ) lblfarout = 0.67;
if( lblfarout < 0.001 && strnicmp( labelmode, "line", 4 )==0 ) lblfarout = 1.3;

/* -------------------------- */
/* now do the plotting work.. */
/* -------------------------- */

/* compute total.. */
if( total <= 0.0 ) {
	total = 0.0;
	for( i = 0; i < Nrecords[Dsel]; i++ ) {
		total += atof( da( i, df ) );
		}
	}

ibb = 0;


for( j = 0; j < 2; j++ ) { /* first time - colors; 2nd time, lines */
	theta = starttheta + 6.28319;
	if( j == 1 ) {
		/* set line details for outline.. */
		linedet( "outlinedetails", outlinedetails, 0.5 );
		}
	for( i = 0; i < Nrecords[Dsel]; i++ ) {
		/* take val as % of total and convert to rads */
		fval = (atof( da( i, df ) ) / total ) * 6.28319; 
		stop = theta - fval;
		if( nexpl <= 0 ) {
			ux = 0.0;
			uy = 0.0;
			}
		else if( i >= nexpl ) {
			ux = (expl[nexpl-1]*radius) * cos( theta+(fval/-2.0) );
			uy = (expl[nexpl-1]*radius) * sin( theta+(fval/-2.0) );
			}
		else if( i < nexpl ) {
			ux = (expl[i]*radius) * cos( theta+(fval/-2.0) );
			uy = (expl[i]*radius) * sin( theta+(fval/-2.0) );
			}
			
		first = 1;
		stheta = theta;
		if( j == 1 && strnicmp( outlinedetails, "no", 2 )==0 ) goto DOLAB;
		for( ; theta > stop; theta -= 0.03 ) {
			if( theta - stop < 0.03 ) theta = stop;
			x = cx + (radius * cos( theta ));
			y = cy + (radius * sin( theta ));
			if( j == 0 ) {
				if( first ) { first = 0; oldx = x; oldy = y; continue; }

				/* done this way to avoid gd fill bug.. */
				Emov( cx+ux, cy+uy ); 
				Epath( oldx+ux, oldy+uy ); 
				Epath( x+ux, y+uy ); 
				if( i < ncolors ) Ecolorfill( color[i] );
				else if( ncolors > 0 ) Ecolorfill( color[ncolors-1] );
				else Ecolorfill( "0.8" );

				/* patch to compensate for gd fill bug, which sometimes leaves a
				   white line from center to 3oclock position.. */
				if( close_to( cy, y, 0.03 ) && x > cx ) {
					if( i < ncolors ) Ecolor( color[i] );
					else if( ncolors > 0 ) Ecolor( color[ncolors-1] );
					else Ecolor( "0.8" );
					Emov( cx+ux, cy+uy ); Elin( x+ux, cy+uy );
					}

				oldx = x; oldy = y;
				}
			else if( j == 1 ) {
				if( first ) { Emov( cx+ux, cy+uy ); Elin( x+ux, y+uy ); first = 0; }
				Elin( x+ux, y+uy );
				}
			}
		if( j == 1 ) Elin( cx+ux, cy+uy ); 

		/* labeling */
		DOLAB:
		if( j == 1 && strlen( labelmode ) > 0 ) {
			strcpy( label, "" );
			if( lblfld >= 0 ) strcpy( label, da( i, lblfld ) );
			else if( lbltextgiven ) Egetseg( label, Bigbuf, &ibb, "\n" );

			sprintf( pctstr, pctfmt, (atof( da( i, df ) ) / total)*100.0 );
			Evarsub( label, "@PCT", pctstr );
			Econvertnl( label );

			if( strnicmp( labelmode, "legend", 6 )==0 ) {
				add_legent( LEGEND_COLOR, label, color[i], "", "" );
				}
			else if( strnicmp( labelmode, "labelonly", 5 ) ==0 ) {
				double htheta;
				htheta = stop + ((stheta - stop) / 2.0 );
				x = cx + ( (radius * lblfarout) * cos( htheta ) );
				y = cy + ( (radius * lblfarout) * sin( htheta ) );
				textdet( "textdetails", textdetails, &align, &adjx, &adjy, -2,"R");
				Emov( x+ux, y+uy );
				Ecentext( label );
				linedet( "linedetails", outlinedetails, 0.5 ); /* restore */
				}
			else if( strnicmp( labelmode, "line+label", 4 ) ==0 ) {
				double htheta, px, py, w, z;

				/* w = radius * lblfarout;
				 * if( w < (1.1 * radius) ) z = lblfarout;
				 */
				w = radius * lblfarout;
				if( w < (1.1 * radius) ) z = lblfarout;
				else z = 1.1;

				htheta = stop + ((stheta - stop) / 2.0 );
				px = cx + ( (radius * 0.9 ) * cos( htheta ) );
				py = cy + ( (radius * 0.9 ) * sin( htheta ) );
		
				x = cx + ( (radius * z ) * cos( htheta ) );
				y = cy + ( (radius * z ) * sin( htheta ) );
				linedet( "lablinedetails", lablinedetails, 0.5 ); 
				Emov( px+ux, py+uy );
				Elin( x+ux, y+uy );
				if( x+ux < cx ) {
					Elin( (cx+ux)-w, y+uy );
					textdet( "textdetails", textdetails, &align, &adjx, &adjy, -2,"R");
					Emov( ((cx+ux)-w)-0.1, y+uy );
					Erightjust( label );
					}
				else 	{
					Elin( (cx+ux)+w, y+uy );
					textdet( "textdetails", textdetails, &align, &adjx, &adjy, -2,"R");
					Emov( cx+ux+w+0.1, y+uy );
					Etext( label );
					}
				linedet( "outlinedetails", outlinedetails, 0.5 ); /* restore */
				}
			}


		theta = stop;
		}
	}

return( 0 );
}
