/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* PLTAB - take an ascii table and display it in a proportional font,
   and apply optional embellishments, following a set of parsing rules. 

*/

/* #include "graphcore.h" */
#include "pl.h" 

#ifdef LOCALE
 #include <locale.h>
#endif


#define MAX_ROW	 150
#define MAX_COL	 200

#define LINE 0
#define TWO_BLANK 1

#define LEFT 0
#define CENTER 1
#define RIGHT 2
#define ROW 3

#define DEFAULT_FONT "/Helvetica"
#define DEFAULT_TEXTSIZE 10

static char t[MAX_COL][MAX_ROW];	/* text array */
static int nrows, maxc; /* dimensions of text array */
static int plotrowstart = 0, plotrowstop = 0;   /* range of text array applicable to plotting */
static double plotx1, ploty1, plotx2, ploty2; 	/* absolute plot location in Y */
static double catslide = 0.0;
static int curstub = 0;
static int npages;	/* number of pages as defined by #pagebreak's */
static char shchar[20];
static char delimchars[30] = "()";

vermessage()
{
fprintf( Diagfp, "\n\
Usage: pltab tablefile -large|-medium|-small|-landscape [options]\n\
\n\
ploticus 1.41 data display software (c) 1998-2001 Stephen C. Grubb\n\
Home page is www.sgpr.net.\n\
\n\
This program is free software; you can redistribute it and/or modify it\n\
under the terms of the GNU General Public License as published by the\n\
Free Software Foundation; either version 2 of the License, or (at your\n\
option) any later version.\n\
\n\
This program is distributed in the hope that it will be useful, but WITHOUT\n\
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n\
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.\n" );
exit( 1 );
}

main( argc, argv )
int argc;
char **argv;
{
char font[128];
double topmark;
double centerline;
double width;
double height;
int titlesize;
int bodysize;
int notitle;
double atof();
char buf[1024], buf2[1024];
char graycolor[40];
char color[40], bkcolor[40];
char outfile[128];
char tok[120], parm[120], plotfile[120];
int stat;

double left, right, bottom;

int i, iln, j, k, r, c, n;	/* counters, row, col */
int isnum;			/* true if token is a number */
int p;				/* position of decimal point in number */
int nblank;			/* blank line counter */
int first;			/* first time thru */
int coord[4];			/* returned by getphrase(), 
			holds col position of string left, center, and right, and row */
int shadeblocks;		/* table row shading option */
int shade;			/* shade flag */
int shadeblocksize;
int shadeopt;			/* 1 if -shade specified */
char area[60];			/* paper area */
int shademode1;			/* if 1, switch shades when line begins w/ character 
						other than blank */
int shademode2;			/* if 1, lines beginning w/ character other than blank 
					are shaded, others aren't */
int headerstart, bodystart;	/* line #s where header starts and where body starts */

int morestuff;	/* 1 if pagebreak issued */
int shmode; 	/* 1 if doing section headers */
int shsize;	/* size for section heading */
char shfont[40]; /* font for section heading */
char titlefont[40]; /* font for section heading */
char headerfont[40]; /* font for section heading */
int headersize;	/* size offset for column header */
int firstbody;	/* 1 if first #body has been encountered */
int vertruleline; /* line containing vertical rule info.. */
int chartmode;  /* 1 if making h rule on every line */
int autobreak;  /* 0 if no autobreak; otherwise it is the approx max # of body lines per page */
int nlines, cutline; /* related to autobreak */
int deducttop, nrulechar, n_notrulechar;
int chartmode2;			/* if 1, do ruling only for table lines that begin w/
					character other than blank */
int valused, found;
int nt;

char data[50][DATAMAXLEN+1]; /* to appease scripter()  - but not used for anything.. */
int musthave_fail;
char recordid[80];     /* '' */
struct scripterstate ss;
int firstnrows;
int singlepage; 
char continuemode[20];
int recurs;
int bodystart_specified;
int landscape;
double pbh;
char curfont[100]; /* keep track of current font - necessary to prevent interaction between #font
			and section headers, etc.   2/27/01 */

#ifdef LOCALE
 setlocale(LC_CTYPE, "" );
 setlocale(LC_COLLATE, "" );
#endif

strcpy( recordid, "" );

/* set pre-profile defaults.. */
Device = 'c'; /* default output is postscript */
strcpy( Viewer, "" );

/* read and process profile, etc. */
strcpy( Cmdlineparms, "" );
do_preliminaries();


strcpy( Eprogname, "pltab" );

/* usage message */
if( argc < 2 ) {
	vermessage();
	}

/* ---------------------------------------------- */
/* ---------------------------------------------- */
/* ---------------------------------------------- */

/* set variables to defaults.. */
strcpy( font, DEFAULT_FONT );
strcpy( curfont, font );
topmark = 11.5 - 1.2;
centerline = 4.3;
width = -1.0;
height = -1.0;
bodysize = DEFAULT_TEXTSIZE;
notitle = 0;
strcpy( Inputfile, "" );
strcpy( color, "black" );
strcpy( bkcolor, "white" );
strcpy( graycolor, "gray(0.98)" );
shade = 0;
shadeopt = 0;
shadeblocks = 0;
shadeblocksize = 3;
shademode1 = 0;
shademode2 = 0;
shmode = 0;
strcpy( shfont, "" );
shsize = 0;
headersize = 0;
strcpy( headerfont, "" );
titlesize = 3;
strcpy( titlefont, "B" );
chartmode = 0;
chartmode2 = 0;
autobreak = 0;
Debug = 0;
Winx = 100;
Winy = 0;
Winh = 11.0;
Winw = 8.5;
strcpy( plotfile, "" );
Bkcolorgiven = 0;
strcpy( Cmdlineparms, "" );
singlepage = 0;
strcpy( continuemode, "c" );
strcpy( Flags, "I" ); /* make -id the default */
recurs = 0;
strcpy( shchar, "" );
landscape = 0;
 

/* process command line options.. */
for( i = 1; i < argc; i++ ) {
	if( strcmp( argv[i], "-small" )==0 ) {
		/* pltab myfile -top 1.5 -bodysize 14 -title 1 B */
		topmark = 9.5;
		bodysize = 14;
		titlesize = 1;
		}
	else if( strcmp( argv[i], "-medium" )==0 ) {
		topmark = 10.0;
		bodysize = 12;
		titlesize = 2;
		}
	else if( strcmp( argv[i], "-large" )==0 ) {
		topmark = 10.3;
		bodysize = 10;
		titlesize = 2;
		height = 9.8;
		}
	else if( strcmp( argv[i], "-landscape" )==0 ) {
		topmark = 7.6;
		centerline = 5.3;
		width = 10.0;
		/* height = 7.2; */
		landscape = 1;
		Winh = 8.5;
		Winw = 11.0;
		}

	else if( strcmp( argv[i], "-font" )==0 ) {
		i++;
		if( i < argc ) {
			strcpy( font, argv[i] );
			strcpy( Estandard_font, argv[i] ); /* set the overall default as well.. */
			strcpy( curfont, font );
			}
		}
	else if( strcmp( argv[i], "-top" )==0 ) {
		i++;
		if( i < argc ) topmark = 11 - atof( argv[i] );
		}
	else if( strcmp( argv[i], "-midline" )==0 ) {
		i++;
		if( i < argc ) centerline = atof( argv[i] );
		}
	else if( strcmp( argv[i], "-width" )==0 ) {
		i++;
		if( i < argc ) width = atof( argv[i] );
		}
	else if( strcmp( argv[i], "-height" )==0 ) {
		i++;
		if( i < argc ) height = atof( argv[i] );
		}
	else if( strcmp( argv[i], "-title" )==0 ) {
		i++;
		if( i < argc ) titlesize = atoi( argv[i] ); 
		i++;
		if( i < argc ) strcpy( titlefont, argv[i] );
		}
	else if( strcmp( argv[i], "-header" )==0 ) {
		i++;
		if( i < argc ) headersize = atoi( argv[i] );
		i++;
		if( i < argc ) strcpy( headerfont, argv[i] );
		}
	else if( strcmp( argv[i], "-bodysize" )==0 ) {
		i++;
		if( i < argc ) bodysize = atoi( argv[i] );
		}
	else if( strcmp( argv[i], "-shadecolor" )==0 ) {
		i++;
		if( i < argc ) strcpy( graycolor, argv[i] );
		}



	/* other options.. */
	else if( strcmp( argv[i], "-shade" )==0 ) shadeopt = 1;
	else if( strcmp( argv[i], "-shadesize" )==0 ) {
		i++;
		if( i < argc ) shadeblocksize = atoi( argv[i] );
		if( shadeblocksize < 1 || shadeblocksize > 40 ) {
			shadeblocksize = 3;
			Eerr( 2589, "pltab: bad shadesize", argv[i] );
			}
		}
	else if( strcmp( argv[i], "-shademode1" )==0 ) {
		shademode1 = 1;
		shadeopt = 1;
		}
	else if( strcmp( argv[i], "-shademode2" )==0 ) {
		shademode2 = 1;
		shadeopt = 1;
		}
	else if( strcmp( argv[i], "-sectionheading" )==0 ) {
		i++;
		shmode = 1;
		if( i < argc ) shsize = atoi( argv[i] );
		i++;
		if( i < argc ) strcpy( shfont, argv[i] );
		}
	else if( strcmp( argv[i], "-notitle" )==0 ) notitle = 1;
	else if( strcmp( argv[i], "-chartmode" )==0 ) chartmode = 1;
	else if( strcmp( argv[i], "-chartmode2" )==0 ) chartmode2 = 1;
	else if( strcmp( argv[i], "-multipage" )==0 ) autobreak = 40;
	else if( strcmp( argv[i], "-singlepage" )==0 ) { autobreak = 0; singlepage = 1; }
	else if( strcmp( argv[i], "-linesperpage" )==0 ) {
		i++;
		if( i < argc ) autobreak = atoi( argv[i] );
		if( autobreak > 0 && autobreak < 10 ) autobreak = 40;
		}
	else if( strcmp( argv[i], "-continue" )==0 ) {
		i++;
		if( i < argc ) strcpy( continuemode, argv[i] );
		}
	else if( strcmp( argv[i], "-recurs" )==0 ) recurs = 1;
	else if( strcmp( argv[i], "-shchar" )==0 ) {
		i++;
		if( i < argc ) strcpy( shchar, argv[i] );
		}
	else if( strcmp( argv[i], "-delimchars" )==0 ) {
		i++;
		if( i < argc ) strcpy( delimchars, argv[i] );
		if( strcmp( delimchars, "none" )==0 ) strcpy( delimchars, "" );
		}

	else if( strcmp( argv[i], "-showargs" )==0 ) {
		int ii;
		fprintf( Diagfp, "-------------\n" );
		for( ii = 0; ii < argc; ii++ ) fprintf( Diagfp, "%s ", argv[ii] );
		fprintf( Diagfp, "\n" );
		}

	else if( strcmp( argv[i], "-png" )==0 && !devavail( "png" ) ) { 
		char *p[100];
		if( strcmp( &argv[0][strlen(argv[0])-8], "pltabpng" )==0 ) {
			Eerr( 5279, "png not available in pltabpng", "" );
			exit(1);
			}
		p[0] = "pltabpng";
		for( i = 1; i < argc; i++ ) p[i] = argv[i];
		p[argc] = NULL;
		execvp( "pltabpng", argv );
		fprintf( Errfp, "PNG not supported in this build (pltabpng not found)\n" );
                exit(1);
		}

	/* standard ploticus args.. */
	else	{
 		if( i+1 < argc ) process_arg( argv[i], argv[i+1], &valused, &found );
                else process_arg( argv[i], "", &valused, &found );

		/* table file.. */
                if( !found ) { 
			if( strlen( Inputfile ) > 0 || argv[i][0] == '-' ) 
				Eerr( 2958, "pltab: parameter error", argv[i] );
			strcpy( Inputfile, argv[i] );
			}
                i += valused;
		}
	}


/* ---------------------------------------------- */
/* ---------------------------------------------- */
/* ---------------------------------------------- */

if( strlen( Cmdline_outfile ) > 0 &&
    strcmp( Cmdline_outfile, "-" )!= 0 )
        Esetoutfilename( Cmdline_outfile );
else 	{   /* set default output file names for gif and eps.. */
        if( Device == 'e' ) { 
		strcpy( outfile, "out.eps" ); 
		fprintf( Diagfp, "creating out.eps\n" ); 
		}
        else if( Device == 'g' ) {
		if( Png ) {
			strcpy( outfile, "out.png" );
			fprintf( Diagfp, "creating out.png\n" ); 
			}
		else	{
			strcpy( outfile, "out.gif" );
			fprintf( Diagfp, "creating out.gif\n" ); 
			}
		}
	Esetoutfilename( outfile );
        }


/* if a viewcommand is specified and an outfile has not been and the device
   is paginated postscript, we need to set the output file to out.ps */
if( strlen( Viewer ) > 0 && strlen( outfile ) < 1 ) {
        if( Device == 'c' || Device == 'p' ) strcpy( outfile, "out.ps" );
        }


/* DEVICE variable.. */
if( getvar( "DEVICE", buf ) != 0 ) { /* DEVICE not given on command line, set DEVICE */
        declarevar( "DEVICE" );
        devnamemap( &Device, buf, 2 );
        setvar( "DEVICE", buf );
        }
else    { /* DEVICE given on command line, set Device from DEVICE */
        getvar( "DEVICE", buf );
        devnamemap( &Device, buf, 1 );
        }

 
Esetsize( Winw, Winh, Winx, Winy );


/* do size offsets */
titlesize = bodysize+titlesize;
headersize = bodysize+headersize;
shsize = bodysize+shsize;
EPSfontname( font, titlefont );
if( Debug )fprintf( Diagfp, "title font: %s  size: %d\n", titlefont, titlesize ); 
EPSfontname( font, headerfont );
if( Debug )fprintf( Diagfp, "header font: %s  size: %d\n", headerfont, headersize ); 
EPSfontname( font, shfont );
if( Debug )fprintf( Diagfp, "s.h. font: %s  size: %d\n", shfont, shsize );


/* sanity check.. */
if( topmark <= 0.0 || centerline <= 0.0 || titlesize <= 4 || bodysize <= 3 || shsize <= 3 ||
    width == 0.0 || height == 0.0 || width > 11.0 || height > 11.0 ||
    topmark > 11.0 || centerline > 8.5 ) {
	Eerr( 8921, "argument out of range", "" );
	exit(1);
	}

/* initialize graphics package */
Einit( Inputfile, Device );
Initialized = 1;
if( strcmp( color, "black" ) != 0 ) Ecolor( color );
Ebackcolor( bkcolor );
Enormline();
if( landscape ) Epaper(1);

Eclr();

/* if( !eps ) */
/* if( Device != 'e' ) Ecblock( 0.0, 0.0, 8.5, 11.0, bkcolor, 0 ); */ /* needed? commented out scg 1/22/01 */
	
/* open ascii table file */
stat = scripter_open( &ss, Inputfile );
if( stat != 0 ) { Eerr( 22, "Cannot open table text file", Inputfile ); vermessage(); }


/* Get some info about table.. count number of lines in file, find where bodystart is,
   then closeit, (it will be re-opened later..)   */
bodystart = 0;
bodystart_specified = 0;
nlines = 0;
musthave_fail = 0;
/* while( fgets( buf, 1023, fp ) != NULL ) { */
while( 1 ) {
	stat = scripter( &ss, buf, recordid, data );
 
        if( stat == 408 ) {
                musthave_fail = 1;
                continue;
                }
        else if( stat > 255 || musthave_fail ) {
                Eerr( 23, "Fatal script processing error.", "" );
		exit(1);
		}
 
        else if( stat != 0 ) break;

	if( strncmp( buf, "#pltaboptions", 13 )==0 ) {
		/* parse line, set **p to point to args, call execvp() to reinvoke */
		int ix, ia;
		char *p[100], *strtok();

		if( recurs ) continue;

		ix = 0; ia = 0;
		Egetok( buf, &ix ); /* get #options */
		strtok( buf, " \t\n" );
		p[ia++] = argv[0];
		p[ia++] = "-recurs";
		while( 1 ) {
			p[ia++] = strtok( NULL, " \t\n" );
			if( p[ia-1] == NULL ) { ia--; break; }
			}
		/* add args from command line */
		for( i = 1; i < argc; i++ ) p[ia++] = argv[i];
		p[ia] = NULL;
		/* for( i = 0; i < ia; i++ ) fprintf( stderr, "%s ", p[i] ); 
		 * fprintf( stderr, "\n" ); 
		 */

		execvp( "pltab", p );
		}

	/* try to find where header ends and body begins */
	if( strncmp( buf, "#body", 5 ) ==0 ) {
		bodystart = nlines+1; /* +1 scg */
		bodystart_specified = 1;
		}
	nrulechar = n_notrulechar = 0;
	for( j = 0; j < strlen( buf ) -1; j++ ) {
		if( Emember( buf[j], "_-=" )) nrulechar++;
		else if( !Emember( buf[j], "_-= " )) n_notrulechar++;
		}

	/* scg 12/1/99 */
	if( !bodystart_specified &&
	    nlines < 20 && nrulechar > 10 && n_notrulechar < 5 ) bodystart = nlines+1; /* +1scg*/

	if( slmember( buf, "*#plotbegin*" ) && nlines - bodystart <= 3 ) bodystart = nlines+1;
		/* above to solve plotting problem on pp 2 and up  scg 6/29 */

	nlines++;	
	}
/* fclose( fp ); */
scripter_close( &ss ); 


/* try to automatically determine if multipage -added scg 6/22/99 */
if( (nlines * (bodysize/72.0)) > 9.5 && autobreak == 0 && !singlepage ) autobreak = 40; 


if( autobreak ) {
	Doing_multipage_pltab = 1; /* for pl procs */

	/* deduct # of lines in title/header.. */
	/* nlines -= (bodystart - 1);  */
	nlines -= (bodystart - 2);  /* -2scg*/
	/* find number of pages to do.. */
	npages = nlines / autobreak; 
	if( nlines % autobreak != 0 ) npages++; 

	/* find point at which to automatically break each page (title and headers
		will be included in line count) */
	cutline = (nlines / npages) + bodystart + 1;
	
	if( Debug ) fprintf( Diagfp, "Auto pg breaks: %d lines of tbl body/page (%d lines in all)\n", 
		(nlines/npages), nlines );
	}

/* read ascii file into t array, converting tabs to spaces */
/* initialize storage */
for( i = 0; i < 132; i++ ) {
	for( j = 0; j < 132; j++ ) t[i][j] = ' ';
	t[i][131] = '\0';
	}

npages = 1;
headerstart = 0;
maxc = 0;

/* open table file.. */
stat = scripter_open( &ss, Inputfile );
if( stat != 0 ) { Eerr( 22, "Cannot open table text file", Inputfile ); vermessage(); }



/* Each page begins here.. */
PAGE:

r = 0; 
if( npages == 1 ) curstub = 0;
else curstub = -1;

/* for continuation pages.. */
if( npages > 1 ) {
	r = bodystart;  /* this preserves title and header  (was +1) */


	if( Edev == 'x' ) do_x_button( "More.." );

	else if( Emember( Device, "ge" )) {
		/* gif & eps: terminate current page and start a new one, 
		   using  outpN as the result file name where N is page #. */
                Eshow();
                Eendoffile();
		if( Device == 'g' ) {
			if( Png ) sprintf( buf, "outp%d.png", npages );
			else sprintf( buf, "outp%d.gif", npages );
			}
		else if( Device == 'e' ) sprintf( buf, "outp%d.eps", npages );
		fprintf( Diagfp, "Creating %s\n", buf );
                Esetoutfilename( buf );
                Einit( Inputfile, Device );
                }
	else Eprint();


	if( landscape ) Epaper( 1 );

	/* if( !eps ) Ecblock( 0.0, 0.0, 8.5, 11.0, bkcolor, 0 ); */
	Eclr();
	
	/* print new page number at top of next page */
	sprintf( buf, "(page %d)", npages );
	Ecolor( Estandard_color );
	Etextsize( 12 );
	Emov( 7.8, 10.0 );
	Erightjust( buf );
	shadeblocks = 0;
	}
morestuff = 0;

/* rewind( fp ); */




musthave_fail = 0;
/* while( fgets( buf2, 1023, fp ) != NULL ) { */
while( 1 ) {
	stat = scripter( &ss, buf2, recordid, data );
        if( stat == 408 ) {
                musthave_fail = 1;
                continue;
                }
        else if( stat > 255 || musthave_fail ) {
                Eerr( 23, "Fatal script processing error.", "" );
		exit(1);
		}
 
        else if( stat != 0 ) break;

	if( strncmp( buf2, "#pltaboptions", 13 )==0 ) continue; /* skip completely */

	Eexpand_tabs( buf, buf2 );
	if( r >= MAX_ROW ) {
		Eerr( 8922, "Sorry, too many lines. Table truncated", "" );
		break;
		}
	if( strlen(buf) >= MAX_COL ) {
		Eerr( 2749, "Sorry, line too long; skipping it", buf );
		continue;
		}
	c = 0;
	for( i = 0; i < strlen( buf ); i++ ) t[ r ][ c++ ] = buf[ i ];
	t[r][c] = '\0';

	/* remember longest line, but lines beginning with # don't count */
	if( c > maxc && t[r][0] != '#' ) maxc = c; 

	r++;

	if( autobreak > 0 && r > cutline ) {  /* insert the word #pagebreak */
		strcpy( buf, "#pagebreak" ); 
		strcpy( t[r], "#pagebreak" );
		r++;
		}

	if( strncmp( buf, "#pagebreak", 10 )==0 ) {
		if( Device == 'e' ) {
			Eerr( 4929, "warning: #pagebreak used in eps mode", "" );
			}
		morestuff = 1;
		break;
		}

	}

nrows = r;

/* try to preserve correct line spacing on last page of a multi-page doc.. */
if( npages <= 1 ) firstnrows = nrows;
if( nrows != firstnrows ) height = (height/(double)firstnrows) * nrows;
	
/* nrows now holds # of rows (lines);  maxc holds length of longest line. */
if( Debug ) fprintf( Diagfp, "page %d: %d rows, longest one is %d chars.\n", 
	npages, nrows, maxc );


	

/* determine scale area for table rendering.. */
Etextsize( bodysize );

if( height < 0.0 ) height = (nrows * (Ecurtextheight+0.02));

/* guess at width based on longest line and point size of body text.. */
if( width < 0.0 ) {
	if( Emember( Device, "xg" )) width = maxc * ((double)bodysize/115.0);
	else width = maxc * ((double)bodysize/150.0);
	}

left = centerline - (width/2.0);
right = centerline + (width/2.0);
bottom = topmark - height;
if( Debug )
 fprintf( Diagfp, "page %d: table will be rendered in X [%g to %g] and Y [%g to %g]\n",
	npages, left, right, bottom, topmark );
Escale_x( left, right, 0.0, (double)maxc );
Escale_y( bottom, topmark, 0.0, (double)nrows );

if( notitle ) { 
	r = 0; 
	goto BODY; 
	}

/* get ready to do title.. */
Efont( titlefont );
Etextsize( titlesize );

/* do the title.. */
nblank = -999;
for( r = 0; r < nrows ; r++ ) {
	/* count blank lines */
	if( headerstart > 0 && r >= headerstart ) break;
	if( bodystart > 0 && r >= bodystart ) break;
	if( sscanf( t[r], "%s", buf ) < 1 ) {
		nblank++;
		if( nblank >= 2 ) break;
		else continue;
		}
	else nblank = 0;
	if( strncmp( t[r], "#header", 7 )==0 ) break;
	if( strncmp( t[r], "#body", 5 )==0 ) break;
	if( strncmp( t[r], "#center", 7 )==0 ) {
		getphrase( buf, &t[r][8], coord, &c, LINE );
		coord[ROW] = r;
		coord[LEFT] = 0;
		coord[RIGHT] = maxc;
		center( buf, coord ); 
		continue;
		}
	getphrase( buf, t[r], coord, &c, LINE );
	coord[ROW] = r;
	center( buf, coord ); 
	}
if( r > 15 ) 
	Eerr( 2059, "Warning: use at least two blank lines to separate table title from body", "" );

/* go to header settings.. */
if( strlen( headerfont ) > 0 )Efont( headerfont );
else Efont( font );
Etextsize( headersize );

/* do BODY of table */
BODY:

shade = 0;
headerstart = r;
if( bodystart == 0 ) bodystart = r;
vertruleline = -1;
if( shadeopt ) iln = -1;
firstbody = 0;
if( Debug ) {
	fprintf( Diagfp, "table header start detected at line %d\n", headerstart+1 );
	fprintf( Diagfp, "table body start detected at line %d\n", bodystart+1 );
	}
for( ; r < nrows; r++ ) {
	c = 0;

	if( r == bodystart ) {
		firstbody = 1;
		if( shadeopt ) shadeblocks = 1;
		Efont( font );
		Etextsize( bodysize );
		}

	/* PLOT operators */
	nt = sscanf( t[r], "%s %s", tok, parm );
	if( nt < 1 ) strcpy( tok, "" );
	if( strcmp( tok, "#plotbegin" )==0 ) {
		if( nt != 2 && strlen( plotfile ) < 1 ) {
			Eerr( 2954, "#plotbegin: expecting name of a ploticus control file", "" );
			continue;
			}
		strcpy( plotfile, parm );
		plotrowstart = r+1;
		plotrowstop = nrows-1; /* as a fallback, should be set with #plotend or pagebreak */
		/* ploty2 = Eay( EDYhi-((double)r-0.5) ); */
		ploty2 = Eay( EDYhi-((double)r) ); 
		for( i = 0; i < strlen( t[r] ); i++ ) if( t[r][i] == '#' ) break;
		plotx1 = Eax( (double)i );
		/* see if there is a right side.. */
		plotx2 = 8.0; /* fallback */
		ploty1 = 1.0; /* fallback */
		for( i++ ; i < strlen( t[r] ); i++ ) if( t[r][i] == '#' ) break;
		if( i != strlen( t[r] ) ) plotx2 = Eax( (double)i+1 ); /* adding 1 
									for proper placement*/
		continue;
		}
	else if( strcmp( tok, "#plotend" )==0 ) {  /* this needs to be replicated at page brks */
		int faker; /* need to 'virtually' add one */
		faker = r+1;
		plotrowstop = faker-1; 
		ploty1 = Eay( EDYhi-((double)faker-2) ); /* r-2 for proper placement */
		continue;
		}

	/* special operators */
	if( t[r][0] == '#' ) {
		if( strncmp( t[r], "#shadeblocks-on", 15 )==0 ) { shadeblocks = 1; iln = -1; }
		else if( strncmp( t[r], "#shadeblocks-off", 16 )==0 ) { 
			shadeblocks = 0; 
			shade = NO; 
			}
		else if( strncmp( t[r], "#shade-on", 9 )==0 ) shade = 1;
		else if( strncmp( t[r], "#shade-off", 10 )==0 ) shade = 0; 
		else if( strncmp( t[r], "#header", 7 )==0 ) headerstart = r;
		else if( strncmp( t[r], "#pagebreak", 10 )==0 ) {
			if( vertruleline >=0 ) 
				drawvertrules( t[vertruleline], vertruleline, nrows-2 );
			if( tolower(continuemode[0]) == 'c' ) {
				Emov( 4.25, EYlo-0.15 ); Etextsize( 8 ); Ecentext( "(Continued..)" );
				}
			else if( tolower(continuemode[0]) == 'l' ) {
				Emov( EXlo, EYlo-0.15 ); Etextsize( 8 ); Etext( "(Continued..)" );
				}
			else if( tolower(continuemode[0]) == 'r' ) {
				Emov( EXhi, EYlo-0.15 ); Etextsize( 8 ); Erightjust( "(Continued..)" );
				}

			/* PAGEBREAK: PLOT here, also set outfilename for next page (gif, eps) */
			/* Do #plotend stuff here.. */
			plotrowstop = r; /* used to be r-1 scg */
			ploty1 = Eay( EDYhi-((double)r-1) ); /* was r-2 */

			do_logo();

			Eflush();
			if( strlen( plotfile ) > 0 ) {
				stat = exec_plfile( plotfile );
				if( stat != 0 ) Eerr( 3882, "Error while executing plotfile", plotfile );
				}

			npages++;
			goto PAGE;
			}
		else if( strncmp( t[r], "#textsize", 9 )==0 ) {
			char stok[20];
			sscanf( t[r], "%*s %s", stok );
			if( Emember( stok[0], "+-" )) bodysize += atoi( stok );
			else bodysize = atoi( stok );
			Etextsize( bodysize );
			}
		else if( strncmp( t[r], "#font", 5 )==0 ) {
			char stok[80];
			sscanf( t[r], "%*s %s", stok );
			EPSfontname( font, stok );
			Efont( stok );
			strcpy( curfont, stok );
			}
		else if( strncmp( t[r], "#stopvr", 7 )==0 ) {
			if( vertruleline >=0 ) {
				drawvertrules( t[vertruleline], vertruleline, r );
				vertruleline = -1;
				}
			}
		else if( strncmp( t[r], "#buttvr", 7 )==0 ) {
			if( vertruleline >=0 ) {
				drawvertrules( t[vertruleline], vertruleline, r );
				vertruleline = -1;
				}
			Emovu( 0.0, (double)(nrows-r) );
			Elinu( (double)maxc-1, (double)(nrows-r) );
			}
		else if( strncmp( t[r], "#body", 5 )==0 ) ;
			/* { 
			 * bodystart = r;
			 * firstbody = 1;
			 * if( shadeopt ) shadeblocks = 1;
			 * Efont( font );
			 * Etextsize( bodysize );
			 * }
			 */
		else if( strncmp( t[r], "#hr-light", 9 )==0 ) {
			Elinetype( 0, 0.4, 1.0 );
			Emovu( 0.0, (double)(nrows-r) );
			Elinu( (double)maxc-1, (double)(nrows-r) );
			Enormline();
			}
		else if( strncmp( t[r], "#hr-heavy", 9 )==0 ) {
			Elinetype( 0, 1.4, 1.0 );
			Emovu( 0.0, (double)(nrows-r) );
			Elinu( (double)maxc-1, (double)(nrows-r) );
			Enormline();
			}
		else if( strncmp( t[r], "#hr-double", 10 )==0 ) {
			Elinetype( 0, 0.4, 1.0 );
			Emovu( 0.0, ((double)(nrows-r))-0.1 );
			Elinu( (double)maxc-1, ((double)(nrows-r))-0.1 );
			Emovu( 0.0, ((double)(nrows-r))+0.1 );
			Elinu( (double)maxc-1, ((double)(nrows-r))+0.1 );
			Enormline();
			}
		else if( strncmp( t[r], "#hr", 3 )==0 ) {
			Emovu( 0.0, (double)(nrows-r) );
			Elinu( (double)maxc-1, (double)(nrows-r) );
			}

		if( strncmp( t[r], "#body", 5 )==0 || strncmp( t[r], "#header", 6 )==0 ||
		    strncmp( t[r], "#hr", 3 )==0 || strncmp( t[r], "#vr", 3 ) ==0 ) {
			/* look for vert ruling marks */
			for( i = 5; i < strlen( t[r] ); i++ ) {
				if( t[r][i] == '|' || t[r][i] == 'H' ) vertruleline = r;
				}
			}
		continue;
		}
			
	if( shadeblocks ) {
		iln++;
		if( shademode1 ) {
			if( t[r][0] != ' ' ) shade = !shade;
			}
		else if( shademode2 ) {
			if( firstbody && is_sh( t[r][0] ) )
				/* t[r][0] != ' ' && t[r][0] != '\0' && t[r][0] != '\n' )  */
				shade = !shade;
			}
		else	{
			/* if( ( iln / shadeblocksize ) == 0 ) shade = !shade; */
			if( ( iln % shadeblocksize ) == 0 ) shade = !shade;
			}
		}
	
	if( shade ) {
		Ecblock( EXlo, Eay((double)(EDYhi-r)-.15), 
			EXhi, Eay((double)(EDYhi-r)+.85), graycolor, 0 );
		if( shademode2 ) shade = 0;
		}

	if( ( chartmode && firstbody ) || 
		( chartmode2 && firstbody && t[r][0] != ' ' && t[r][0] != '\n' ) ) {  
		/* chartmode: do a thin horizontal rule.. */
		Elinetype( 0, 0.4, 1.0 );
		Emovu( 0.0, (double)((nrows-r)-0.1) );
		Elinu( (double)maxc-1, ((double)(nrows-r)-0.1) );
		Enormline();
		}

			
	if( sscanf( t[r], "%s", buf ) < 1 ) continue; /* blank lines.. do nothing */


	if( shmode && firstbody && is_sh( t[r][0] ) ) {
		Efont( shfont );
		Etextsize( shsize );
		}
	if( shmode && firstbody && !is_sh( t[r][0] ) ) {
		Efont( curfont );
		Etextsize( bodysize );
		}
		


	/* now process fields in body */
	first = 1;
	while( 1 ) {

		/* get a phrase, and position of beginning and end.. */
		getphrase( buf, t[r], coord, &c, TWO_BLANK, first );
		coord[ROW] = r;
		if( buf[0] == '\0' ) break;


		/* see if its a number */
		isnum = Egoodnum( buf, &p );

		/* horizontal rulings (entire token must be -,_,or =) */
		for( i = 0; i < strlen( buf ); i++ ) if( !Emember( buf[i], "-_=" )) break;

		if( strlen( buf ) > 2 && i == strlen( buf )) ruler( buf, coord );

		/* vertical rulings */
		else if( buf[0] == '|' ) vert_ruler( coord, nrows, t );

		/* if phrase is 1st thing on line (within 8 chars), left justify */
		else if( first && coord[LEFT] < 10 ) ljust( buf, coord );

		/* if phrase is in the BODY (not header), alpha and 
		   longer than 2 chars: left justify */
		else if( strlen( buf ) > 2 && !isnum 
			&& r >= bodystart ) /* added scg 2/29/00 */
			/* isalpha( buf[0] ) && r > 10 ) */  ljust( buf, coord );
		
		/* scg 3-7-96 - now handling decimal numbers separately from int ..*/
		/* else if( isnum ) rjust( buf, coord ); */

		/* integer numbers.. right-just */ 
		else if( isnum && p < 0 ) rjust( buf, coord );

		/* decimal numbers - right align front portion, center dec pt, left align back portion*/
		else if( isnum && p >= 0 ) {
			char front[20];
			int cd[4], coordp;
			strcpy( front, buf );
			front[p] = '\0';
			cd[0] = coord[0]; cd[1] = coord[1]; cd[2] = coord[2]; cd[3] = coord[3];
			cd[RIGHT] = coord[RIGHT] - (strlen( buf ) - p );
			rjust( front, cd );
			coordp = coord[LEFT] + p;
			cd[LEFT] = coordp; cd[RIGHT] = coordp+1;
			center( ".", cd );
			cd[LEFT] = coordp+1; cd[RIGHT] = coord[RIGHT];
			ljust( &buf[p+1], cd );
			}


		/* everything else gets centered */
		else center( buf, coord );

		first = 0;
		}
	}
if( vertruleline >=0 ) 
	drawvertrules( t[vertruleline], vertruleline, nrows-1 ); /* -1 becuase there's
									no #pagebreak line */
do_logo();

/* PLOT here */
Eflush();
if( strlen( plotfile ) > 0 ) {
	stat = exec_plfile( plotfile );
	if( stat != 0 ) Eerr( 3882, "Error while executing plotfile", plotfile );
	}

if( Edev == 'x' ) do_x_button( "End." );
else Eprint();
Eendoffile();

if( !Debug ) {
        sprintf( buf, "rm %s* 2> /dev/null", Tmpname );  /* PORT */
        system( buf );
        }
 
if( strlen( Viewer )  > 0 && Device != 'x' ) {
        Egetoutfilename( buf );
        if( strlen( buf ) > 0 ) {
                strcat( Viewer, " " );
                strcat( Viewer, buf );
                fprintf( Diagfp, "Executing '%s' to view results..\n", Viewer );
                system( Viewer );
                }
        }
 
exit(0);

}


/* ====================================== */
getphrase( out, line, coord, c, mode, first )
char out[], 	/* returned phrase */
     line[];	/* input line */
int coord[4];	/* returned location */
int *c;		/* position in line */
int mode;	/* tells if phrases are 
			1) == LINE everything on a line; 
			2) == TWO_BLANK delimited by 2 spaces or 1 tab 
		  */
int first;	/* behaves differently if getting 1st phrase on a line. */
{
int i;

if( mode == LINE ) {	/* working with whole line */
	if( sscanf( line, "%s", out ) < 1 ) { out[0] = '\0'; return( 0 ); } /* blank line */
	for( i = 0; ; i++ ) if( !Emember( line[i], "\t \n" ) ) break; /* skip white space on left */
	for( (*c) = strlen( line ); ; (*c)-- ) if( !Emember( line[i], "\t \n" ) ) break; /* skip white space on right */
	}

else if( mode == TWO_BLANK ) {  
	if( sscanf( &line[ (*c) ], "%s", out ) < 1 ) { out[0] = '\0'; return( 0 ); }
	for( i = (*c); i < strlen( line ) ; i++ ) 
		if( !Emember( line[i], "\t \n" ) ) break; /* skip white space on left */
	/* go until we hit a stopping rule */
	for( (*c) = i; (*c)<strlen( line ) ; (*c)++ ) {
		if( line[ (*c) ] == ' ' && line[ (*c)-1 ] == ' ' ) break;
		else if( first ) continue;
		else if( Emember( line[ (*c) ], delimchars ) && (*c)==i && 
		    line[ (*c) -1 ] != '\\' ) {  /* SCG 5/1/98 added check for preceding \ */
			(*c)++; break; 
			} /* first token-- return it */
		else if( ( Emember( line[ (*c) ], delimchars ) || line[ (*c) ] == '\t')
		        && line[ (*c)-1 ] != '\\' ) 
			/* SCG 5/1/98 added check for preceding \ */
			break; /* not first token, return and get it next time */
		}
	}

strncpy( out, &line[i], (*c)-i );
for( ; ; (*c)-- ) if( ! Emember( out[ ((*c)-i)-1 ], " \t\n" ) ) break;
out[ (*c)-i ] = '\0';

coord[ LEFT ] = i; coord[ CENTER ] = ((*c) + i ) / 2;  coord[ RIGHT ] = (*c);
	
}
/* ====================================== */
center( str, coord )
char str[];
int coord[4];
{
/* fprintf( stderr, "centering %s at %g,%g\n", str, 
 * (double)((coord[LEFT]+coord[RIGHT])/2.0), (double)(EDYhi-coord[ROW]) );
 */
Emovu( (double)((coord[LEFT]+coord[RIGHT])/2.0), (double)EDYhi-coord[ROW] );
Ecentext( str );
}
/* ====================================== */
center_float( str, coord )
char str[];
int coord[4];
{

}
/* ====================================== */
ljust( str, coord )
char str[];
int coord[4];
{
/* fprintf( stderr, "ljust %s at %g,%g\n", str, 
 *  (double)coord[LEFT], (double)(EDYhi-coord[ROW]) );
 */
Emovu( (double)coord[LEFT], (double)(EDYhi-coord[ROW]) );
Etext( str );
}
/* ====================================== */
rjust( str, coord )
char str[];
int coord[4];
{
Emovu( (double)coord[RIGHT], (double)(EDYhi-coord[ROW]) );
Erightjust( str );
}
/* ================================== */
ruler( str, coord )
char str[];
int coord[4];
{ 
if( str[0] == '-' ) Enormline();
else Elinetype( 0, 0.4, 1.0 );
Emovu( (double)coord[LEFT], (double)(EDYhi-coord[ROW]+.5) );
Elinu( (double)coord[RIGHT], (double)(EDYhi-coord[ROW]+.5) );
if( str[0] == '=' ) { /* double line */
	Emovu( (double)coord[LEFT], (double)(EDYhi-coord[ROW]+.3) );
	Elinu( (double)coord[RIGHT], (double)(EDYhi-coord[ROW]+.3) );
	}
Enormline();
}

/* ================================= */
vert_ruler( coord, nrows, t )
int coord[4], nrows;
char t[][MAX_ROW];
{
int i;
Emovu( (double)(coord[LEFT]+.5), (double)((EDYhi-coord[ROW])+1) );
for( i = coord[ROW]+1; i < nrows; i++ ) {
	if( t[i][ coord[LEFT] ] == '|' ) { 
		Elinu( (double)(coord[LEFT]+.5), (double)(EDYhi-i) ); 
		return( 1 ); 
		}
	}
}

/* ======== */
drawvertrules( line, toprow, bottomrow )
char *line;
int toprow, bottomrow; /* actually these seem to be reversed.. */
{
int i;
double Eay();
if( toprow > 0 )toprow--;
Elinetype( 0, 0.4, 1.0 );
for( i = 0; i < strlen( line ); i++ ) {
	if( line[i] == '|' ) { /* single line */
		Emovu( (double)i, (double)(EDYhi-toprow) );
		Elinu( (double)i, (double)(EDYhi-bottomrow) );
		}
	if( line[i] == 'L' ) { /* single line at left edge of table */
		Emovu( -1.0, (double)(EDYhi-toprow) );
		Elinu( -1.0, (double)(EDYhi-bottomrow));
		if( line[i+1] == 'H' ) {
			Emovu( -1.4, (double)(EDYhi-toprow));
			Elinu( -1.4, (double)(EDYhi-bottomrow));
			i++;
			continue;
			}
		}
	if( line[i] == 'H' ) { /* double line */
		Emovu( ((double)i)-0.2, (double)(EDYhi-toprow) );
		Elinu( ((double)i)-0.2, (double)(EDYhi-bottomrow) );
		Emovu( ((double)i)+0.2, (double)(EDYhi-toprow) );
		Elinu( ((double)i)+0.2, (double)(EDYhi-bottomrow) );
		}
	}
Enormline();
return( 0 );
}
/* ================ */
is_sh( c )
char c;
{
if( shchar[0] == '\0' && c != ' ' && c != '\n' && c != '\0' ) return( 1 );
else if( shchar[0] != '\0' && c == shchar[0] ) return( 1 );
else return( 0 );
}

/* ========================= */
/* PLTAB_ROW - given a string, return the row that begins with it.
 	This is called from ploticus to get Y position.
	Returns 0 if matching label was found; 1 if not.

	Remember, when this is executing, scale is set for the plot, NOT the table,
	so the Ea*() scale funtions will operate with respect to the plot area.
*/
pltab_row( s, ypos )
char *s;
double *ypos;
{
int i, j, len, slen, start;
double h;
slen = strlen( s );

if( npages == 1 ) start = plotrowstart;
else start = plotrowstart-1; 

for( i = start; i < plotrowstop; i++ ) {
	len = strlen( t[i] );
	for( j = 0; j < len; j++ ) if( t[i][j] != ' ' ) break;
	if( strncmp( s, &t[i][j], slen )==0 ) { 
		*ypos = (double)(plotrowstop - i) - 0.63 + catslide; /* adjust for exact middle of row */
		/* fprintf( stderr, "[%s FOUND, row %d, returning %g]", s, i, *ypos );  */
		return( 0 );
		}
	}
if( Debug ) fprintf( Diagfp, "stub %s not found\n", s ); 
return( 1 );
}

/* ======================== */
/* PLTAB_PLOTAREA - report the boundary of the pltab plotting area, and the 
   number of rows.  Areadef should use these to set up the plotting area
   when yscaletype is pltab (yrange from 0 to ymax). */

pltab_plotarea( x1, y1, x2, y2, ymax )
double *x1, *y1, *x2, *y2, *ymax;
{
if( plotrowstop == 0 ) return( 1 ); /* plotting area not defined */
*ymax = (double)(plotrowstop - plotrowstart);
*x1 = plotx1;
*y1 = ploty1;
*x2 = plotx2;
*y2 = ploty2;
if( Debug ) fprintf( Diagfp, "pltab plotarea is: %g %g   %g %g  ymax=%g\n", 
	*x1, *y1, *x2, *y2, *ymax ); 
return( 0 );
}

/* ========================= */
/* PLTAB_SETCATSLIDE - adjust the category location */
pltab_setcatslide( amount )
double amount;
{
catslide = amount;
return( 0 );
}

/* ========================= */
/* PLTAB_NEXTSTUB - get next stub.  Stub is truncated to maxlen.
   When there are no more, a zero length stub is returned and function returns 1.
   Otherwise, function returns 0.
*/
pltab_nextstub( stub, maxlen )
char *stub;
int maxlen;
{
int i, j;
int tlen;

strcpy( stub, "" );
while( 1 ) {
	i = plotrowstart + curstub;
	if( i >= plotrowstop || i > nrows ) {
		strcpy( stub, "" );
		return( 1 );
		}
	tlen = strlen( t[i] );
	for( j = 0; j < tlen; j++ ) if( t[i][j] != ' ' && t[i][j] != '\n' ) break;
	if( j == tlen ) { /* blank line */
		curstub++;
		continue;
		}
	if( t[i][j] == '#' ) { /* # operator.. skip   scg 6/28/99 */
		curstub++;
		continue;
		}
	
	strncpy( stub, &t[i][j], maxlen );
	stub[maxlen] = '\0';
	curstub++;
	return( 0 );
	}
return( 1 );
}
/* =============== */
pltab_in_effect()
{
return( 1 );
}

