/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include "pl.x"

do_preliminaries()
{
char buf[512];
FILE *fp;
char *filename, *getenv();
char tag[80];
char value[512];
int valused, found;
int nt, stat;


/* set first part of name for for tmp files.. */
make_unique_string( buf, 0 );
sprintf( Tmpname, "/usr/tmp/plo%s", buf ); 

Errfp = stderr; 	/* portability? */
Diagfp = stderr;	/* portability? */

Png = 0;

Bignumspacer = '\0'; 	/* use standard number notation */


suppress_convmsg( 1 ); /* suppress unplottable data msgs by default - user can turn on */
checkdatelengths( 0 ); /* don't be strict about the length of date items */


/* the remaining code reads the config file.. */

filename = getenv( "PLOTICUS_CONFIG" );
if( filename == NULL ) return( 0 );
if( strlen( filename ) < 1 ) return( 0 );

fp = fopen( filename, "r" );
if( fp == NULL ) {
	err( 15060, 0, "Cannot open ploticus config file", filename );
	return( 0 );
	}

/* get user settings.. */
while( fgets( buf, 511, fp ) != NULL ) {
	buf[ strlen( buf ) -1 ] = '\0';
	nt = sscanf( buf, "%s %s", tag, value );
	if( nt < 1 ) continue; /* skip blank lines */
	if( Emember( tag[0], "#/" )) continue; /* skip comments of various kinds */
	if( nt < 2 ) strcpy( value, "" );

	if( slmember( tag, "dateformat: pivotyear: months* weekdays:" )) {
		stat = setdateparms( tag, &buf[strlen(tag)+1] );
		if( stat != 0 ) return( Eerr( stat, "date parm error in config file", buf ) );
		}

	else if( slmember( tag, "units:" )) {
		if( tolower( value[0] ) == 'c' ) Using_cm = 1;
		else Using_cm = 0;
		}

	else if( slmember( tag, "tmpdir:" )) {
		make_unique_string( buf, 0 );
		sprintf( Tmpname, "%s/plo%s", value, buf ); 
		}

	else if( slmember( tag, "option:" )) {
		value[0] = '\0';
		sscanf( buf, "%*s %s %s", tag, value );
		process_arg( tag, value, &valused, &found );	
		if( !found ) Eerr( 2784, "invalid option in config file", tag );
		}

	else if( slmember( tag, "numbernotation:" )) {
		sscanf( buf, "%*s %s", value );
		if( stricmp( value, "us" )==0 ) Bignumspacer = ',';
		else if( stricmp( value, "euro" )==0 ) Bignumspacer = '.';
		else Bignumspacer = '\0';
		}

	else 	{ 
		err( 2749, 0, "Unrecognized tag in config file", tag );
		continue;
		}
	}
return( 0 );
}
