/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* ================== */
/* PROCESS_ARG - process one command line argument from either pl or pltab */

#include "pl.x"

process_arg( opt_in, val, valused, found )
char *opt_in; /* option */
char *val; /* argument that follows option */
int *valused; /* 1 if val was used, 0 if not */
int *found; /* 1 if option found here, 0 if not */
{
int i, j;
int vardec;
char buf[256];
int nt;
char opt[256];

*found = 1;
if( opt_in[0] == '-' ) strcpy( opt, &opt_in[1] );
else strcpy( opt, opt_in );

/* add to list of parameters set on command line.. */
strcat( Cmdlineparms, opt );
strcat( Cmdlineparms, " " );

vardec = 0;
for( j = 0; j < strlen( opt ); j++ ) {
	if( opt[j] == '=' ) {
		vardec = 1;
		break;
		}
	}

*valused = 1;

if( slmember( opt, "x11 gif png ps bwps eps" )) { devnamemap( &Device, opt, 1 ); *valused = 0; }
else if( strcmp( opt, "debug" )==0) { Debug = 1; *valused = 0; }
else if( strcmp( opt, "tightcrop" )==0 ) { Etightbb(1); *valused = 0; }
else if( strcmp( opt, "viewer" )==0 || strcmp( opt, "v" )==0 ) strcpy( Viewer, val ); 
else if( strncmp( opt, "crop", 4 )==0 ) {
	double cropx1, cropy1, cropx2, cropy2;
	nt = sscanf( val, "%lf,%lf,%lf,%lf", &cropx1, &cropy1, &cropx2, &cropy2 );
	if( nt != 4 ) 
		fprintf( Errfp, "pl argument error, correct usage is: -crop x1,y1,x2,y2 OR -croprel left,bottom,right,top\n" );
	else {
		if( Using_cm ) { cropx1 /= 2.54; cropy1 /= 2.54; cropx2 /= 2.54; cropy2 /= 2.54; }
		if( strcmp( opt, "croprel" )==0 ) Especifycrop( 2, cropx1, cropy1, cropx2, cropy2 ); /* relative to tight */
		else Especifycrop( 1, cropx1, cropy1, cropx2, cropy2 ); /* absolute */
		}
	}
else if( strcmp( opt, "pagesize" )==0 ) {
	nt =  sscanf( val, "%lf,%lf", &Winw, &Winh );
	if( nt != 2 ) 
		fprintf( Errfp, "pl argument error, correct usage is: -pagesize width,height (inches or cm)\n" );
	else Winsizegiven = 1;
	if( Using_cm ) { Winw /= 2.54; Winh /= 2.54; }
	}
else if( strcmp( opt, "winloc" )==0 ) {
	nt = sscanf( val, "%d,%d", &Winx, &Winy );
	if( nt != 2 ) {
		Winx = 100; Winy = 0;
		fprintf( Errfp, "pl argument error, correct usage is: -winloc x,y\n" );
		}
	}

else if( strcmp( opt, "o" )==0 ) {
	strcpy( Cmdline_outfile, val );
	}

else if( strcmp( opt, "scale" )==0) {
	double sx, sy;
	nt =  sscanf( val, "%lf,%lf", &sx, &sy );
	if( nt == 1 ) Esetglobalscale( sx, sx );
	else if( nt == 2 ) Esetglobalscale( sx, sy );
	}

else if( strcmp( opt, "posteroffset" )==0) {
	double x, y;
	nt = sscanf( val, "%lf,%lf", &x, &y );
	if( nt != 2 ) fprintf( Errfp, "pl argument error, correct usage is -posteroffset x,y\n" );
	else	{
		if( Using_cm ) { x /= 2.54; y /= 2.54; }
		Esetposterofs( x, y );
		}
	}

else if( strcmp( opt, "landscape" )==0 ) { Landscape = 1; *valused = 0; }
else if( strcmp( opt, "portrait" )==0 ) { Landscape = 0; *valused = 0; Winw = 8.5; Winh = 11.0; } 
								/* winw and winh added scg 11/14/00 */

else if( strcmp( opt, "showbad" )==0 ) { suppress_convmsg( 0 ); *valused = 0; }

else if( strcmp( opt, "font" )==0) strcpy( Estandard_font, val );
		
else if( strcmp( opt, "textsize" )==0) Estandard_textsize = atoi( val );

else if( strcmp( opt, "linewidth" )==0) Estandard_lwscale = atof( val );

else if( strcmp( opt, "color" )==0) strcpy( Estandard_color, val );
else if( strcmp( opt, "cm" )==0) { Using_cm = 1; *valused = 0; }
else if( strcmp( opt, "inches" )==0) { Using_cm = 0; *valused = 0; }

else if( strcmp( opt, "backcolor" )==0) {
	strcpy( Estandard_bkcolor, val );
	Ebackcolor( val );
	Bkcolorgiven = 1;
	}
else if( strcmp( opt, "id" )==0 ) {
	strcat( Flags, "I" );
	*valused = 0;
	}
else if( strcmp( opt, "noid" )==0 ) {
	strcat( Flags, "N" );
	*valused = 0;
	}

else if( strcmp( opt, "?" )==0 || 
	strcmp( opt, "help" )==0  ||
	strcmp( opt, "ver" )==0 ||
	strcmp( opt, "version" )==0 ) {
	vermessage();
	exit(0);
	}

else if( strcmp( opt, "diagfile" )==0 ) {
	if( smember( val, "stderr STDERR" )) Diagfp = stderr; /* portability  */
	else if( smember( val, "stdout STDOUT" )) Diagfp = stdout; /* portability  */
	else	{
                Diagfp = fopen( val, "w" );
                if( Diagfp == NULL ) {
                        fprintf( Errfp, "warning, cannot open -diagfile %s, using stderr\n", val );
                        Diagfp = stderr;
                        }
                }
	}

else if( strcmp( opt, "errfile" )==0 ) {
	if( smember( val, "stderr STDERR" )) Errfp = stderr; /* portability  */
	else if( smember( val, "stdout STDOUT" )) Errfp = stdout; /* portability  */
	else	{
                Errfp = fopen( val, "w" );
                if( Errfp == NULL ) {
                        fprintf( stderr, "warning, cannot open -errfile %s, using stderr\n", val );
                        Errfp = stderr;
                        }
                }
	}
 
else if( vardec ) {	/* its a VAR=VALUE pair */
	int ix;
	char varname[40];
       	strcpy( buf, opt );
       	ix = 0;
        Egetseg( varname, buf, &ix, "=" );
        declarevar( varname );
        setvar( varname, &buf[ix] );
	*valused = 0;
        }
else 	{ *found = 0; *valused = 0; }
return( 0 );
}
