/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include <stdio.h>

static int getdecplaces();
static int encode();

/* =========================== */
char *
getok( s, i )
char *s;
int *i;
{
char *Egetok();
return( Egetok( s, i ) );
}
/* ========================== */
/* FUZZYMATCH - Do a 'fuzzy' match.  See if s1 is similar to s2.

   This routine may alter s1 and s2.

   A wildcard character '*' at the beginning or end of s2 indicates
   that a match can occur anywhere within s1 (eg. smith* would match
   smithington and *smith would match harrowsmith; *smith* would match
   both).  If there are no asterisks, lengths
   of the two strings must be similar.

   Degree changes the 'looseness' of the match.  
   5 = strict, 4 = medium-strict, 3 = medium 2= loose, 1 = very loose.
   
   Returns 1 on a match, 0 not. 

   SCG '94,'95
 */

fuzzymatch( s1, s2, len2, degree )
char *s1; /* data value */
char *s2; /* query value */
int len2; /* length of s2 */
int degree; /* dgree of tightness */
{
int i, j, k, len1, pts;
int bestpts, score;
char c;
int openfront, openend, goal;

/* exact match.. no need to do more.. */
if( strcmp( s2, s1 ) ==0 ) return( 1 );

len1 = strlen( s1 );

/* see what type of match, open or closed */
if( s2[0] == '*' ) { openfront = 1; s2 = &s2[1]; len2--; }
else openfront = 0;
if( s2[ len2-1 ] == '*' ) { openend = 1; len2--; }
else openend = 0;

/* set goal */
if( len2 < 4 )goal = len2 * degree;
else if( len2 < 10 ) goal = (len2 -2) * degree;
else goal = (len2 -3) * degree;


/* reject disparate lengths if doing a closed match */
if( !openfront && !openend && len1 - len2 > 3 ) return( 0 );

/* truncate end of s1 if openended match */
if( openend && !openfront && len1 > len2+2 ) len1 = len2+2;

/* chop of front of s1 if openfronted match */
if( openfront && !openend && len1 > len2+2 ) s1 = &s1[ len1-(len2+2) ];

/* printf( "[%d-%d]", openfront, openend ); */

/* go thru s2.. */
score = 0;
for( i = 0; i < len2; i++ ) {
	c = s2[i];

	/* find each occurrence of c in s1.. */
	bestpts = 0;
	pts = 0;
	for( j = 0; j < len1; j++ ) {

		if( s1[j] == c ) {

			/* check adjacent characters, count matches.. */
			pts = 1;

			if( i + j == 0 ) pts+=2; /* both 1st char */
			
			else /* go left.. */
			  for( k = 1; k <= 4; k++ ) {
				if( j-k < 0 || i-k < 0 ) break;
				else if( s1[j-k] == s2[i-k] ) pts++;
				else break;
				}
			if( i == len2-1 && j == len1-1 )
				pts+=2; /* both last char */

			else /* go right.. */
			  for( k = 1; k <= 4; k++ ) {
				if( j+k > len1-1 || i+k > len2-1 ) break;
				else if( s1[j+k] == s2[i+k] ) pts++;
				else break;
				}
			}
		if( pts > bestpts )bestpts = pts;
		}
	if( bestpts > 6 ) score += 6;
	else score += bestpts;
	/* printf( "[%c:%d]", c, bestpts ); */
	if( score > goal ) return( 1 ); /* optimize */
	}
/* printf( "<%s v. %s : %d>\n", s2, s1, score ); */
if( score > goal ) return( 1 );
else return( 0 );
}
/* ====================== */
/* smember() - returns 1 if string s is present as a token in string t,
		0 otherwise.  Auth: Steve Grubb */
smember( s, t )
char s[], t[];
{
char tok[100], *getok();
int i;
i = 0;
while( 1 ) {
	strcpy( tok, getok( t, &i ) );
	if( tok[0] == '\0' ) break;
	if( strcmp( tok, s ) == 0 ) return( 1 );
	}
return( 0 );
}


/* ====================== */
/* SUBSTRING - another substring routine.  This one has the same functionality
	as the substr shell utility in ../cmisc.

	substring( result, str, fromchar, nchar )
	char *result;   // substring is copied into this variable
	char *str;	// the original string
	int fromchar;   // starting point from which to take the substring
	int nchar;	// length of the substring

	In all cases the first char is 1, not 0.

	Two ways it can operate:
	   If <fromchar> is greater than 0, the result will be the portion of <string>
		beginning at position <fromchar>, for a length of <nchar>, or until
		the end is reached.
	   If <fromchar> is less than 0, and <nchar> is greater than 0:
		we will begin counting from the end of <string>, 
		leftward.  for (abs)<fromchar> characters.  Then, we will take the 
		substring beginning from that character 
		for a length of <nchar>, or until the end is reached.

	
	Examples: substring( result, "02001.fv02", 7, 4 )    -- result would be "fv02"
	          substring( result, "02001.fv02", -4, 99 )   -- result would be "fv02"

	History
	----------------
	2-11-97 scg created from ../cmisc/substr.c
*/
#include <stdio.h>

substring( result, str, fromchar, nchar )
char *result;
char *str;
int fromchar;
int nchar;
{
int i, j;
int len;

len = strlen( str );
if( fromchar > 0 ) fromchar -= 1;
else if( fromchar < 0 ) fromchar += len;

for( i = fromchar, j = 0; i < fromchar + nchar; i++ ) { 
	if( i > len-1 ) break;
	result[j++] = str[i];
	}
result[j] = '\0';
return( 0 );
}

#ifdef TESTING
main()
{
char result[50];
substring( result, "02001.fv02", 7, 4 ); /*   -- result should be "fv02" */
printf( "%s\n", result );
substring( result, "02001.fv02", -4, 99 ); /*   -- result should be "fv02" */
printf( "%s\n", result );
} 
#endif


/* =============================================== */
/* SYSTIME - get current time */
#ifndef DOS
systime( hour, min, sec )
int	*hour, *min, *sec ;
{
	int	tvec[2], *dtime ;

	time( tvec );
	dtime = (int *)localtime( tvec );
	*hour = *(dtime+2) ;
	*min = *(dtime+1)  ;
	*sec = *(dtime)  ;
}
#endif
/* =============================================== */
/* SLMEMBER - Return 1 if str is matches any items in list.
	List is a space-delimited list of tokens.  The tokens in the
	list may contain ? or * wildcard characters.  The match is
	NOT case-sensitive.

	scg 3-19-97
*/
slmember( str, list )
char *str;
char *list;
{
char tok[100], *getok();
int i;
i = 0;
while( 1 ) {
        strcpy( tok, getok( list, &i ) );
        if( tok[0] == '\0' ) break;
        if( wildcmp( str, tok, strlen(tok), 0 ) == 0 ) return( 1 );
        }
return( 0 );
}

/* =============================================== */
/* WILDCMP - compare two strings s1 and s2.  S2 may contain 
   wildcards (* and ?).

   Function returns 0 on a match;  < 0 if s1 < s2;   > 0 if s1 > s2
   Prints an error message and returns -999 on error.

   * wildcard limited to the following uses: *ppp; ppp*; pp*pp; *ppp*
   ? can be used anywhere.

   Double asterisks at beginning and end are also handled (means the
   same as single asterisk).

   scg 3-4-96 (written elsewhere)

 */

#include <stdio.h> 
#include <ctype.h>

wildcmp( s1, s2, len, casecare )
char *s1, *s2;
int len, casecare;

/* s1  = data value
   s2  = query value which can contain wildcards - not null terminated.
   len = length of s2
   casecare = 0 for case-insensitive, 1 for case-sensitive
 */
{
int i, nwc, wcp, stat, stat2;


if( len == 0 ) return( strlen( s1 ) );
else if( s2[0] == '*' ) {
	if( len == 1 ) return( 0 ); /* everything matches */
	}
else if( s2[0] == '?' ) ; /* can't tell yet */
else if( tolower( s1[0] ) < tolower( s2[0] ) ) return( -1 ); /* way off */
else if( tolower( s1[0] ) > tolower( s2[0] ) ) return( 1 );  /* way off */

/* strip off extraneous * at beginning and end.. */
if( s2[0] == '*' && s2[1] == '*' ) { s2 = &s2[1]; len--; }
if( s2[len-1] == '*' && s2[len-2] == '*' ) len--;

/* see if any wild cards were used.. */
nwc = 0;
for( i = 0; i < len; i++ ) if( s2[i] == '*' ) { nwc++; wcp = i; }

if( nwc < 1 ) {  /* straight match */
	if( strlen( s1 ) > len ) return( wcmp( s1, s2, strlen( s1 ), casecare));
	else return( wcmp( s1, s2, len, casecare ) ); 
	}

else if( nwc == 1 ) {                /* wildcard match */
	/* find beginning of what we need to compare */
	i = strlen( s1 ) - (len - (wcp+1) );

	/* case 1: wc at end.. */
	if( wcp == len-1 ) {
		return( wcmp( s1, s2, len-1, casecare ) );
		}

	/* case 2: wc at beginning.. */
	if( wcp == 0 ) {
		return( wcmp( &s1[i], &s2[ 1 ], len-1, casecare ) );
		}

	/* case 3: wc in middle.. */
	else	{
		int frontlen, backlen;

		frontlen = wcp;

		/* do front compare.. */
		stat = wcmp( s1, s2, frontlen, casecare ); 
		if( stat != 0 ) return( stat );

		backlen = strlen( s2 ) - (frontlen + 1);
		if( strlen( s1 )  < frontlen + backlen ) return( 1 );  /* fail if s1 too short */

		/* do back compare.. */
		stat = wcmp( &s1[ strlen( s1 ) - backlen ], &s2[ strlen( s2 ) - backlen ], backlen, casecare );
		return( stat );

		}
	}

else if( nwc == 2 ) {
	int stop;
	/* case 4: wc at beginning and end.. */
	if( wcp != (len-1) ) goto ERR;
	else if( s2[0] != '*' ) goto ERR;
	stop = ( strlen( s1 ) - len ) + 2;
	for( i = 0; i <= stop; i++ ) {
		if( wcmp( &s1[i], &s2[1], len-2, casecare ) == 0 ) return( 0 );
		}

	return( -1 );
	}
else 	{
	ERR:
#ifndef PLOTICUS
	fprintf( stderr, "Wild card match error (%s vs %s).\n", s1, s2 );
#endif
	return( -999 );
	}
}

/* ============================= */
/* WCMP - compare two strings.  S2 may contain ? wildcards which matches any
       single character.  Len is the # of characters to check.
 */

wcmp( s1, s2, len, casecare )
char *s1, *s2;
int len, casecare;
{
int i;
for( i = 0; i < len; i++ ) {
	if( ! casecare ) {
		if( tolower(s1[i]) < tolower(s2[i]) && s2[i] != '?' ) 
			return( -1 );
		else if( tolower(s1[i]) > tolower(s2[i]) && s2[i] != '?' ) 
			return( 1 );
		}
	else	{
		if( s1[i] < s2[i] && s2[i] != '?' ) return( -1 );
		else if( s1[i] > s2[i] && s2[i] != '?' ) return( 1 );
		}
	}
return( 0 );
}
/* ============================= */

/* ------------------------------- */
/* AUTOROUND - round a value reasonably based on its magnitude.
   val is the value, represented as a string.
   decoffset controls the precision of the rounded result as follows:
	If decoffset = 0 then nothing happens.  
	If decoffset = 1 then rounding will go to 1 additional decimal place.  
	decoffset = -1 then rounding will go to one less decimal place than normal.

   The rounded result is copied back into val.

   If val is non-numeric or a whole number then it is left unchanged.
*/
autoround( val, decoffset )
char *val;
int decoffset; 
{
int precision, decplaces, stat;
char roundingfmt[50];
double g, atof();

stat = Egoodnum( val, &precision );
if( stat && precision > 0 ) {
	g = atof( val );
	decplaces = getdecplaces( g );
	if( decplaces > -99 ) {
		if( decplaces < 0 ) decplaces = 0;
		sprintf( roundingfmt, "%%.%df", decplaces + decoffset );
		sprintf( val, roundingfmt, g );
		}
	}
return( 0 );
}
/* --------------------- */
/* AUTOROUNDF - variant of autoround(), takes val as a double, return value is character rep..*/
char *
autoroundf( val, decoffset )
double val;
int decoffset;
{
int precision, decplaces, stat;
char roundingfmt[50];
static char result[50];

sprintf( result, "%g", val ); /* fallback */
decplaces = getdecplaces( val );
if( decplaces > -99 ) {
	if( decplaces < 0 ) decplaces = 0;
	sprintf( roundingfmt, "%%.%df", decplaces + decoffset );
	sprintf( result, roundingfmt, val );
	}
return( result );
}

/* -------------- */
static int getdecplaces( val )
double val;
{
int decplaces;
double g, fabs();

g = fabs( val );
decplaces = -99;
if( g >= 1000 ) decplaces = 0; 
else if( g >= 100 ) decplaces = 0; 
else if( g >= 10 ) decplaces = 1;
else if( g >= 1.0 ) decplaces = 2;
else if( g >= 0.1 ) decplaces = 3;
else if( g >= 0.01 ) decplaces = 4;
else if( g >= 0.001 ) decplaces = 5;
else if( g >= 0.0001 ) decplaces = 6;

return( decplaces );
}

/* ----------------- */
/* MAKE_UNIQUE_STRING - generate an identifier using date, time, and pid */

make_unique_string( s, i )
char *s;
int i;  /* may be sent as an integer.. if 0 getpid() will be used.. */
{
int mon, day, yr, hr, min, sec, pid, a, b, c;
sysdate( &mon, &day, &yr );
systime( &hr, &min, &sec );
s[0] = encode( yr % 100 );
s[1] = encode( mon );
s[2] = encode( day );
s[3] = encode( hr );
s[4] = encode( min );
s[5] = encode( sec );
if( i == 0 ) pid = getpid();
else pid = i;
s[6] = encode( pid % 62 );
pid = pid / 62;
s[7] = encode( pid % 62 );
s[8] = encode( pid / 62 );
s[9] = '\0';

return( 0 );
}

/* encode - derive a character representation of a number */
static int encode( a )
int a;
{
if( a >= 0 && a <= 9 ) return( a + '0' );
else if( a > 35 ) return( (a-36) + 'A' ); /* A-Z    26 letters + 9 = 35 */
else if( a > 9 ) return( (a-10) + 'a' ); /* a-z */
else return( '0' );
}
