/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* Set up a graphics area with scaling */
#include "pl.x"

static double xofst[38] = { 0, 0, 4, 0, -4, 4, -4, -4, 4,
        0, -8, 0, 8, 4, -8, 4, 8, -4, -8, -4, 8,
        0, 0, 12, -12, 4, 4, 12, -12, -4, -4, 12, -12,
        8, -8, -8, 8 };

static double yofst[38] = { 0, 4, 0, -4, 0, 4, -4, 4, -4,
        -8, 0, 8, 0, -8, 4, 8, 4, -8, -4, 8, -4,
        12, -12, 0, 0, 12, -12, 4, 4, 12, -12, -4, -4,
        8, -8, 8, -8 };

static double dofst[17] = { 0, 2, -2, 4, -4, 6, -6, 8, -8, 10, -10, 12, -12, 14, -14, 16, -16 };

#ifdef NONANSI
static int ptcompare();
#else
static int ptcompare(double *f, double *g);
#endif

scatterplot( specfile, clonefiles, nclonefiles  )
char *specfile, clonefiles[][ MAXPATH ];
int nclonefiles;
{
int i;
FILE *fp;
char buf[256];
char attr[40], val[256];
char lineval[256];
int nt;
FILE *openclonefile();
int icf;

int stat;
int align;
double adjx, adjy;

int xfield, yfield;
char symbol[256];
double linelen;
char linedetails[256];
double xloc, yloc;
int cluster;
double radius;
char symcode[80];
double x, y;
char text[256];
double cx, cy;
double hlinelen;
char textdetails[256];
int lblfield;
char selex[256];
int result;
double sizescale;
int sizefield;
double ox[38], oy[38];
double od[17];
double clusterfact;
double oldx, oldy;
int dupcount;
int subdupcount;
int clustevery;
int verttext;
int irow;
char legendlabel[120];
char xrange[80], yrange[80];
double xlo, xhi, ylo, yhi;
char rhi[40], rlo[40];
double clusterdiff;
int realrow;

strcpy( Eprogname, "pl proc scatterplot" );


/* -------------------------- */
/* DEFAULTS */
/* All variables must start out with default values, */
/* since no variable is guaranteed to be set */
/* -------------------------- */
xfield = -1; yfield = -1;
strcpy( symbol, "" );
linelen = -1.0;
strcpy( linedetails, "" );
xloc = 0.0;
yloc = 0.0;
cluster = 1;
strcpy( text, "" );
strcpy( textdetails, "" );
lblfield = -1;
strcpy( selex, "" );
sizefield = -1;
sizescale = 0.5/72.0; /* correspond roughly with pt size */
clusterfact = 0.01;
verttext = 0;
strcpy( legendlabel, "" );
strcpy( xrange, "" );
strcpy( yrange, "" );
clustevery = 0;
clusterdiff = 0.001;

/* -------------------------- */
/* open and process all clonefiles and lastly, specfile */
/* -------------------------- */
for( icf = 0; icf <= nclonefiles; icf++ ) {
 
    if( icf == nclonefiles ) {
        fp = fopen( specfile, "r" );
        if( fp == NULL ) return( Eerr( 117, "Could not open to read", specfile ) );
        }
    else        {
        fp = openclonefile( clonefiles[icf] );
        if( fp == NULL ) continue;
        }

    while( fgets( buf, 255, fp ) != NULL ) {
	nt = sscanf( buf, "%s %s", attr, val );		/* val has 2nd token */
	if( nt < 1 ) continue;
	if( nt == 1 ) strcpy( val, "" );
	if( icf < nclonefiles && stricmp( attr, "#proc" )==0 ) continue;
	i = 0;
        Egetok( buf, &i );
	while( buf[i] == ' ' || buf[i] == '\t' ) i++; /* skip over white space */
        strcpy( lineval, &buf[ i ] ); /* lineval has remainder of line, for sscanf-ing */
        lineval[ strlen( lineval ) - 1 ] = '\0'; /* strip off newline */
 
	/* strip off trailing colon if any.. */
	if( attr[ strlen( attr ) - 1 ] == ':' ) attr[ strlen( attr ) - 1 ] = '\0';


	/* attributes.. */

	if( stricmp( attr, "xfield" )==0 ) xfield = fref( val ) -1;
		
	else if( stricmp( attr, "yfield" )==0 ) yfield = fref( val ) -1;
	else if( stricmp( attr, "labelfield" )==0 ) lblfield = fref( val ) -1;
		
	else if( stricmp( attr, "symbol" )==0 ) strcpy( symbol, lineval );
	else if( stricmp( attr, "text" )==0 ) strcpy( text, val );
	else if( stricmp( attr, "textdetails" )==0 ) strcpy( textdetails, lineval );

	else if( stricmp( attr, "sizefield" )==0 ) sizefield = fref( val ) -1;
	else if( stricmp( attr, "sizescale" )==0 ) sizescale = atof( val ) * 0.5/72.0;
	else if( stricmp( attr, "xrange" )==0 ) strcpy( xrange, lineval );
	else if( stricmp( attr, "yrange" )==0 ) strcpy( yrange, lineval );
	else if( stricmp( attr, "clusterdiff" )==0 ) clusterdiff = atof( val );
	else if( stricmp( attr, "linelen" )==0 ) {
		if( strlen( val ) < 1 ) linelen = -1.0;
		else	{
			linelen = atof( val );
			if( Using_cm ) linelen /= 2.54;
			}
		}
	else if( stricmp( attr, "linedetails" )==0 ) strcpy( linedetails, lineval );
		
	else if( stricmp( attr, "xlocation" )==0 ) {
		Eposex( lineval, X, &xloc ); /* val -> lineval scg 5/3/99 */
		if( Econv_error() ) Eerr( 2394, "invalid xlocation", val );
		}
		
	else if( stricmp( attr, "ylocation" )==0 ) {
		Eposex( lineval, Y, &yloc ); /* val -> lineval 5/3/99 */
		if( Econv_error() ) Eerr( 2395, "invalid ylocation", val );
		}
		
	else if( stricmp( attr, "select" )==0 ) strcpy( selex, lineval );
	else if( stricmp( attr, "legendlabel" )==0 ) strcpy( legendlabel, lineval );

	else if( stricmp( attr, "cluster" )==0 ) {
		if( strnicmp( val, YESANS, 1 )==0 ) cluster = 1;
		else cluster = 0;
		}
	else if( stricmp( attr, "clusterfact" )==0 ) (clusterfact = atof( val ) * .01);
	else if( stricmp( attr, "clustevery" )==0 ) {
		clustevery = atoi( val );
		if( clustevery < 1 ) clustevery = 1;
		}
	else if( stricmp( attr, "verticaltext" )==0 ) {
		if( strnicmp( val, YESANS, 1 )==0 ) verttext = 1;
		else verttext = 0;
		}
	

	else Eerr( 1, "attribute not recognized", attr );
	}
    fclose( fp );
    }

/* -------------------------- */
/* overrides and degenerate cases */
/* -------------------------- */

if( Nrecords[Dsel] < 1 ) return( Eerr( 17, "No data has been read yet w/ proc getdata", "" ) );
if( !scalebeenset() )
         return( Eerr( 51, "No scaled plotting area has been defined yet w/ proc areadef", "" ) );

if( xfield < 0 && yfield < 0 ) return( Eerr( 2205, "Niether xfield nor yfield defined", "" ));

if( lblfield >= 0 ) cluster = 0;  /* added scg 12/21/00 */

/* -------------------------- */
/* now do the plotting work.. */
/* -------------------------- */

if( cluster ) {
	/* make offsets */
	for( i = 0; i < 38; i++ ) {
		ox[i] = xofst[i] * clusterfact;
		oy[i] = yofst[i] * clusterfact;
		}
	for( i = 0; i < 17; i++ ) od[i] = dofst[i] * clusterfact;
	}

/* ranges */
xlo = EDXlo;
xhi = EDXhi;
ylo = EDYlo;
yhi = EDYhi;
if( strlen( xrange ) > 0 ) {
	nt = sscanf( xrange, "%s %s", rlo, rhi );
	xlo = Econv( X, rlo );
	if( Econv_error() ) { Eerr( 3958, "xrange bad format", rlo ); xlo = EDXlo; }
	if( nt == 2 ) xhi = Econv( X, rhi );
	if( Econv_error() ) { Eerr( 3958, "xrange bad format", rhi ); xhi = EDXhi; }
	}
if( strlen( yrange ) > 0 ) {
	nt = sscanf( yrange, "%s %s", rlo, rhi );
	ylo = Econv( Y, rlo );
	if( Econv_error() ) { Eerr( 3958, "yrange bad format", rlo ); ylo = EDYlo; }
	if( nt == 2 ) yhi = Econv( Y, rhi );
	if( Econv_error() ) { Eerr( 3958, "yrange bad format", rhi ); yhi = EDYhi; }
	}





irow = 0;
for( i = 0; i < Nrecords[Dsel]; i++ ) {

	if( strlen( selex ) > 0 ) { /* process against selection condition if any.. */
                stat = do_select( selex, i, &result );
                if( stat != 0 ) { Eerr( stat, "Select error", selex ); continue; }
                if( result == 0 ) continue; /* reject */
                }

	/* get x value.. */
	if( xfield >= 0 ) {
		x = fda( i, xfield, 'x' );
        	if( Econv_error() ) { conv_msg( i, xfield, "xfield" ); continue; }
		if( x < xlo || x > xhi ) continue;
		}

	/* get y value.. */
	if( yfield >= 0 ) {
		y = fda( i, yfield, 'y' );
        	if( Econv_error() ) { conv_msg( i, yfield, "yfield" ); continue; }
		if( y < ylo || y > yhi ) continue;
		}

	/* go to absolute units.. */
	if( xfield < 0 ) x = xloc;
	else x = Eax(x);
	if( yfield < 0 ) y = yloc;
	else y = Eay(y);


	/* put (x,y) into Dat array so points can be sorted.. */
	if( irow >= THIRDMAXDAT ) {
		fprintf( Errfp, "scatterplot: sorry, holding capacity exceeded, skipping data point\n" );
		continue;
		}
	dat3d( irow, 0 ) = x;
	dat3d( irow, 1 ) = y;
	dat3d( irow, 2 ) = (double)i;  /* added scg 12/21/00 - went from dat2d to dat3d */
				       /* need to keep track of actual location in data array for labels, sizefield, etc.. */
	irow++;
	}


/* if clustering and not using a label field, sort Dat array */
if( cluster && lblfield < 0 && sizefield < 0 ) {
	if( Debug ) fprintf( Diagfp, "sorting points for scatterplot\n" );
	qsort( Dat, irow, sizeof(double)*3, ptcompare );
	}


if( verttext ) Etextdir( 90 );

/* these are used in clustering.. */
oldx = NEGHUGE;
oldy = NEGHUGE;
dupcount = 0;
subdupcount = 0;

strcpy( symcode, "sym6a" );
radius = 0.04;

/* in the following, text must come before symbol.. */
if( strlen( text ) > 0 || lblfield >= 0 ) {
	textdet( "textdetails", textdetails, &align, &adjx, &adjy, -3, "R" );
	}
if( strlen( symbol ) > 0 ) {
	symdet( "symbol", symbol, symcode, &radius );
	}
if( linelen > 0.0 ) {
	linedet( "linedetails", linedetails, 0.5 );
	}
cx = Ecurtextwidth * 0.3;
cy = Ecurtextheight * 0.3;
hlinelen = linelen * 0.5;


/* now display points.. */
for( i = 0; i < irow; i++ ) {
	x = dat3d( i, 0 );
	y = dat3d( i, 1 );
	realrow = (int)dat3d( i, 2 ); /* added scg 12/21/00 */

	if( cluster ) {
		if( close_to( x, oldx, clusterdiff ) && close_to( y, oldy, clusterdiff ) ) {
			subdupcount++;
			if( subdupcount >= clustevery ) {
				dupcount++;
				subdupcount = 0;
				}
			if( dupcount > 37 ) dupcount = 0;
			if( yfield < 0 ) x += od[dupcount%10]; /* 1D */
			else if( xfield < 0 ) y += od[dupcount%10]; /* 1D */
			else 	{
				if( xfield >= 0 ) x += ox[dupcount];
				if( yfield >= 0 ) y += oy[dupcount];
				}
			}
		else {
			oldx = x;
			oldy = y;
			dupcount = 0;
			subdupcount = 0;
			}
		}



	/* render text, mark or line.. */
	/* text can be combined with mark if text and symbol both specified */

	/* mark */
	if( strlen( symbol ) > 0 || (strlen( text ) < 1 && linelen <= 0.0 && lblfield < 0 ) ) {
		if( sizefield >= 0 ) 
			radius = sqrt((atof( da( realrow, sizefield ) ) * sizescale)/3.1415927);
			         /* sizefield scales up the AREA of symbol, not the diameter */
		Emark( x, y, symcode, radius );
		}

	/* text */
	if( strlen( text ) > 0 ) {
		if( symbol[0] != '\0' )  /* set text color etc... */
			textdet( "textdetails", textdetails, &align, &adjx, &adjy, -3, "R" );
		if( sizefield >= 0 ) Etextsize( (int) (atof( da( realrow, sizefield ) ) * sizescale) );
		if( verttext ) Emov( (x+cy)+adjx, y ); /* cy puts midheight of character on point */
		else Emov( x+adjx, (y-cy)+adjy );
		if( align == '?' ) Edotext( text, 'C' );
		else Edotext( text, align );
		if( symbol[0] != '\0'  )  /* restore symbol color etc... */
			symdet( "symbol", symbol, symcode, &radius );
		}

	/* label from data */
	else if( lblfield  >= 0 ) {
		if( sizefield >= 0 ) Etextsize( (int) (atof( da( realrow, sizefield ) ) * sizescale) );
		if( verttext) Emov( (x+cy)+adjx, y ); /* cy puts midheight of character on point */
		else Emov( x+adjx, (y-cy)+adjy );

		if( align == '?' ) Edotext( da( realrow, lblfield ), 'C' );
		else Edotext( da( realrow, lblfield ), align );
		/* fprintf( stderr, "[%g %g %s]", x+adjx, (y-cy)+adjy, da( i, lblfield) ); */
		}

	/* line */
	else if( linelen > 0.0 ) {
		if( sizefield >= 0 ) hlinelen = linelen * 0.5 * atof( da( realrow, sizefield ) ); 
					/* sizefield acts as a scale factor to linelen */
		if( yfield < 0 ) {
			Emov( x, y-hlinelen );
			Elin( x, y+hlinelen );
			}
		else	{
			Emov( x-hlinelen, y );
			Elin( x+hlinelen, y );
			}
		}

	}
if( verttext ) Etextdir( 0 );

if( strlen( legendlabel ) > 0 ) {
	char s[40];
	sprintf( s, "%d", irow );
	Evarsub( legendlabel, "@NVALUES", s );
	if( linelen <= 0.0 && lblfield < 0 && strlen( text ) < 1 )
		add_legent( LEGEND_SYMBOL, legendlabel, symbol, "", "" );
	else if( strlen( symbol ) > 0 && (strlen( text ) > 0 ) )
		add_legent( LEGEND_SYMBOL+LEGEND_TEXT, legendlabel, text, textdetails, symbol );
	}

setintvar( "NVALUES", irow );

return( 0 );
}
/* ======================= */
static int
ptcompare( f, g )
double *f;
double *g;
{
double *f2, *g2;
if( *f > *g ) return( 1 );
else if( *f < *g ) return( -1 );
else	{
	/* advance to Y component */
	f2 = f+1;
	g2 = g+1;
	if( *f2 > *g2 ) return( 1 );
	else if( *f2 < *g2 ) return( -1 );
	else return( 0 ); /* same */
	}
}
