/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* SCRIPTER - process scripts. 

   See README.scripter for scripter syntax, etc.

   Caller must open script file for reading by calling scripter_open().
   Then it must call scripter() repeatedly.  Scripter
   will evaluate script and pass back one line at a time in the
   'line' argument (line should have a size of at least 256 chars).
   (The trailing newline will be suppressed for script lines which end in \c )

   Returns 0 when there is more script to execute.
	   -1 (EOF) when end of script is reached.
	   or the code returned via #exit (255 if #exit is used without a code)

   scg 8/16/00 - make #if, #loop, etc. work in #included file
*/
#include "dms.x"


scripter( ss, line, recordid, data )
struct scripterstate *ss; /* scripter state structure - initialize using scripter_open(()) */
char *line;	/* returned- the next line after evaluation */
char *recordid; /* recordid which identifies data array, 
			or "" if no data items will be accessed by name.
			Note: if #loaddata or #datadef is used this must be 
				a changable C variable. */
char data[][DATAMAXLEN+1];  /* data array */

{
char buf[256];
char tok[256];
int stat;
int i, j;
int nt;
char itemname[80];
char *getok();
char label[256];
long ftell(); /* loop offset */
int len;
FILE *popen();



while( 1 ) {
	if( ss->incflag ) { /* handle included file.. */
		if( fgets( line, 255, (ss->incfp) ) == NULL ) { 
			ss->incflag = 0; 
			fclose( ss->incfp );
			ss->incfp = NULL;
			continue; 
			}
		}
	else 	{   /* regular primary file.. */
		if( fgets( line, 255, (ss->fp) ) == NULL ) break;
		}

	/* check for \c - suppress newline */
	if( line[ strlen( line ) -3 ] == '\\' && line[ strlen( line ) -2 ] == 'c' )
		line[ strlen( line ) - 3 ] = '\0';

	/* expand_tabs( buf, line ); */ /* publib */ /* convert tabs to spaces.. */
	strcpy( buf, line );

	/* get first token.. */
	nt = sscanf( buf, "%s", tok );
	if( nt < 1 ) strcpy( tok, "" );
	if( strncmp( tok, "//", 2 )==0 ) continue; /* comment */

	/* remove trailing newline and comments from statements.. */
	if( tok[0] == '#' ) {
		buf[ strlen( buf )-1 ] = '\0';
		rm_comments( buf );
		}

	/* now convert @items to literals, copying back to 'line' */
	if( smember( tok, "#if" ) )  /* changed, scg 1/28/98, used to include #set as well */
		stat = value_subst( line, buf, data, recordid, FOR_CONDEX ); 
	else stat = value_subst( line, buf, data, recordid, NORMAL );  /* dmslib */
	if( stat > 1 ) return( err( stat, 0, "Unrecognized variable or data item", buf ) );
	if( strlen( line ) > 255 ) return( err( 401, 0, "Line too long after value subst.", "" ) );


	/* output text.. */
	if( tok[0] != '#' ) {
		/* convert _ to ' ' */
		if( ss->hideund ) /* convert _ to ' ' */
                        for( i = 0; i < strlen( line ); i++ ) if( line[i] == '_' )line[i] = ' ';

		return( 0 );
		}





	/* operators.. */

	if( strcmp( tok, "#set" ) ==0 ) {
		stat = do_set( line, data, recordid );
		if( stat != 0 ) return( stat );
		}


	else if( strcmp( tok, "#if" ) == 0 ) {
		i = 0;
		getok( line, &i );
		stat = condex( &line[i] ); /* dmslib */
		if( stat < 0 ) return( err( 403, 0, line, "" ) );
		if( stat == 0 ) {  /* skip to matching #endif */
			while( 1 ) {
				if( ss->incflag ) { if( fgets( line, 511, ss->incfp ) == NULL ) break; }
				else  { if( fgets( line, 511, ss->fp ) == NULL ) break; }
		
				nt = sscanf( line, "%s", tok );
				if( nt < 1 ) continue;
				if( strcmp( tok, "#if" )==0 ) ss->ifnest++;
				else if( smember( tok, "#endif #fi" )) {
					if( ss->ifnest <= 0 ) break;
					else ss->ifnest--;
					}
				}
			}
		continue;
		}
	else if( smember( tok, "#endif #fi" )) continue;

	else if( strcmp( tok, "#hideund" )==0 ) ss->hideund = 1;
	else if( strcmp( tok, "#showund" )==0 ) ss->hideund = 0;

	else if( strcmp( tok, "#declare" )==0 ) { /* declare a temp variable */
		stat = do_declare( line, data, recordid );
		if( stat != 0 ) return( stat );
		}

	else if( strcmp( tok, "#musthave" )==0 ) { /* require that a variable/parameter exist 
							and have a length > 0 */
		stat = do_musthave( line, data, recordid );
		if( stat != 0 ) return( stat );
		}
		

	else if( strcmp( tok, "#exit" )==0 ) { /* quit and return a status code */
		i = 0;
		getok( line, &i ); /* get #exit.. */
		strcpy( buf, getok( line, &i ) ); /* get status code if any */
		if( strcmp( buf, "0" )==0 ) 
			return( err( 411, 0, "#exit status code of 0 is not allowed.", line ) );
		if( atoi( buf ) != 0 ) return( atoi( buf ) );
		else return( 255 );
		}

	else if( strcmp( tok, "#jumpto" )==0 ) { /* jump forward to a label.. */
		int skipping;
		i = 0;
		getok( line, &i ); /* get #jumpto.. */
		strcpy( buf, getok( line, &i ) ); /* get label */
		skipping = 1;
		while( 1 ) {
			if( ss->incflag ) { if( fgets( line, 511, ss->incfp ) == NULL ) break; }
			else  { if( fgets( line, 511, ss->fp ) == NULL ) break; }

			sscanf( line, "%s %s", tok, label );
			if( strncmp( tok, "#label", 6 ) ==0 ) {
				if( strcmp( label, buf )==0 ) {
					skipping = 0;
					break;
					}
				}
			}
		if( skipping ) return( err( 405, 0, "Missing #jumpto label", buf ) );
		}

	else if( strcmp( tok, "#label" )==0 ) continue;

	else if( strcmp( tok, "#loop" )==0 ) {
		if( ss->incflag ) ss->lofs = ftell( ss->incfp );
		else ss->lofs = ftell( ss->fp );
		continue;
		}

	else if( strcmp( tok, "#endloop" )==0 ) {  /* jump back to begining of loop */
		if( ss->lofs < 0L ) return( err( 406, 0, "#endloop: no matching #loop statement", buf ) );

		if( ss->incflag ) fseek( ss->incfp, ss->lofs, 0 );
		else fseek( ss->fp, ss->lofs, 0 /* SEEK_SET BSD */ );

		continue;
		}

	else if( strcmp( tok, "#continue" )==0 ) { /* jump back to beginning of loop */
		if( ss->lofs < 0L ) return( err( 406, 0, "#continue: not within a #loop", buf ) );

		if( ss->incflag ) fseek( ss->incfp, ss->lofs, 0 );
		else fseek( ss->fp, ss->lofs, 0 /* SEEK_SET BSD */ );

		continue;
		}

	else if( strcmp( tok, "#break" )==0 ) {    /* skip to #endloop */
		if( ss->lofs < 0L ) return( err( 407, 0, "Break is not within a #loop", buf ) );

		while( 1 ) {
			if( ss->incflag ) { if( fgets( line, 511, ss->incfp ) == NULL ) break; }
			else { if( fgets( line, 511, ss->fp ) == NULL ) break; }

			sscanf( line, "%s", tok );
			if( strcmp( tok, "#endloop" ) ==0 ) break;
			}
		ss->lofs = -1;
		continue;
		}

#ifndef PLOTICUS
	else if( strcmp( tok, "#loaddata" )==0 ) {   /* load data array using results
							from most recent $dba_get() */
		stat = dba_copy_current_record( data, recordid, &i );
		if( stat != 0 ) return( err( 409, 0, "#loaddata: no recent $dba_get", buf ) );
		}
#endif PLOTICUS

	else if( strcmp( tok, "#datadef" )==0 ) { /* set recordid, data already loaded */
		i = 0;
		getok( line, &i ); /* get #datadef */
		strcpy( recordid, getok( line, &i ) ); /* get id */
		}

	else if( strcmp( tok, "#line" )==0 ) { /* draw a hor. line (since native underscores vanish) */
		i = 0;
		getok( line, &i ); /* get #line */
		strcpy( buf, getok( line, &i ) ); /* get length */
		len = atoi( buf );
		for( i = 0; i < len; i++ ) line[i] = '_';
		line[i] = '\0';
		return( 0 );
		}

	else if( strcmp( tok, "#include" ) ==0 || strcmp( tok, "#includesh" ) == 0 ) { /* include another file */
								/* #includesh added scg 11/22/00  */
		if( ss->incflag ) err( 412, 0, "nested #include not allowed, ignoring.", line );
		i = 0;
		getok( line, &i ); /* get #line */
		
		if( strcmp( tok, "#includesh" )==0 ) {
			ss->incfp = popen( &line[i+1], "r" );  /* this really should be closed by pclose() */
			if( ss->incfp == NULL ) {
				err( 410, 0, "Warning, cannot #includesh results of command", &line[i+1] ); 
				continue; 
				}
			}
		else 	{
			strcpy( tok, getok( line, &i ) ); /* get filename */
			ss->incfp = fopen( tok, "r" );
			if( ss->incfp == NULL ) { 
				err( 410, 0, "Warning, cannot #include file", tok ); 
				continue; 
				}
			}
		ss->incflag = 1;
		continue;
		}

	else if( strcmp( tok, "#shell" )==0 ) {
		i = 0;
		fflush( stdout );
		getok( line, &i ); /* get #shell */
		system( &line[i] );
		continue;
		}

	else if( strcmp( tok, "#shellmore" )==0 ) { /* added SCG 3/30/99 */
		FILE *popen(), *pfp;
		int ix, nshlines;
		i = 0;
		nshlines = 0;
		getok( line, &i ); /* get #shell */
		ix = i;
		strcpy( tok, getok( line, &ix ) ); /* get optional arg */
		pfp = popen( &line[i], "r" );
		if( pfp == NULL ) continue;
		while( fgets( buf, 511, pfp ) != NULL ) {
			nshlines++;
			printf( "%s", buf );
			}
		pclose( pfp );
		/* declare and set _NSHLINES variable to the number of lines generated.. */
		declarevar( "_NSHLINES" );
		sprintf( buf, "%d", nshlines );
		setvar( "_NSHLINES", buf );
		continue;
		}

	else if( strcmp( tok, "#put" )==0 ) {
		i = 0;
		getok( line, &i ); /* get #put */
		sscanf( &line[i], "%s", tok );
		strcpy( line, tok );
		return( 0 );
		}

	else if( strcmp( tok, "#print" )==0 ) {
		i = 0;
		getok( line, &i ); /* get #print */
		i++;
		strcpy( line, &line[i] );
		strcat( line, "\n" );
		return( 0 );
		}

	else if( tok[0] == '#' ) {
		/* Just return with line .. */
		strcat( line, "\n" );
		return( 0 );
		}

	}
return( EOF );
}


/* ===================== */
/* SCRIPTER_OPEN - initialize scripterstate structure.
   Returns -1 if filename could not be opened or if file is empty. */
scripter_open( ss, filename )
struct scripterstate *ss;
char *filename;
{
char buf[256];
int stat;

/* first try file name as is.. */
ss->fp = fopen( filename, "r" );
if( ss->fp == NULL ) {
#ifndef PLOTICUS
	/* now try Scripterdir */
	sprintf( buf, "%s/%s", Scripterdir, filename );
	ss->fp = fopen( buf, "r" );
#endif
	if( ss->fp == NULL ) {
		return( -1 );
		}
	}

/* check for empty file.. */
if( fgets( buf, 255, ss->fp ) == NULL ) {
	fclose( ss->fp );
	return( -1 );
	}
rewind( ss->fp );

ss->ifnest = 0;
ss->lofs = -1;
ss->loopnest = 0;
ss->incflag = 0;
ss->hideund = 0;
ss->incfp = NULL; /* scg 10/12/00 */
#ifndef PLOTICUS
itemfld_silence();
declarevar( "_SCRIPT_TYPE" );
setvar( "_SCRIPT_TYPE", "batch" );
declarevar( "_DATAPATH" );
setvar( "_DATAPATH", Datapath );
#endif
return( 0 );
}

/* ===================== */
/* SCRIPTER_CLOSE - close script file */
scripter_close( ss )
struct scripterstate *ss;
{
fclose( ss->fp );
if( ss->incfp != NULL ) fclose( ss->incfp ); /* added scg 8/16/00 */
return( 0 );
}

/* ====================== */
/* SCRIPTER_CALL - convenience routine for executing script 
	and return first line of results */
scripter_call( value, scriptfile )
char *value;   /* should be large enough for one line of results */
char *scriptfile;
{
char data[MAXITEMS][DATAMAXLEN+1];
struct scripterstate ss;
int stat;
char recordid[40];
char buf[256];
char oldscripttype[40];

strcpy( recordid, "" );
strcpy( value, "" );
stat = getvar( "_SCRIPT_TYPE", oldscripttype );
if( stat != 0 ) strcpy( oldscripttype, "" );
stat = scripter_open( &ss, scriptfile );
if( stat != 0 ) return( stat );
while( 1 ) {
	stat = scripter( &ss, buf, recordid, data );
	if( stat != 0 ) break;
	strcat( value, buf );
	if( buf[ strlen( buf ) -1 ] == '\n' ) break;
	}
scripter_close( &ss );
setvar( "_SCRIPT_TYPE", oldscripttype ); /* restore _SCRIPT_TYPE */
if( strlen( value ) < 1 ) return( -1 );
else return( 0 );
}



