/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* SCRIPTOPS - operations that are general to all script types. */
/* SCG */

#include "dms.x"


/* =========================== */
/* DO_SET - handle a #set line */

do_set( line, data, recordid )
char *line;
char data[MAXITEMS][DATAMAXLEN+1];
char *recordid;
{
int i, ix;
char itemname[80];
int stat;
char *getok();
char tok[80];
char buf[256];
char value[256];

i = 0;
getok( line, &i ); /* get #set.. */
strcpy( itemname, getok( line, &i ) ); /* get itemname */
if( itemname[0] == '@' ) strcpy( itemname, &itemname[1] );

/* skip the optional equals sign.. */
ix = i; if( strcmp( getok( line, &ix ), "=" )==0 ) i = ix; 
while( line[i] == ' ' ) i++; /* skip over white space.. */

if( line[i] == '"' ) {  /* do value substitution only.. */
	stat =  assignvalue( itemname, &line[i+1], data, recordid );
	if( stat != 0 ) return( err( stat, 0, "Variable to be set is undeclared", line ) );
	}

else	{   /* do condex evaluation */

	/* updated scg 11/13/00 */
	if( line[i] == '$' ) {   /* a standalone function.. */
		strcpy( buf, &line[i] );
		stat = function_call( buf );
		if( stat < 0 ) return( err( 409, 0, "Function error", line ) );
		if( buf[ strlen( buf )-1 ] == ' ' ) { buf[ strlen( buf ) -1 ] = '\0'; } /* forced alpha */
		stat = assignvalue( itemname, buf, data, recordid ); 
		if( stat != 0 ) return( err( stat, 0, "Variable to be set is undeclared", line ) );
		}
	else    {   /* do condex evaluation */
		stat = condex( &line[i] );
		if( stat < 0 ) return( err( 402, 0, "Expression error", line ) );
		stat = assignvalue( itemname, condex_results(), data, recordid ); 
		if( stat != 0 ) return( err( stat, 0, "Variable to be set is undeclared", line ) );
		}

	/* stat = condex( &line[i] );
	 * if( stat < 0 ) return( err( 402, 0, "Expression error", line ) );
	 * stat = assignvalue( itemname, condex_results(), data, recordid ); 
	 * if( stat != 0 ) return( err( stat, 0, "Variable to be set is undeclared", line ) );
	 */
	}
return( 0 );
}

/* =========================== */
/* DO_DECLARE - handle a #declare line */
/* 		data may be passed in as NULL - if so no comparison against
						#datarec item names is done. */

/* 020400 scg - support for lists added, e.g. #declare ID, VISIT, NAMECODE */

do_declare( line, data, recordid )
char *line;
char data[MAXITEMS][DATAMAXLEN+1];
char *recordid;
{
int i, ii;
char itemname[80];
int stat;
char *getok();
char opt[256];

/* handle var name list.. convert commas to blanks.. */
for( i = 0; i < strlen( line ); i++ ) if( line[i] == ',' ) line[i] = ' ';

i = 0;
getok( line, &i ); /* get #declare.. */

while( 1 ) {

	if( i >= strlen( line ) ) break;

	strcpy( itemname, getok( line, &i ) ); /* get itemname */
	if( strlen( itemname ) < 1 ) break;
	if( data != NULL ) {
#ifndef PLOTICUS
		if( itemfld( recordid, itemname ) >= 0 ) 
			return( err( 404, 0, "#declare: data item name conflict", itemname ) );
#endif
	;
		}
	ii = i; /* lookahead tentatively.. */
	strcpy( opt, getok( line, &ii ) ); /* get opt */
	if( strcmp( opt, "soft" )==0 ) { /* soft option.. only declare/initialize if not
					declared already */
		i = ii; /* eat token */
		stat = 0;
		if( getvar( itemname, opt ) != 0 ) stat = declarevar( itemname );
		if( stat != 0 ) return( err( stat, 0, "error in #declare soft", line ) );
		continue;
		}
	stat = declarevar( itemname );
	if( stat != 0 ) return( err( stat, 0, "#declare error", line ) );
	}
return( 0 );
}

/* =========================== */
/* DO_MUSTHAVE - handle a #musthave line */

do_musthave( line, data, recordid )
char *line;
char data[MAXITEMS][DATAMAXLEN+1];
char *recordid;
{
int i;
char itemname[80];
int stat;
char buf[255];
char *getok();

for( i = 0; i < strlen( line ); i++ ) if( line[i] == ',' ) line[i] = ' ';
i = 0;
getok( line, &i ); /* get #musthave.. */
while( 1 ) {
	strcpy( itemname, getok( line, &i ) ); /* get itemname */
	if( strlen( itemname ) < 1 ) break;
	stat = getvalue( buf, itemname, data, recordid );
	if( stat != 0  || strlen( buf ) < 1 ) 
		return( err( 408, 0, "Required variable not set.", itemname ) );
	}
return( 0 );
}


/* ============================= */
/* RM_COMMENTS - remove comments given to the right of statements */

rm_comments( line )
char *line;
{
int i;
char tok[256];

/* don't do this for shell commands, they may need to contain // */
sscanf( line, "%s", tok );
if( strncmp( tok, "#shell", 6 )==0 ) return( 0 ); 

for( i = 1; i < strlen( line ); i++ ) {
	if( line[i] == '/' && line[i+1] == '/' && (line[i-1] == ' ' || line[i-1] == '\t') ) {
		line[i] = '\n';
		line[i+1] = '\0';
		break;
		}
	}
return( 0 );
}
