/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include <stdio.h>
/* TIMES - time notation handling and conversion

standard units = (double) minutes since 0:00:00.0
		 --seconds are expressed as decimal portion of 1 minute

input formats for which arithmetic is supported:    
format	input			output
HHMM 	hh:mm or hh:mm:ss     	(output is hh:mm)
HHMMSS	hh:mm or hh:mm:ss	(output is hh:mm:ss)
MMSS	mm:ss  			(output is mm:ss)

output formats:   hh:mm hh:mma  hh:mm:ss hh:mm:ssa mm:ss  
	(a=use am/pm notation; A yeilds capitals e.g. AM or PM)

To get current time in seconds past midnight, use tomin( "now", &sec );
To get current time in current format use frommin( -1.0, result );

*/
#define HHMM 0
#define MMSS 1
#define HHMMSS 2
#define stricmp(s,t) strcasecmp(s,t)

static int format = HHMM;
static int Hr = 0, Min = 0;
static double Sec = 0.0;
static double Dtwinbegin = 0.0, Dtwinend = 1440.0, Dtwinsize = 1440.0; /* default: 24 hours */
				/* ^ changed scg 12-28-99, was 0.0 */
static int Suppress_twin_warn = 0;

/* ============================= */
/* SETTIMEFMT - set the "current time format" */
settimefmt( fmt )
char *fmt;
{
if( stricmp( fmt, "hh:mm" )==0 ) format = HHMM;
else if( stricmp( fmt, "mm:ss" )==0 ) format = MMSS;
else if( stricmp( fmt, "hh:mm:ss" )==0 ) format = HHMMSS;
else return( -1 );
return( 0 );
}

/* ============================= */
/* TOMIN - take a time in the "current format" and return # minutes since 00:00:00 */
/* Return 0 on valid input; return non-zero on invalid input */
tomin( s, result )
char *s; /* time */
double *result; /* returned */
{
int i;
char t[128];
int nt;

if( stricmp( s, "now" )==0 ) {
	systime( &Hr, &Min, &i );
	Sec = (double)i;
	maketime( Hr, Min, Sec, t );
	}
else strcpy( t, s );

/* validate s */
if( !isdigit(t[0]) || !isdigit(t[strlen(t)-1]) /* || !isdigit(t[strlen(t)-2]) */ )  /* relaxed 2/21/01 */
	{ *result = 0.0; return( 1 ); }

if( format == HHMM || format == HHMMSS ) {
	if( strlen( t ) < 3 ) { *result = 0.0; return( 2 ); } /* sanity check */
	nt = sscanf( t, "%d:%d:%lf", &Hr, &Min, &Sec );
	if( nt == 2 ) *result = (Hr * 60.0) + Min;
	else if( nt == 3 ) *result = (Hr * 60.0) + Min + (Sec/60.0);
	else { *result = 0.0; return( 3 ); } /* error */
	return( 0 );
	}
else if( format == MMSS ) {
	if( strlen( s ) < 4 ) { *result = 0.0; return( 4 ); } /* sanity check */
	nt = sscanf( s, "%d:%lf", &Min, &Sec );
	if( nt != 2 ) { *result = 0.0; return( 6 ); } /* error */
	*result = Min + (Sec/60.0);
	}
return( 0 );
}

/* ============================ */
/* FROMMIN - take # seconds since 00:00:00 and return time string in current format */
frommin( s, result )
double s; /* sec may be < 0 for "now" */
char *result;
{
int i;
if( s < 0.0 ) {
	systime( &Hr, &Min, &i );
	Sec = (double)i;
	}
else	{
	Hr = (int)(s/60.0);
	Min = ((int)(s)-(Hr*60));
	Sec = s - (double)((Hr*60) + (Min));
	}
maketime( Hr, Min, Sec, result );
return( 0 );
}

/* ============================= */
/* GETHMS - get hour, min, sec components of most recent time 
	processed by tomin() for frommin() */
gethms( hour, min, sec )
int *hour, *min;
double *sec;
{
*hour = Hr;
*min = Min;
*sec = Sec;
return( 0 );
}

/* ============================= */
/* MAKETIME - given h m s, build time string using current format. */
maketime( hr, min, sec, result )
int hr, min;
double sec;
char *result;
{
hr = hr % 24;
if( format == HHMM ) sprintf( result, "%d:%02d", hr, min );
else if( format == MMSS ) sprintf( result,  "%d:%02g", min, sec );
else if( format == HHMMSS ) sprintf( result,  "%d:%02d:%02g", hr, min, sec ); 
return( 0 );
}


/* ============================== */
/* take time in current format and convert to outformat */
formattime( s, outformat, result )
char *s, *outformat, *result;
{
int oldformat;
double min;
int hampm;
char c;
char ampm[4];
char fmt[40];

strcpy( fmt, outformat );

tomin( s, &min );

/* in case we do am/pm */
strcpy( ampm, "" );
c = fmt[ strlen( fmt ) - 1];
if( c == 'a' || c == 'A' ) {
	if( c == 'a' ) { 
		if( Hr >= 12 && Hr < 24 ) strcpy( ampm, "pm" );
		else strcpy( ampm, "am" );
		}
	else if( c == 'A' ) { 
		if( Hr >= 12 && Hr < 24 ) strcpy( ampm, "PM" );
		else strcpy( ampm, "AM" );
		}
	fmt[ strlen( fmt ) - 1 ] = '\0';
	if( Hr > 12 ) {
		min -= 720.0;
		Hr -= 12;
		}
	}

if( stricmp( fmt, "hh" )==0 ) sprintf( result, "%02d", Hr );
else if( stricmp( fmt, "h" )==0 ) sprintf( result, "%0d", Hr );
else if( stricmp( fmt, "mm" )==0 ) sprintf( result, "%02d", Min );
else if( stricmp( fmt, "m" )==0 ) sprintf( result, "%d", Min );
else if( stricmp( fmt, "ss" )==0 ) sprintf( result, "%g", Sec );
else	{
	oldformat = format;
	settimefmt( fmt );
	frommin( min, result );
	format = oldformat;
	}

if( strlen( ampm ) > 0 ) {
	if( result[0] == '0' ) sprintf( result, "%s%s", &result[1], ampm );
	else strcat( result, ampm );
	}
return( 0 );
}


/* =============================== */
/* TIMEDIFF - find difference between two times */
timediff( s1, s2, diff )
char *s1, *s2;
double *diff; /* s1-s2 difference in minutes */
{
double min1, min2;

tomin( s1, &min1 );
tomin( s2, &min2 );
*diff = min1 - min2;
return( 0 );
}

/* ================================== */
/* ================================== */
/* Datetime routines                  */
/* ================================== */

/* ================================== */
/* SETDATETIMEFMT - set the format for the date and time 
   	components of datetime */
setdatetimefmt( fmt, window )
char *fmt;
char *window; /* Specifies a 'window' of hours to be shown for each day.
		 Notation nn-mm where nn is hour to begin with and mm is 
		hour to end with (0-24) */
{
char datepart[20], timepart[20];
char beginpart[20], endpart[20];
int i;

i = 0;
Egetchunk( datepart, fmt, &i, "." );
Egetchunk( timepart, fmt, &i, "." );
setdatefmt( datepart );
settimefmt( timepart );

if( strlen( window ) > 0 ) { /* parse 'window' specification */
	i = 0;
	Egetchunk( beginpart, window, &i, "-" );
	Egetchunk( endpart, window, &i, "-" );
	Dtwinbegin = atoi( beginpart ) * 60.0;
	Dtwinend = atoi( endpart ) * 60.0;
	Dtwinsize = (Dtwinend - Dtwinbegin);
	}
if( strlen( window ) == 0 || Dtwinsize == 0.0 ) {
	Dtwinbegin = 0.0;
	Dtwinend = 1440.0;
	Dtwinsize = 1440.0;
	}

return( 0 );
}


/* ================================== */
/* DATETIME2DAYS - convert a datetime to julian date with 
	decimal portion representing the time component.

   Returns 0 if ok, 1 on error */
datetime2days( dt, days )
char *dt;
double *days;
{
char datepart[20], timepart[20];
long jul;
int i, stat;
double mins;

i = 0;
Egetchunk( datepart, dt, &i, "." );
Egetchunk( timepart, dt, &i, "." );
stat = jdate( datepart, &jul );
if( stat != 0 ) return( 1 );

stat = tomin( timepart, &mins );
if( stat != 0 ) return( 1 );

if( mins < Dtwinbegin || mins > Dtwinend ) {
	if( !Suppress_twin_warn ) 
	   err( 4279, 0, "warning: time is outside of window range- displaying it at window boundary", dt );
	if( mins < Dtwinbegin ) mins = Dtwinbegin;
	if( mins > Dtwinend ) mins = Dtwinend;
	}

mins = (mins-Dtwinbegin) / Dtwinsize; /* normalize to  0.0 - 1.0 */
*days = jul+mins;
return( 0 );
}
/* ================================== */
/* DAYS2DATETIME - convert decimal julian date to datetime */
days2datetime( days, dt )
double days;
char *dt;
{
char s[30], t[30];
double datepart, timepart, f, floor(), fmod(), rint();

datepart = floor( days );
timepart = days - datepart;

timepart = (timepart * Dtwinsize) + Dtwinbegin; /* convert from 0.0 - 1.0 to #min past midnite */
timepart = rint( timepart );  /* remove any rounding error introduced by above */

fromjul( (long)datepart, s );
frommin( timepart, t );

sprintf( dt, "%s.%s", s, t );
return( 0 );
}
/* ================================== */
/* FORMATDATETIME - convert a datetime from current format 
	to a new format for display.  Either the date or
	time component may be omitted for display purposes.
*/
formatdatetime( dt, format, newdt )
char *dt, *format, *newdt;
{
char datepart[20], timepart[20];
char datefmt[20], timefmt[20];
char s[30], t[30];
int i;

i = 0;
strcpy( datefmt, "" );
strcpy( timefmt, "" );
if( format[0] != '.' && tolower(format[0]) != 'h' ) Egetchunk( datefmt, format, &i, "." );
Egetchunk( timefmt, format, &i, "." );

i = 0;
Egetchunk( datepart, dt, &i, "." );
Egetchunk( timepart, dt, &i, "." );

if( strlen( datefmt ) > 0 ) {
	formatdate( datepart, datefmt, s );
	}
else strcpy( s, "" );
if( strlen( timefmt ) > 0 ) {
	if( strlen( datefmt ) > 0 )strcpy( t, "." );	
	else strcpy( t, "" );
	formattime( timepart, timefmt, &t[strlen(t)] );
	}
else strcpy( t, "" );

sprintf( newdt, "%s%s", s, t );

return( 0 );
}

/* ================================= */
/* SUPPRESS_TWIN_WARN - allow warning message 'value outside datetime
	window' to be suppressed */
suppress_twin_warn( mode )
int mode;
{
Suppress_twin_warn = mode;
return( 0 );
}


#ifdef TESTING
/* ================================== */
main()
{
double min;
char result[80];
char result2[80];

tomin( "now", &min );
frommin( min, result );
formattime( result, "hh:mmA", result2 );
printf( "%s\n", result2 );
}
#endif


