/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* VALIDATION - routines for validating data items. */

/* IF a type is added here, it must also be added in fdfcheck.c */


/* VALIDATE - given an input value and a data type string, determines
	which validation routine to execute (all are contained herein),
	and calls it.  Returns 1 if value ok, 0 if not (plus error() gets
	called with an error message).  Also, each error has an identifying
	number; the number of the most recently occurring error may be
	accessed by calling get_validate_error().

	All messages with the same meaning should use the same error number.

	History
	------------------------------------------------------
	Auth: Steve Grubb 

	2-11-97 scg converted to c2j().
	2-10-97 scg converted to new dates infrastructure.
*/



/* Note: a link to this module exists in fip/src for use by the
	generic fipview system.  Compile flag -FIP will be used
	in that context */
#ifdef PLOTICUS
#define goodnum(a,b) Egoodnum(a,b)
#define member(a,b) Emember(a,b)
#define us_to_blank(a)  strlen(a)
#define FIP
#endif
#ifdef FIP
#include <stdio.h>
#define NO 0
#define YES 1
#define DATAMAXLEN 80 

#else
#include "dms.x"
#endif




#include <ctype.h>
#define BADCHAR( c ) ( (c!='+'&&c!='-'&&c!='.') && (c<'0' || (c>'9'&&c<'A') || (c>'Z'&&c<'a') || c>'z'||c=='#'))
#define MAXLIST 50
#define MAXPATH 100
#define MAXCTREC 256

static int checkcmt( ), checkspec( ), checkchar( ) , checkans( ), checkx( ),
  checkitem( ), checkalpha( ), checkdt( ), checkfloat( ), checkopt( ),
  checknumer( ), checklist( ), checklookup( ), checkmulti( ),
	checkrecentdt(), checktimems();
#ifndef FIP
static int checkdba(), check_scriptercall(), checkfunc();
#endif

static int val_errno = 10;

/* ===================== */
/* GET_VALIDATE_ERRNO - access the most recent validation error number */
get_validate_errno()
{
return( val_errno );
}
/* ===================== */
/* SET_VALIDATE_ERRNO - allow a custom validation routine to set the error number */
set_validate_errno( n )
int n;
{
val_errno = n;
return( 0 );
}



/* ===================== */

validate( uin, typestr )
char uin[], typestr[];
{
int status;
char ans[10], group[10];
char typ[40];

sscanf( typestr, "%s", typ );
if( strlen( uin ) < 1 ) return( 0 );

status = NO;
/* data types */
if( strcmp( typ, "list" )==0 ) status = checklist( uin, typestr );
else if( strcmp( typ, "ans" )==0 ) status = checkans( uin );
else if( strcmp( typ, "x" )==0 ) status = checkx( uin );
else if( smember( typ, "date mmyy yymm lazydate" ) ) status = checkdt( uin, typestr );
else if( smember( typ, "recentdate" )) status = checkrecentdt( uin, typestr );
else if( smember( typ, "numeric numfill" ) ) status = checknumer( uin, typestr );
else if( strcmp( typ, "float" )==0 ) status = checkfloat( uin, typestr );
else if( strcmp( typ, "table" )==0 ) status = checklookup( uin, typestr );
else if( strcmp( typ, "alpha" )==0 ) status = checkalpha( uin, typestr );
else if( strcmp( typ, "multi" )==0 ) status = checkmulti( uin, typestr );
else if( strcmp( typ, "char" )==0 ) status = checkchar( uin );
else if( strcmp( typ, "item" )==0 ) status = checkitem( uin, typestr );
else if( strcmp( typ, "comment" )==0 ) status = checkcmt( uin );
else if( smember( typ, "timems" )) status = checktimems( uin, typestr );
else status = custom_validation( uin, typestr );
if( status == -1 ) { 
	fprintf( stderr, "[303 Trace: value was %s]", uin );
	return( err( 303, 0, "Invalid data type", typestr ) );
	}
return( status );
}


/* ============================ */
static int checkcmt( uin )
char uin[];
{
int i;
for( i = 0; i < strlen( uin ); i++ ) if( member( uin[i], "\033`~\\" ) ) {
	val_errno = 11;
	return( NO );
	}
return( YES );
}



/* ============================ */
static int checkchar( uin ) 
char uin[];
{
 if( !BADCHAR( uin[0] ) ) return( YES );
 else { error( "Invalid char." ); val_errno = 12; return( NO ); }
}




/* ============================ */
static int checkans( uin )
char uin[];
{
if( uin[0] == 'y' || uin[0] == 'n' ) return( YES );
else { val_errno = 13; return( NO ); }
}



/* ============================ */
static int checkx( uin )
char uin[];
{
   if( uin[0] == 'x' ) return( YES ); 
   else { error( "If data is present, input should be 'x'." ); val_errno = 14; return( NO ); }
}



/* ============================ */
/* Handles these types:  item, item list, item dba  */
static int checkitem( uin, typestr )
char uin[];
char typestr[];
{
int i, j;
char tok[80], *getok();


i = 0;
strcpy( tok, getok( typestr, &i ) ); /* "item" */
j = i;
strcpy( tok, getok( typestr, &i ) ); /* "list" or "dba_get" or nothing */
if( strcmp( tok, "list" )==0 ) {
	return( checklist( uin, &typestr[j] ) ); /* run a list check */
	}

#ifndef FIP
else if( strcmp( tok, "function" )==0 ) return( checkfunc( uin, &typestr[i] ) );
else if( strcmp( tok, "dba_get" ) == 0 ) return( checkdba( uin, &typestr[i] ) );
else if( strcmp( tok, "scripter_call" )==0 ) return( check_scriptercall( uin, &typestr[i] ) );
else if( strcmp( tok, "numeric" )==0 ) {
	int prec, stat, lo, hi;
	sscanf( typestr, "%*s %*s %d %d", &lo, &hi );
	stat = goodnum( uin, &prec );
	if( stat == 1 && prec == -1 ) {
		if( atoi( uin ) <= hi && atoi( uin ) >= lo ) return( 1 );
		else { val_errno = 131; return( 0 ); }
		}
	else { val_errno = 130; return( 0 ); }
	}
else if( strcmp( tok, "float" )==0 ) {
	int prec, stat;
	double lo, hi, atof();
	sscanf( typestr, "%*s %*s %lf %lf", &lo, &hi );
	stat = goodnum( uin, &prec );
	if( stat == 1 ) {
		if( atof( uin ) >= lo && atof( uin ) <= hi ) return( 1 );
		else { val_errno = 127; return( 0 ); }
		}
	else { val_errno = 128; return( 0 ); }
	}
#endif

if( strcmp( uin, "x" )==0 ) {
	error( "x not allowed." );
	val_errno = 30;
	return( NO );
	}
for( i = 0; i < strlen( uin ); i++ ) {
	if( uin[i] < 32 || uin[i] > 126 ) {
		error( "Non-ascii character(s) in write-in response" );
		val_errno = 31;
		return( NO );
		}
	}
return( YES );
}


#ifndef FIP
/* ======================= 
CHECKDBA 
The "item dba_get" and "alpha dba_get" types allow validation by retrieval 
from a database file.  The dbacom argument is a DBA command (see dbaccess.c).  
@VAL stands for the value to be validated.
For example:
	item dba_get number from cert where @num = @VAL && @pos in CC,VA
The retrieved item will be placed into the input variable, so translations may
be done, e.g.: dba_get newname from tranlist where @oldname = @@VAL

An input value of "" is converted here to _null_
Note that in scripter and interview tables, double at-signs must be used.  
*/

static int checkdba( uin, dbacom )
char *uin;
char *dbacom;
{
int i;
char buf[256], s[4], tok[100];
int stat;
char tablename[40];


sscanf( dbacom, "%*s %*s %s", tablename );
strcpy( buf, "" );
if( strlen( uin ) < 1 ) strcpy( tok, "_null_" );
else strcpy( tok, uin );
for( i = 0; i < strlen( dbacom ); i++ ) {
	if( strncmp( &dbacom[i], "@VAL", 4 )==0 ) {
		strcat( buf, tok );
		i += 3;
		}
	else 	{
		sprintf( s, "%c", dbacom[i] );
		strcat( buf, s );
		}
	}
	
stat = dba_get( buf, buf );
if( stat == 1 ) {
	error( "Response not found in %s list.", tablename );
	val_errno = 9;
	return( NO );
	}
else if( stat == 0 ) {
	strcpy( uin, buf );
	if( Fullscreenapp ) {  /* display the results.. */
		dba_display_current_record( buf );
		for( i = 0; i < strlen( buf ); i++ ) if( buf[i] == '_' ) buf[i] = ' ';
		message( buf );
		}
	return( YES );
	}
else return( err( stat, 0, "validation dba_get error", dbacom ) );
}

/* ============================ */
/* CHECK_SCRIPTERCALL - 
The "item scripter_call" and "alpha scripter_call" types allow validation by 
running a scripter script.  The input value is placed into the tmp variable VAL. 
The script should produce a result of 1 if VAL is ok, or 0 if not. 
Script should also set VAL to an error message to be displayed on invalid.
Scriptcom argument should be the scriptfile, optionally followed by one or
more var=value pairs. */

static int check_scriptercall( uin, scriptcom )
char *uin;
char *scriptcom;
{
char filename[80];
int i;
char buf[256];
int ix;
char varname[80];
char *getok();

declarevar( "VAL" );

/* convert any embedded spaces to _, (spaces will bomb out in script) - SCG 2/6/98 */
strcpy( buf, uin );
for( i = 0; i < strlen( buf ); i++ ) if( buf[i] == ' ' ) buf[i] = '_';
setvar( "VAL", buf );

i = 0;
strcpy( filename, getok( scriptcom, &i ) );
while( 1 ) {
	strcpy( buf, getok( scriptcom, &i ) );
	if( strlen( buf ) < 1 ) break;

        ix = 0;
        getseg( varname, buf, &ix, "=" );
        declarevar( varname );
        setvar( varname, &buf[ix] );
        }

scripter_call( buf, filename );
if( buf[0] == '1' ) return( YES );
else 	{
	getvar( "VAL", buf );
	error( buf );
	return( NO );
	}
}

/* =========================== */
/* CHECK_FUNC - validate an value by executing the given condex function.
   The value to be tested is placed into the tmp variable VAL.
   The given condex function must return 1 if valid, 0 if not. */

static int checkfunc( uin, func )
char *uin;
char *func;
{
int stat;
char buf[512];
char tok[80];
int i;
char s[10];
if( strlen( uin ) < 1 ) strcpy( tok, "_null_" );
else strcpy( tok, uin );
strcpy( buf, "" );
for( i = 0; i < strlen( func ); i++ ) {
	if( strncmp( &func[i], "@VAL", 4 )==0 ) {
		strcat( buf, tok );
		i += 3;
		}
	else 	{
		sprintf( s, "%c", func[i] );
		strcat( buf, s );
		}
	}
stat = condex( buf );
if( stat == YES ) return( YES );
else	{
	error( "Invalid (%s)", buf ); 
	return( NO );
	}
}
#endif /* FIP */


	

/* ============================ */
static int checkalpha( uin, typline )
char uin[], typline[];
{
char c, format[ DATAMAXLEN+1 ];
char tok[40];
int i;
char *getok();

strcpy( tok, "" );
if( sscanf( typline, "%*s %s %s", format, tok ) < 1 ) return( YES );
if( strlen( uin ) != strlen( format ) ) {
	error( "Response does not conform to required format: %s", format ); 
	val_errno = 16;
	return( NO ); 
	}
for( i = 0; i < strlen( uin ); i++ ) {
	c = uin[i];
	if( !(  ( isupper( c ) && isupper( format[ i ] )) || 
		( islower( c ) && islower( format[ i ] )) ||
		( isdigit( c ) && isdigit( format[ i ] )) ||
		( format[i] == '?' ))) { 
			error( "Response does not conform to required format: %s", format ); 
			val_errno = 16;
			return( NO ); 
			}
	}
/* check for alpha dba_get.. */
if( strcmp( tok, "dba_get" )==0 ) {
	i = 0;
	getok( typline, &i ); /* alpha */
	getok( typline, &i ); /* format */
	getok( typline, &i ); /* dba_get */
#ifndef FIP
	return( checkdba( uin, &typline[i] ) );
#endif
	fprintf( stderr, "dba_get datatype not supported in FIP context\n" );
	return( NO );
	}
else if( strcmp( tok, "scripter_call" )==0 ) {
	getok( typline, &i ); /* alpha */
	getok( typline, &i ); /* format */
	getok( typline, &i ); /* scripter_call */
#ifndef FIP
	return( check_scriptercall( uin, &typline[i] ) );
#endif
	fprintf( stderr, "scripter_call datatype not supported in FIP context\n" );
	return( NO );
	}
else if( strcmp( tok, "function" )==0 ) {
	getok( typline, &i ); /* alpha */
	getok( typline, &i ); /* format */
	getok( typline, &i ); /* function */
#ifndef FIP
	return( checkfunc( uin, &typline[i] ) );
#endif
	}

return( YES );
}





/* ============================ */

static int checkdt( uin, typline )
char uin[], typline[];
{
char type[DATAMAXLEN+1], todaystr[20];
char testdate[20];
char lowdate[20], hidate[20];
char parm1[20], parm2[20];
int i, mon, day, year, tmp, lowyr, hiyr;
int norange; /* 1 if no range being checked.. */
long diff, diff2, jul;
int stat;
int pivotyear;
int nt;

todaysdate( todaystr );
sysdate( &mon, &day, &year ); /* to get year */

/* process year range arguments.. */
strcpy( parm1, "" ); strcpy( parm2, "" );
i = sscanf( typline, "%s %s %s", type, parm1, parm2 );
if( i == 1 ) norange = 1;
else 	{
	norange = 0;

	if( parm1[0] == 'y' ) lowyr = year + atoi( &parm1[2] );
	else lowyr = atoi( parm1 );
	if( parm2[0] == 'y' ) hiyr = year + atoi( &parm2[2] );
	else hiyr = atoi( parm2 );
	}


if( strcmp( type, "mmyy" )==0 ) { /* mmyy date type */
	nt = sscanf( uin, "%2d%2d", &mon, &year );
	if( nt != 2 ) { error( "Invalid date" ); val_errno = 18; return( NO ); }
	day = 1;
	stat = makedate( year, mon, day, "", testdate );
	}
else if( strcmp( type, "yymm" )==0 ) { /* yymm date type */
	nt = sscanf( uin, "%2d%2d", &year, &mon );
	if( nt != 2 ) { error( "Invalid date" ); val_errno = 18; return( NO ); }
	day = 1;
	stat = makedate( year, mon, day, "", testdate );
	}

else strcpy( testdate, uin );

if( strcmp( type, "lazydate" )==0 ) setlazydates( 1, 1 );


stat = jdate( testdate, &jul );

/* the following are allowed by jdate() so trap them here.. */
if( strcmp( testdate, "today" )==0 || strcmp( testdate, "TODAY" )==0 ) stat = 1;

if( stat != 0 ) {
	error( "Invalid date" ); 
	val_errno = 18;
	return( NO ); 
	}
if( strcmp( parm2, "yr" )==0 ) {  /* if max limit is 'yr', dates after today not allowed */
	stat = daysdiff( testdate, todaystr, &diff );
	if( diff > 0 ) {
		error( "Date is later than today (%s)", todaystr ); 
		val_errno = 19;
		return( NO ); 
		}
	}

if( norange ) {
	if( strcmp( type, "lazydate" )==0 ) setlazydates( 0, 0 );
	return( YES ); /* no need to check range.. */
	}

makedate( lowyr, 1, 1, "", lowdate );
makedate( hiyr, 12, 31, "", hidate );
daysdiff( lowdate, testdate, &diff );
daysdiff( testdate, hidate, &diff2 );
if( diff > 0 || diff2 > 0 ) {
	error( "Date's year is out of range. Min=%d, Max=%d", lowyr, hiyr );
	val_errno = 111;
	return( NO ); 
	}

if( strcmp( type, "lazydate" )==0 ) setlazydates( 0, 0 ); /* turn off lazy mode */
	
return( YES );
}

/* ============================ */
/* recentdate N  where N is the allowable number of days ago from the present. */
static int checkrecentdt( uin, typline )
char *uin, *typline;
{
int daysago, mon, day, year;
char todaystr[15];
int tmp;
long diff, jul;
int stat;

stat = jdate( uin, &jul );
if( stat != 0 ) {
	error( "Invalid date" ); 
	val_errno = 18;
	return( NO ); 
	}

sscanf( typline, "%*s %d", &daysago );
todaysdate( todaystr );

stat = daysdiff( todaystr, uin, &diff );
if( stat != 0 ) {
	error( "Invalid date" ); 
	val_errno = 18;
	return( NO ); 
	}
if( diff < 0 ) {
	error( "Date is later than today (%s).", todaystr );
	val_errno = 19;
	return( NO );
	}
else if( diff > daysago ) {
	error( "A date from within %d days of %s is expected.", daysago, todaystr );
	val_errno = 112;
	return( NO );
	}
else return( YES );
}
/* ============================ */
/* timems mm:ss MM:SS  where mm:ss is the minimum allowable time, and
	MM:SS is the maximum allowable time.  Colon must be specified in range parameters
	as well as input.  Both range paramters must be the same format and length.
	The first minute place may be omitted; if so the input is constrained accordingly.
 */
static int checktimems( uin, typline )
char *uin, *typline;
{
int nt;
char low[12], hi[12], tok[20];
int lowsec, hisec, insec, m, s, len;
nt = sscanf( typline, "%*s %s %s", low, hi );
if( nt != 2 || strlen( low ) != strlen( hi ) ) {
	err( 308, 0, "Invalid fdf timems specification", typline );
	return( NO );
	}
/* convert low to seconds */
sscanf( low, "%d:%d", &m, &s );
lowsec = (m * 60) + s;

/* convert hi to seconds */
sscanf( hi, "%d:%d", &m, &s );
hisec = (m * 60) + s;

/* convert input to seconds */
if( strlen( uin ) != strlen( low ) ) {
	error( "Response does not conform to required format, e.g. %s", low );
	val_errno = 120;
	return( NO );
	}
if( uin[ strlen( uin ) -3 ] != ':' ) {
	error( "Response does not conform to required format, e.g. %s (missing or misplaced colon)", low );
	val_errno = 120;
	return( NO );
	}
sscanf( uin, "%d:%d", &m, &s );
if( m > 59 || s > 59 ) {
	error( "Invalid mm:ss time" );
	val_errno = 121;
	return( NO );
	}
insec = (m * 60) + s;

if( insec < lowsec || insec > hisec ) {
	error( "Response out of range; range is %s to %s", low, hi );
	val_errno = 122;
	return( NO );
	}
return( YES );
}


/* ============================ */
static int checkfloat( uin, typline )
char uin[], typline[];
{
double atof(), min, max; 
char c, tok[ DATAMAXLEN+1 ], templat[ DATAMAXLEN+1 ]; 
int i, prec;

/* reconstruct template, since missings will obliterate sign and dec. pt. */
sscanf( typline, "%*s %s", tok );
goodnum( tok, &prec ); /* get precision */
for( i = 0; i < strlen( tok ); i++ ) {
	if( i == prec ) templat[i] = '.';
	else if( i == 0 && ( tok[i] == '+' || tok[i] == '-' ))templat[i] = 's';
	else templat[i] = '_';
	}
templat[i] = '\0';
if( ! goodnum( uin, &i ) ) {
	error( "Response is not a valid number" );
	val_errno = 128;
	return( NO );
	}
sscanf( typline, "%*s %lf %lf", &min, &max );
if(!( min <= atof( uin ) && atof( uin ) <= max )) { 
	error( "Response is out of legal range (%g - %g)", min, max ); 
	val_errno = 127;
	return( NO ); 
	}

if(  templat[0] == 's' && !(( uin[0]=='+' || uin[0]== '-' )) ) { 
	error( "A sign (+/-) is required (use + for 0 )" ); 
	val_errno = 124;
	return( NO ); 
	}
else if( templat[0] != 's' && ( uin[0] == '+' || uin[0] == '-' ) ) { 
	error( "Sign (+/-) is not permitted" ); 
	val_errno = 125;
	return( NO ); 
	}
for( i = 0; i < strlen( uin ); i++ ) {
	c = uin[i];

	if(( c != '.' && templat[i] == '.' )||
	   ( c == '.' && templat[i] != '.' )) { 
		error( "Response does not conform to required decimal format (%s)", templat ); 
		val_errno = 126;
		return( NO ); 
		}

	else if( (isdigit(c) && templat[i] == '_' ) ||
		 ( i == 0 && (c == '+' || c == '-') ) || 
		 ( templat[i] == '.' && c == '.' ) );  /* OK condition */

	else 	{ 
		error( "Response does not conform to required decimal format (%s)", templat ); 
		val_errno = 126;
		return( NO ); 
		}
	}

if( strlen( tok ) != strlen( uin ) ) { 
	error( "Response does not conform to the required length and format" ); 
	val_errno = 123;
	return( NO ); 
	}
return( YES );
}





/* ============================ */
static int checknumer( uin, typline )
char uin[], typline[];
{
int i, min, max;
char minstr[40], maxstr[40];
sscanf( typline, "%*s %d %d", &min, &max );
sscanf( typline, "%*s %s %s", minstr, maxstr );
for( i = 0; i < strlen( uin ); i++ ) {
	if( !(  isdigit( uin[i] )  )) { 
		error( "Response may contain numeric characters only" ); 
		val_errno = 130;
		return( NO ); 
		}
	}
if( !( min <= atoi( uin ) && atoi( uin ) <= max ) ) { 
	error( "Response is out of legal range. (%s - %s)", minstr, maxstr ); 
	val_errno = 131;
	return( NO ); 
	}

if( strlen( uin ) != strlen( minstr ) ) { 
	error( "Response does not conform to the correct length (e.g. %s)", minstr ); 
	val_errno = 132;
	return( NO ); 
	}
return( YES );
}





/* ============================ */
static int checklist( uin, typline )
char uin[], typline[];
{
char *getok(), type[ DATAMAXLEN+1 ], parm[ MAXLIST ][ 12 ] ;
int i, j, np;


i = 0; strcpy( type, getok( typline, &i ) ); j = i;
np = 0;
while( 1 ) {
	strcpy( parm[np], getok( typline, &i ));
	if( parm[np][0] == '\0' ) break;
	np++;
	}
for( i = 0; i < np; i++ ) if( strcmp( parm[i], uin )==0 ) break;
if( i == np ) { 
	if( typline[ strlen( typline ) -1 ] == '\n' )
		typline[ strlen( typline ) -1 ] = '\0'; /* strip off newline */
	error( "Response must be one of: %s", &typline[j] ); 
	val_errno = 135;
	return( NO ); 
	}
return( YES );
}





/* ============================ */
static int checklookup( in, typline )
char in[], typline[];
{
FILE *tfp;
char tab[ DATAMAXLEN+1 ], tblfilenm[ MAXPATH ], key[ DATAMAXLEN+1 ], line[ MAXCTREC ];
char type[15];
char f[8][DATAMAXLEN+1], key2[ 40 ], uin[12], buf[30];
int i, nf;

for( i = 0; i < 8; i++ ) f[i][0] = '\0';
nf = sscanf( typline, "%s %s %s %s %s %s %s %s %s %s",
	type, tab, f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[7] ) - 2;

#ifdef FIP
sprintf( tblfilenm, "../defs/%s", tab );
#else
sprintf( tblfilenm, "%s/%s", Lookupspath, tab );
#endif

tfp = fopen( tblfilenm, "r" );
if( tfp == NULL ) return( err( 304, 0, "Cannot find lookup table %s.", tab ) );

while( fgets( line, MAXCTREC, tfp ) != NULL ) {
	if( sscanf( line, "%s %s" , key, key2 ) < 1 ) continue;
		
	if( strcmp( key, in )==0 ) {
		for( i = 0; i < nf; i++ ) if( strcmp( f[i], key2 )==0 ) break;
		if( nf > 0 && i == nf ) continue;
		line[ strlen( line ) -1 ] = '\0';
		us_to_blank( line );
		message( "%s", line ); 
		fclose( tfp ); 
		return( YES );
		}
	}
fclose( tfp );
sprintf( buf, "%s %s %s %s %s %s %s %s",
	f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[7] );
error( "Response not in %s list.", tab );
val_errno = 140;

return( NO );
}


/* ============================ */
static int checkmulti( uin, typline )
char *uin;
char *typline;
{
int i, len;
sscanf( typline, "%*s %d", &len );
for( i = 0; i < len; i++ ) {
	if( uin[i] == 'x' ) uin[i] = '1';
	if( uin[i] == ' ' || uin[i] == '_' ) uin[i] = '0';
	if( uin[i] != '1' && uin[i] != '0' ) {
		error( "Invalid multichoice checkbox response." );
		val_errno = 146;
		return( NO );
		}
	}
return( YES );
}

#ifdef FIP
/* stubs to allow fipview to link */
custom_validation( uin, typestr )
char *uin;
char *typestr;
{ return( -1 ); }

error( msg )
char *msg;
{ return( 0 ); }

message( msg )
char *msg;
{ return( 0 ); }

#endif
