/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include "dms.x"

/* Routines for assigning and getting value of variables.  

	A "variable" can be either a data item (member of data array) 
	or a "tmp var" (a variable defined using declarevar().  

	These routines: 
		return 0 if ok, or an error code > 0.

		accept a recordid which is either 1) a literal recordid,
			2) a numeric field specifier @N, which indicates that the 
			recordid is available on the Nth field (e.g. @4 indicates 
			that the recordid is on the 4th field), or 3) "", for cases 
			where no accesses by name will be done.
*/

/* ======================== */
/* ASSIGNVALUE - set the named variable to the given value.
		Returns 0 if ok, 1 if itemname not recognized.  */

assignvalue( itemname, value, data, recordid )
char *itemname, *value;
char data[ MAXITEMS ][ DATAMAXLEN+1 ];	/* data array */
char *recordid;  /* see top of this file */
{
int j;
int ival;
int stat;
int p;

/* see if itemname is an integer field number; 
   if so set it and return.. scg 2-18-98 */
stat = goodnum( itemname, &p );
if( stat && p <= 0 ) {
	strcpy( data[ atoi( itemname ) -1 ], value );
	return( 0 );
	}

if( strcmp( recordid, "" )==0 ) j = -1;
#ifndef PLOTICUS
else if( recordid[0] == '@' ) {
	ival = atoi( &recordid[1] );
	if( ival > 0 && ival < MAXITEMS ) j = itemfld( data[ival-1], itemname );
	else return( 810 ); /* invalid @N */
	}
else j = itemfld( recordid, itemname );
#endif /* PLOTICUS */


if( j < 0 ) {
	stat = setvar( itemname, value );
	return( stat );
	}
else 	{
	if( strlen( value ) > DATAMAXLEN ) return( 808 ); /* value too long */
	strcpy( data[ j ] , value );
	}
return( 0 );
}

/* ========================= */
/* GETVALUE - get the value of the named variable.
 		Returns 0 if ok, 1 if itemname not recognized. */

getvalue( value, itemname, data, recordid )
char *value;
char *itemname;
char data[ MAXITEMS ][ DATAMAXLEN+1 ];	/* data array */
char *recordid;  /* see top of this file */
{
int j;
int stat;
int ival;

if( strcmp( recordid, "" )==0 ) j = -1;
#ifndef PLOTICUS
else if( recordid[0] == '@' ) {
	ival = atoi( &recordid[1] );
	if( ival > 0 && ival < MAXITEMS ) j = itemfld( data[ival-1], itemname );
	else return( 810 ); /* invalid @N */
	}
else j = itemfld( recordid, itemname );
#endif

if( j < 0 ) {
	stat = getvar( itemname, value );
	return( stat );
	}
else strcpy( value, data[ j ] );
return( 0 );
}
