/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


#include "dms.x"

/* VALUE_SUBST - take a text line and substitute values for variables.
		Variables may be: 
			Field numbers like @1, @2, @3, etc. (@1 is first field)
			Data item names like @age

		Returns 1 if one or more substitutions were done, 0 if there were no 
		substitutions necessary, or an error code > 0.

		scg 8/17/00  changed so that unrecognized @tokens are left as is
*/

static int involved[ MAXITEMS ];
static int ninvolved = 0;
static int translate_to_fn = 0;

value_subst( out, in, data, recordid, mode )
char *out; /* result buffer */
char *in;  /* input buffer */
char data[ MAXITEMS ][ DATAMAXLEN+1 ];	/* data array */
char *recordid; /* recordid or "" */
int mode;  /* either FOR_CONDEX, indicating that the line will be passed to condex(),
		   (minor hooks related to this) or NORMAL */
{
int i, k, f;
char itemname[512];
char value[255];
int len;
int found;
int stat;
int infunction;
int ifld;

strcpy( out, "" );
found = 0;
ninvolved = 0;

infunction = 0;
for( i = 0; i < strlen( in ); i++ ) {

	/* handle @@ (escape of @) */
	if( in[i] == '@' && in[i+1] == '@' ) {
		strcat( out, "@" );
		i ++;
		continue;
		}
		
	/* @item or @1 */
	if( in[i] == '@' ) {

		sscanf( &in[i+1], "%s", itemname );
		found = 1;

		/* truncate itemname at first char which is not alphanumeric or _ */
		for( k = 0; k < strlen( itemname ); k++ ) {
			if( !isalnum( itemname[k] ) && itemname[k] != '_' ) { 
				itemname[k] = '\0';
				break;
				}
			}

		/* @1, @2, etc... */
		ifld = atoi( itemname );
		if( ifld > 0 && ifld < MAXITEMS ) {
			if( translate_to_fn ) sprintf( value, "@%d", ifld );
			else strcpy( value, data[ ifld-1 ] );
			involved[ ninvolved++ ] = ifld-1; /* add it to list of involved fields */
			}

		/* @itemname or @varname.. */
		else	{
			stat = getvalue( value, itemname, data, recordid );
			if( stat != 0 ) /* return( stat );  scg 8/17/00  */
				sprintf( value, "@%s", itemname );  /* replace @token on stream unconverted.. */


			/* do the following to add item to list of involved fields */
#ifndef PLOTICUS
			ifld = itemfld( recordid, itemname );
			if( ifld >= 0 ) involved[ ninvolved++ ] = ifld; 

			if( translate_to_fn ) sprintf( value, "@%d", ifld+1 ); /* supercedes
								effect of the previous lines*/
#endif
			}

		/* special case of 0 length data item when in a condex expression but 
		   not within a function arg list.. to prevent condex syntax errors  */
		if( strcmp( value, NULLDATA )==0 && mode == FOR_CONDEX && !infunction )
			strcpy( value, "_null_" ); 


		strcat( out, value );

		i+= strlen( itemname ); /* advance past @itemname */
		}

	/* absolute column specifier - \cN */
	else if( strncmp( &in[i], "\\c", 2 )== 0 ) {
		int col;
		if( ! isdigit( in[i+3] )) 
			err( 15080, 0, "warning: extra characters after \\c", "" );

		else	{
			sscanf( &in[i+2], "%d", &col );
			for( ; len < col; len++ ) out[ len ] = ' ';
			out[ len ] = '\0';
			if( col < 10 ) i+=3;
			else if( col < 100 ) i+= 4;
			}
		}
		
	else	{
		if( in[i] == '$' && isalpha( in[i+1] ) ) infunction = 1;
		if( isspace( in[i] ) ) infunction = 0;
		len = strlen( out );
		out[ len ] = in[i];
		out[ len +1 ] = '\0';
		}
	}
return( found );
}

/* ========================= */
/* GET_INVOLVED_ITEMS  - allow an application to access the field #s of all 
	data items involved in the most recent value_subst call.  Tmp vars
	are not included since they have no field # */
get_involved_items( n, list )
int *n;
int list[ MAXITEMS ];
{
int i;
*n = ninvolved;
for( i = 0; i < ninvolved; i++ ) list[i] = involved[i];
return( 0 );
}

/* ========================= */
/* VS_TRANSLATEMODE - tell value_subst to translate @itemnames to @N (field position)
 	rather than a data value.  */
vs_translatemode( mode )
int mode; /* 1 = on   0 = off */
{
translate_to_fn = mode;
return( 0 );
}

