/* ploticus data display engine.  Software, documentation, and examples.  
 * Copyright 1998-2001 Stephen C. Grubb  (steve@sgpr.net).
 * Covered by GPL; see the file ./Copyright for details. */


/* VARIABLE - maintain named variables  */

#include "dms.d"
#define MAXVAR 128
#define VARMAXLEN 128

static int Ns = 0;
static char Name[MAXVAR][TAGMAXLEN+1];
static char Value[MAXVAR][VARMAXLEN+1];
static int badvarwarn = 0;

/* =================================== */
/* SETVAR - set the named variable to the given value */
setvar( name, value )
char *name, *value;
{
int i;
for( i = 0; i < Ns; i++ ) if( strcmp( Name[i], name )==0 ) break;
if( i == Ns ) {
	if( badvarwarn ) err( 15090, 0, "Undeclared variable", name );
	return( 801 ); /* variable not declared */
	}
if( strlen( value ) > VARMAXLEN ) return( 808 ); /* value too long */
strcpy( Value[i], value );
return( 0 );
}
/* =================================== */
/* GETVAR - get the value of the named variable */
getvar( name, value )
char *name, *value;
{
int i;
for( i = 0; i < Ns; i++ ) if( strcmp( Name[i], name )==0 ) break;
if( i == Ns ) {
	if( badvarwarn )  err( 15090, 0, "Undeclared variable", name );
	return( 802 ); /* variable not declared */
	}
strcpy( value, Value[i] );
return( 0 );
}
/* =================================== */
/* DECLAREVAR - declare a variable and initialize it to null.  Repeat declarations
	are not considered errors though they do null out the value. */
declarevar( name )
char *name;
{
int i;
if( strlen( name ) < 1 || strlen( name ) > TAGMAXLEN ) return( 845 ); /* name too long */
if( Ns >= MAXVAR ) { 	 /* var var table full */
	if( badvarwarn ) err( 16000, 0, "Var table full", name );
	return( 846 );
	}
for( i = 0; i < Ns; i++ ) if( strcmp( Name[i], name )==0 ) break;
if( i != Ns ) {  /* variable already declared - null it out and return.. */
	setvar( name, "" );
	return( 0 ); 
	}
strcpy( Name[Ns], name );
Ns++;
setvar( name, "" );
return( 0 );
}
/* =================================== */
/* CLEARALLVAR - clear the list of variables. */
clearallvar( )
{
Ns = 0;
return( 0 );
}

/* ===================================== */
/* WARN_ON_BADVAR - turn on warning when undeclared vars are encountered. */
warn_on_badvar()
{
badvarwarn = 1;
return( 0 );
}
